//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   26 Dec 05  Brian Frank  Creation
//   19 Aug 06  Brian Frank  Ported from Java to Fan
//

**
** FType is the read/write fcode representation of sys::Type.
**
class FType : FConst
{

//////////////////////////////////////////////////////////////////////////
// Constructor
//////////////////////////////////////////////////////////////////////////

  new make(FPod pod)
  {
    this.pod = pod
    this.facets = FFacet[,]
  }

//////////////////////////////////////////////////////////////////////////
// Meta IO
//////////////////////////////////////////////////////////////////////////

  Void writeMeta(OutStream out)
  {
    out.writeI2(self)
    out.writeI2(base)
    out.writeI2(mixins.size)
    mixins.each |Int m| { out.writeI2(m) }
    out.writeI4(flags & FlagsMask)
  }

  FType readMeta(InStream in)
  {
    self   = in.readU2
    base   = in.readU2
    mixins = Int[,]
    in.readU2.times |,| { mixins.add(in.readU2) }
    flags  = in.readU4
    return this
  }

//////////////////////////////////////////////////////////////////////////
// Body IO
//////////////////////////////////////////////////////////////////////////

  Uri uri()
  {
    return Uri.parse("/" + pod.name(pod.typeRef(self).typeName) + ".fcode")
  }

  Void write()
  {
    out := pod.out(uri)

    out.writeI2(fields.size)
    fields.each |FField f| { f.write(out) }

    out.writeI2(methods.size)
    methods.each |FMethod m| { m.write(out) }

    out.writeI2(facets.size)
    facets.each |FFacet f| { f.write(out) }

    out.close
  }

  Void read()
  {
    in := pod.in(uri)

    fields = FField[,]
    in.readU2.times |,| { fields.add(FField.make.read(in)) }

    methods = FMethod[,]
    in.readU2.times |,| { methods.add(FMethod.make.read(in)) }

    f := facets = FFacet[,]
    in.readU2.times |,| { f.add(FFacet.make.read(in)) }

    in.close
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  Bool hollow := true   // have we only read meta-data
  FPod pod              // parent pod
  Int self              // self typeRef index
  Int flags             // bitmask
  Int base              // base typeRef index
  Int[] mixins          // mixin typeRef indexes
  FField[] fields       // fields
  FMethod[] methods     // methods
  FFacet[] facets       // facets on type itself

}