//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   29 Aug 06  Brian Frank  Creation
//

**
** ImportedSlot is the implementation of CSlot for a slot imported
** from a precompiled pod (as opposed to a SlotDef within the
** compilation units being compiled).
**
abstract class ImportedSlot : CSlot
{
  new make(CType parent, Slot slot)
  {
    this.parent = parent
    this.flags  = (Int)slot->flags // undocumented trap
  }

  override readonly CType parent
  override Str name()      { return slot.name }
  override Str qname()     { return slot.qname }
  override Str signature() { return slot.signature }
  override readonly Int flags
  override Str toStr()     { return signature }

  abstract Slot slot()

}

**************************************************************************
** ImportedField
**************************************************************************

class ImportedField : ImportedSlot, CField
{
  new make(CType parent, Field f)
    : super(parent, f)
  {
    this.f = f
    this.fieldType = ns.importType(f.fieldType)
    if (f.getter != null) this.getter = ns.importMethod(f.getter)
    if (f.setter != null) this.setter = ns.importMethod(f.setter)
  }

  override Slot slot() { return f }

  override readonly CType fieldType
  override readonly CMethod getter
  override readonly CMethod setter
  readonly Field f
}

**************************************************************************
** ImportedMethod
**************************************************************************

class ImportedMethod : ImportedSlot, CMethod
{
  new make(CType parent, Method m)
    : super(parent, m)
  {
    this.m = m
    this.returnType = ns.importType(m.returnType)
    m.params.map(this.params = CParam[,]) |MethodParam p->Obj| { return ImportedParam.make(ns, p) }

    if (!parent.isGeneric) isGeneric = false
    else
    {
      isGeneric = returnType.isGenericParameter
      params.each |CParam p| { isGeneric = isGeneric || p.paramType.isGenericParameter }
    }
  }

  override Slot slot() { return m }

  override CType inheritedReturnType()
  {
    // use trap to access undocumented hook
    if (isOverride)
      return ns.importType((Type)m->inheritedReturnType)
    else
      return returnType
  }

  override readonly CType returnType
  override readonly CParam[] params
  override readonly Bool isGeneric
  readonly Method m
}

**************************************************************************
** ImportedParam
**************************************************************************

class ImportedParam : CParam
{
  new make(Namespace ns, MethodParam p)
  {
    this.p = p
    this.paramType = ns.importType(p.paramType)
  }

  override Str name() { return p.name }
  override Bool hasDefault() { return p.hasDefault }

  override readonly CType paramType
  readonly MethodParam p
}