//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   23 Sept 06  Brian Frank  Creation
//

**
** InheritTest
**
class InheritTest : CompilerTest
{

//////////////////////////////////////////////////////////////////////////
// OrderByInheritance
//////////////////////////////////////////////////////////////////////////

  Void testOrderByInheritance()
  {
    compile("
      class B : A {}
      class D : C {}
      class A {}
      class C : B {}")
     verifyEq(compiler.types[0].name, "A")
     verifyEq(compiler.types[1].name, "B")
     verifyEq(compiler.types[2].name, "C")
     verifyEq(compiler.types[3].name, "D")

    compile("
      mixin MABC : MAB, MC {}
      mixin MAB : MA, MB {}
      mixin MB {}
      mixin MBD : MD, MB {}
      mixin MC {}
      mixin MBC : MB, MC {}
      mixin MA {}
      mixin MD {}
      ")
     //echo(compiler.types.join("\n"))
     verifyEq(compiler.types[0].name, "MA")
     verifyEq(compiler.types[1].name, "MB")
     verifyEq(compiler.types[2].name, "MAB")
     verifyEq(compiler.types[3].name, "MC")
     verifyEq(compiler.types[4].name, "MABC")
     verifyEq(compiler.types[5].name, "MD")
     verifyEq(compiler.types[6].name, "MBD")
     verifyEq(compiler.types[7].name, "MBC")

    // cyclic inheritance errors
    verifyErrors("
      class A : A {}
      class B : C {}
      class C : D {}
      class D : B {}
      mixin X : Z {}
      mixin Y : X {}
      mixin Z : X {}
      ",
       [2, 7, "Cyclic inheritance for 'A'",
        3, 7, "Cyclic inheritance for 'B'",
        6, 7, "Cyclic inheritance for 'X'",
       ])
  }

//////////////////////////////////////////////////////////////////////////
// CheckInheritance
//////////////////////////////////////////////////////////////////////////

  Void testCheckInheritance()
  {
    verifyErrors("
      class C {}
      internal mixin M {}

      mixin C1 : C {}
      mixin C2 : Buf {}

      class D1 : C[] {}
      class D2 : M:C {}
      class D3 : Str[] {}

      class E : |Int x->M| {}
      mixin F : Int:C {}

      class G : Str {}
      class H : ZipEntryFile {}
      class I : M {} // OK!

      enum J : G { a }

      class K : C1, G {}
      class L : C1, Test, C2 {}
      class N : G, C1, K {}
      ",
       [
        5, 7, "Mixin 'C1' cannot extend class '$podName::C'",
        6, 7, "Mixin 'C2' cannot extend class 'sys::Buf'",
        8, 7, "Class 'D1' cannot extend parameterized type '$podName::C[]'",
        9, 7, "Class 'D2' cannot extend parameterized type '[$podName::M:$podName::C]'",
       10, 7, "Class 'D3' cannot extend parameterized type 'sys::Str[]'",
       12, 7, "Class 'E' cannot extend parameterized type '|sys::Int->$podName::M|'",
       13, 7, "Mixin 'F' cannot extend class '[sys::Int:$podName::C]'",
       13, 7, "Class 'F' cannot extend parameterized type '[sys::Int:$podName::C]'",
       15, 7, "Class 'G' cannot extend final class 'sys::Str'",
       16, 7, "Class 'H' cannot access internal scoped class 'sys::ZipEntryFile'",
       19, 7, "Enum 'J' cannot extend class '$podName::G'",
       21, 7, "Invalid inheritance order, ensure class '$podName::G' comes first before mixins",
       22, 7, "Invalid inheritance order, ensure class 'sys::Test' comes first before mixins",
       23, 7, "Class 'N' cannot mixin class '$podName::K'",
       ])
  }

//////////////////////////////////////////////////////////////////////////
// InheritErrors
//////////////////////////////////////////////////////////////////////////

  Void testInheritErrors()
  {
    verifyErrors("
       class A
       {
         new ctor() { return }
         Bool a() { return false }
         virtual Bool b() { return false }
         virtual Bool c() { return false }
         virtual Void d(Bool b) { }
         private Void e() {}
         virtual Int f () { return 9 }
         Bool g
         virtual Int h
         abstract Bool i
       }

       class B : A
       {
         static B ctor() { return null } // ok
         Bool a() { return true }
         Bool b() { return true }
         override Int c() { return 0 }
         override Void d(Bool b, Int x) { }
         private Void e(Int x, Int y) {} // ok
         override Str f
         override Str g
         Int h
         override Str i
       }

       class C : Buf
       {
         override Bool empty() { return false }
         override Bool more(Str x) { return false }
         override Void pos() { }
         Int size() { return 0 }
       }

       mixin ConflictX
       {
         static Void a() {}
         virtual Int b() { return 3 }
         virtual Void c(Str q) { }
         virtual Void d(Str q) { }
       }
       mixin ConflictY
       {
         static Void a() {}
         virtual Bool b() { return true }
         virtual Void c(Str q, Int p) { }
         virtual Void d(Int q) { }
       }
       class Conflict : ConflictX, ConflictY {}

       mixin WhichX { virtual Str name() { return null } }
       mixin WhichY { virtual Str name() { return null } }
       class Which : WhichX, WhichY {}

       class Ouch
       {
         override Void figgle() {}
         override Str foogle
       }
    ",
       [// internal to compilation unit
        19, 10, "Cannot override non-virtual slot '$podName::A.a'",
        20, 10, "Must specify override keyword to override '$podName::A.b'",
        21, 10, "Return type mismatch in override of '$podName::A.c' - 'sys::Bool' != 'sys::Int'",
        22, 10, "Parameter mismatch in override of '$podName::A.d' - 'd(sys::Bool)' != 'd(sys::Bool, sys::Int)'",
        24, 10, "Type mismatch in override of '$podName::A.f' - 'sys::Int' != 'sys::Str'",
        25, 10, "Cannot override non-virtual slot '$podName::A.g'",
        26, 10, "Must specify override keyword to override '$podName::A.h'",
        27, 10, "Type mismatch in override of '$podName::A.i' - 'sys::Bool' != 'sys::Str'",

        // imported inherits
        32, 10, "Cannot override non-virtual slot 'sys::Buf.empty'",
        33, 10, "Parameter mismatch in override of 'sys::Buf.more' - 'more()' != 'more(sys::Str)'",
        34, 10, "Return type mismatch in override of 'sys::Buf.pos' - 'sys::Int' != 'sys::Void'",
        35, 10, "Must specify override keyword to override 'sys::Buf.size'",

        // conflicts
        52, 8, "Inheritance conflict '$podName::ConflictX.a' and '$podName::ConflictY.a'",
        52, 8, "Inherited slots have conflicting signatures '$podName::ConflictX.b' and '$podName::ConflictY.b'",
        52, 8, "Inherited slots have conflicting signatures '$podName::ConflictX.c' and '$podName::ConflictY.c'",
        52, 8, "Inherited slots have conflicting signatures '$podName::ConflictX.d' and '$podName::ConflictY.d'",
        56, 8, "Must override ambiguous inheritance '$podName::WhichX.name' and '$podName::WhichY.name'",

        // overrides of unknown virtuals
        60, 10, "Override of unknown virtual slot 'figgle'",
        61, 10, "Override of unknown virtual slot 'foogle'",
       ])
  }

//////////////////////////////////////////////////////////////////////////
// InheritProtection
//////////////////////////////////////////////////////////////////////////

  Void testInheritProtection()
  {
    verifyErrors("
      class B : A
      {
        override public    Void a1() {} // ok
        override protected Void a2() {}
        override internal  Void a3() {}
        override private   Void a4() {}

        override public    Void b1() {} // ok
        override protected Void b2() {} // ok
        override internal  Void b3() {}
        override private   Void b4() {}

        override public    Void c1() {} // ok
        override protected Void c2() {}
        override internal  Void c3() {} // ok
        override private   Void c4() {}

        override public Void d() {}  // unknown
      }

      class A
      {
        virtual public Void a1() {}
        virtual public Void a2() {}
        virtual public Void a3() {}
        virtual public Void a4() {}

        virtual protected Void b1() {}
        virtual protected Void b2() {}
        virtual protected Void b3() {}
        virtual protected Void b4() {}

        virtual internal Void c1() {}
        virtual internal Void c2() {}
        virtual internal Void c3() {}
        virtual internal Void c4() {}

        private Void d() {}
      }
     ",
     [
       5, 9, "Override narrows protection scope of '$podName::A.a2'",
       6, 9, "Override narrows protection scope of '$podName::A.a3'",
       7, 9, "Override narrows protection scope of '$podName::A.a4'",

      11, 9, "Override narrows protection scope of '$podName::A.b3'",
      12, 9, "Override narrows protection scope of '$podName::A.b4'",

      15, 9, "Override narrows protection scope of '$podName::A.c2'",
      17, 9, "Override narrows protection scope of '$podName::A.c4'",

      19, 9, "Override of unknown virtual slot 'd'",
     ])
  }

//////////////////////////////////////////////////////////////////////////
// IntraUnitInherit
//////////////////////////////////////////////////////////////////////////

  Void testIntraUnitInherit()
  {
    compile("
      class A
      {
        static Str sa() { return \"sa\" }
        Str ia() { return \"ia\" }
        virtual Str vx() { return \"A.vx\" }
      }

      class B : A
      {
        static Str sb() { return sa }
        Str ib() { return ia }
        override Str vx() { return \"B.vx\" }
      }")
     a := compiler.types[0]
     b := compiler.types[1]
     verifyEq(a.name, "A")
     verifyEq(b.name, "B")
     verifyEq(b.base, a)
     verifyEq(b.slot("ib").parent, b)
     verifyEq(b.slot("ia").parent, a)
     verifyEq(b.slot("vx").parent, b)

     typeB := pod.types[1]
     objB := typeB.make
     verifyEq(typeB.method("sb").call([,]), "sa")
     verifyEq(typeB.method("ib").call([objB]), "ia")
     verifyEq(typeB.method("vx").call([objB]), "B.vx")
  }

//////////////////////////////////////////////////////////////////////////
// MultipleInherit
//////////////////////////////////////////////////////////////////////////

  Void testMultipleInherit()
  {
    compile("
      mixin A
      {
        abstract Str a()
        abstract Str b()
        override Int hash() { return 77 }
      }

      class B : A
      {
        override Str a() { return \"B.a\" }
        override Str b := \"B.b\"
      }

      mixin C : A      {
        override Str toStr() { return \"C.toStr\" }
      }

      class D : B, C
      {
      }
      ")

     // verify D's slot map
     astD := compiler.types[3]
     verifyEq(astD.name, "D")
     verifyEq(astD.slot("a").parent.name, "B")      // inherit concrete one
     verifyEq(astD.slot("b").parent.name, "B")      // inherit concrete one
     verifyEq(astD.slot("hash").parent.name, "A")   // inherit most specific one
     verifyEq(astD.slot("toStr").parent.name, "C")  // inherit most specific one

     // verify code works
     typeD := pod.types[3]
     objD := typeD.make
     verifyEq(typeD.name, "D")
     verifyEq(typeD.method("a").call([objD]), "B.a")
     verifyEq(typeD.method("b").call([objD]), "B.b")  // TODO - reflect should be Field
     verifyEq(typeD.method("hash").call([objD]), 77)
     verifyEq(typeD.method("toStr").call([objD]), "C.toStr")
  }

//////////////////////////////////////////////////////////////////////////
// Covariance
//////////////////////////////////////////////////////////////////////////

  Void testCovariance()
  {
    compile("
      mixin Q
      {
        virtual Q m() { return null }
        virtual Q n() { return null }

        virtual Obj[] e() { return [2ns, 3] }
        virtual Num[] f() { return [2, 3.0] }
      }

      mixin P : Q
      {
        override virtual P n() { return this }
        override virtual Int[] f() { return [2, 3] }
      }"
    )
    TypeDef q := compiler.types[0]
    TypeDef p := compiler.types[1]

    Type qt := pod.types[0]
    Type pt := pod.types[1]

    compile("
      using $q.pod.name

      class A
      {
        virtual A x() { return null }
      }

      mixin M
      {
        virtual M y() { return null }
      }

      class B : A, M, P
      {
        override B x() { return this }
        static A xA(A a) { return a.x }
        static B xB(B b) { return b.x }

        override B y() { return this }
        static M yM(M a) { return a.y }
        static B yB(B b) { return b.y }

        override B m() { return this }
        static Q mQ(Q q) { return q.m }
        static B mB(B b) { return b.m }

        override B n() { return this }
        static Q nQ(Q q) { return q.n }
        static P nP(P p) { return p.n }
        static B nB(B b) { return b.n }

        override Str[] e() { return [\"hello\"] }
        static Obj[] eQ(Q q) { return q.e }
        static Str[] eB(B b) { return b.e }

        static Num[] fQ(Q q) { return q.f }
        static Int[] fP(P p) { return p.f }
        static Int[] fB(B b) { return b.f }
      }
      ")

     // compiler.fpod.dump

     //
     // verify CMethod.inheritedType works correctly
     //

     TypeDef a := compiler.types[0]
     TypeDef m := compiler.types[1]
     TypeDef b := compiler.types[2]

     verifyEq(q.name, "Q")
     verifyEq(q.method("m").returnType.name, "Q")
     verifyEq(q.method("m").inheritedReturnType.name, "Q")

     verifyEq(p.name, "P")
     verifyEq(p.method("n").returnType.name, "P")
     verifyEq(p.method("n").inheritedReturnType.name, "Q")
     verifyEq(p.method("f").returnType.signature, "sys::Int[]")
     verifyEq(p.method("f").inheritedReturnType.signature, "sys::Num[]")

     verifyEq(a.name, "A")
     verifyEq(a.method("x").returnType.name, "A")
     verifyEq(a.method("x").inheritedReturnType.name, "A")

     verifyEq(m.name, "M")
     verifyEq(m.method("y").returnType.name, "M")
     verifyEq(m.method("y").inheritedReturnType.name, "M")

     verifyEq(b.name, "B")
     verifyEq(b.method("x").returnType.name, "B")
     verifyEq(b.method("x").inheritedReturnType.name, "A")
     verifyEq(b.method("y").returnType.name, "B")
     verifyEq(b.method("y").inheritedReturnType.name, "M")
     verifyEq(b.method("m").returnType.name, "B")
     verifyEq(b.method("m").inheritedReturnType.name, "Q")
     verifyEq(b.method("n").returnType.name, "B")
     verifyEq(b.method("n").inheritedReturnType.name, "Q")
     verifyEq(b.method("e").returnType.signature, "sys::Str[]")
     verifyEq(b.method("e").inheritedReturnType.signature, "sys::Obj[]")
     verifyEq(b.method("f").returnType.signature, "sys::Int[]")
     verifyEq(b.method("f").inheritedReturnType.signature, "sys::Num[]")

     //
     // verify reflection
     //
     Type at := pod.types[0]
     Type mt := pod.types[1]
     Type bt := pod.types[2]

     verifySame(at.method("x").returnType, at)
     verifySame(mt.method("y").returnType, mt)
     verifySame(qt.method("m").returnType, qt)
     verifySame(qt.method("n").returnType, qt)
     verifySame(qt.method("e").returnType, Obj[].type)
     verifySame(qt.method("f").returnType, Num[].type)
     verifySame(pt.method("m").returnType, qt)
     verifySame(pt.method("n").returnType, pt)
     verifySame(pt.method("e").returnType, Obj[].type)
     verifySame(pt.method("f").returnType, Int[].type)
     verifySame(bt.method("x").returnType, bt)
     verifySame(bt.method("y").returnType, bt)
     verifySame(bt.method("n").returnType, bt)
     verifySame(bt.method("m").returnType, bt)
     verifySame(bt.method("e").returnType, Str[].type)
     verifySame(bt.method("f").returnType, Int[].type)

     //
     // verify methods work
     //
     obj := bt.make
     verifySame(bt.method("xA").call1(obj), obj)
     verifySame(bt.method("xB").call1(obj), obj)
     verifySame(bt.method("yM").call1(obj), obj)
     verifySame(bt.method("yB").call1(obj), obj)
     verifySame(bt.method("mQ").call1(obj), obj)
     verifySame(bt.method("mB").call1(obj), obj)
     verifySame(bt.method("nQ").call1(obj), obj)
     verifySame(bt.method("nP").call1(obj), obj)
     verifySame(bt.method("nB").call1(obj), obj)
     verifyEq(bt.method("eQ").call1(obj), ["hello"])
     verifyEq(bt.method("eB").call1(obj), ["hello"])
     verifyEq(bt.method("fQ").call1(obj), [2, 3])
     verifyEq(bt.method("fP").call1(obj), [2, 3])
     verifyEq(bt.method("fB").call1(obj), [2, 3])
  }

  Void testCovarianceMore()
  {
    compile("
      abstract class AA
      {
        abstract Obj y()
      }

      abstract class A : AA
      {
        abstract Obj x()
        override abstract Num y()
      }

      mixin M
      {
        abstract Obj x()
        virtual Obj y() { return 4 }
      }

      class B : A, M
      {
        override B x() { return this }
        static Obj xA(A a) { return a.x }
        static Obj xM(M m) { return m.x }
        static B xB(B b) { return b.x }

        override Num y() { return 8 }
        static Obj yAA(AA a) { return a.y }
        static Num yA(A a) { return a.y }
        static Obj yM(M m) { return m.y }
        static Num yB(B b) { return b.y }
      }
      ")

     // compiler.fpod.dump

     Type bt := pod.types[3]

     obj := bt.make
     verifySame(bt.method("xA").call1(obj), obj)
     verifySame(bt.method("xM").call1(obj), obj)
     verifySame(bt.method("xB").call1(obj), obj)

     verifySame(bt.method("yAA").call1(obj), 8)
     verifySame(bt.method("yA").call1(obj), 8)
     verifySame(bt.method("yM").call1(obj), 8)
     verifySame(bt.method("yB").call1(obj), 8)
  }

  Void testCovarianceConflict()
  {
    verifyErrors("
      abstract class AA
      {
        abstract Obj f()
      }

      abstract class A : AA
      {
        abstract A x()
        abstract Obj y()
        abstract A z()
        virtual override Num f() { return 3 }
      }

      mixin M
      {
        abstract M x()
        abstract A y()
        abstract Obj z()
        abstract Num f()
      }

      class B : A, M
      {
        override B x() { return this }
        override A y() { return this }
        override B z() { return this }
      }
     ", [
       23, 7, "Inherited slots have conflicting signatures '$podName::A.f' and '$podName::M.f'",
       25, 9, "Conflicting covariant returns: '$podName::A' and '$podName::M'",
       26, 9, "Conflicting covariant returns: 'sys::Obj' and '$podName::A'",
       27, 9, "Conflicting covariant returns: '$podName::A' and 'sys::Obj'",
     ])
  }

//////////////////////////////////////////////////////////////////////////
// FieldInherit
//////////////////////////////////////////////////////////////////////////

  ** abstract method -> field
  Void testFieldInherit1()
  {
    verifyFieldInherit("
        abstract Int p0()
        abstract Int p1()
        abstract Int p2()
        abstract Int p3()
     ", false)
   }

  ** virtual method -> field
  Void testFieldInherit2()
  {
    verifyFieldInherit("
        virtual Int p0() { return 99 }
        virtual Int p1() { return 99 }
        virtual Int p2() { return 99 }
        virtual Int p3() { return 99 }
     ", false)
   }

  ** abstract field -> field
  Void testFieldInherit3()
  {
    verifyFieldInherit("
        abstract Int p0
        abstract Int p1
        abstract Int p2
        abstract Int p3
     ", true)
   }

  ** field -> field (tested in FieldTest - we can't really
  ** use this test approach because the subclass doesn't get
  ** direct accesss to parent class storage

   private Void verifyFieldInherit(Str aFields, Bool doSeta)
   {
     doVerifyFieldInherit(true, aFields, doSeta)
     doVerifyFieldInherit(false, aFields, doSeta)
   }

   private Void doVerifyFieldInherit(Bool aIsClass, Str aFields, Bool doSeta)
   {
     aHeader  := aIsClass ? "abstract class" : "mixin"

     setaCode := !doSeta ? "" :
     "
        static Void seta(A a, Str:Int m)
        {
          a.p0 = m[\"p0\"]
          a.p1 = m[\"p1\"]
          a.p2 = m[\"p2\"]
          a.p3 = m[\"p3\"]
        }
     "

     compile("
      $aHeader A
      {
        $aFields
      }

      class B : A
      {
        static Str:Int geta(A a)
        {
          return [\"p0\":a.p0, \"p1\":a.p1, \"p2\":a.p2, \"p3\":a.p3]
        }

        static Str:Int getb(B b)
        {
          return [\"p0\":b.@p0, \"p1\":b.@p1, \"p2\":b.@p2, \"p3\":b.@p3]
        }

        $setaCode

        static Void setb(B b, Str:Int m)
        {
          b.@p0 = m[\"p0\"]
          b.@p1 = m[\"p1\"]
          b.@p2 = m[\"p2\"]
          b.@p3 = m[\"p3\"]
        }

        static Bool getCounts(B b, Int expected)
        {
          // echo(\"getCounts \$b.p1_get \$b.p2_get\")
          return b.p1_get == expected && b.p2_get == expected
        }

        static Bool setCounts(B b, Int expected)
        {
          // echo(\"setCounts \$b.p2_set \$b.p3_set\")
          return b.p2_set == expected && b.p3_set == expected
        }

        // abstract method -> field
        override Int p0 := 101
        override Int p1 := 102 { get { p1_get++; return @p1 } }
        override Int p2 := 103 { get { p2_get++; return @p2 } set { p2_set++; @p2 = val } }
        override Int p3 := 104 { set { p3_set++; @p3 = val } }
        Int p1_get := 0
        Int p2_get := 0
        Int p2_set := 0
        Int p3_set := 0
      }
      ")

     a := pod.types[0]
     b := pod.types[1]
     geta := b.method("geta")
     getb := b.method("getb")
     seta := b.method("seta", false)
     setb := b.method("setb")
     getCounts := b.method("getCounts")
     setCounts := b.method("setCounts")

     orig := ["p0":101, "p1":102, "p2":103, "p3":104]
     change1 := (Str:Int)orig.map(Str:Int[:]) |Int v->Int| { return 1000+v }
     change2 := (Str:Int)orig.map(Str:Int[:]) |Int v->Int| { return 2000+v }

     obj := b.make
     verifyEq(getCounts.call2(obj, 0), true)
     verifyEq(getCounts.call2(obj, 0), true)

     verifyEq(geta.call1(obj), orig)
     verifyEq(getCounts.call2(obj, 1), true)

     verifyEq(getb.call1(obj), orig)
     verifyEq(getCounts.call2(obj, 1), true)

     setb.call2(obj, change1)
     verifyEq(setCounts.call2(obj, 0), true)

     verifyEq(geta.call1(obj), change1)
     verifyEq(getCounts.call2(obj, 2), true)

     verifyEq(getb.call1(obj), change1)
     verifyEq(getCounts.call2(obj, 2), true)

     if (doSeta)
     {
       seta.call2(obj, change2)
       verifyEq(setCounts.call2(obj, 1), true)

       verifyEq(geta.call1(obj), change2)
       verifyEq(getCounts.call2(obj, 3), true)

       verifyEq(getb.call1(obj), change2)
       verifyEq(getCounts.call2(obj, 3), true)
     }
  }

//////////////////////////////////////////////////////////////////////////
// NamedSuper
//////////////////////////////////////////////////////////////////////////

  Void testNamedSuper()
  {
    compile("
      mixin X
      {
        virtual Str f(Int a) { return \"X.f\" }
        virtual Str s(Int a) { return \"X.s\" }
      }

      mixin Y
      {
        virtual Str f(Int a) { return \"Y.f\" }
      }

      mixin Z : X
      {
        virtual override Str toStr() { return \"Z.toStr\" }
        virtual override Str s(Int a)
        {
          switch (a)
          {
            case 'Z': return \"Z.s\"
            case 'X': return X.super.s(a)
          }
          throw Err.make
        }
      }

      class A
      {
        virtual Str f(Int a) { return \"A.f\" }
      }

      class B : A
      {
        virtual override Str f(Int a) { return \"B.f\" }
      }

      class C : B, Z, Y
      {
        virtual override Str f(Int a)
        {
          switch (a)
          {
            case 'C': return \"C.f\"
            case 'S': return super.f(a)
            case 'B': return B.super.f(a)
            case 'A': return A.super.f(a)
            case 'X': return X.super.f(a)
            case 'Y': return Y.super.f(a)
          }
          throw Err.make
        }

        Str g(Int a)
        {
          switch (a)
          {
            case 'C': return \"C.f\"
            case 'S': return super.f(a)
            case 'B': return B.super.f(a)
            case 'A': return A.super.f(a)
            case 'X': return X.super.f(a)
            case 'Y': return Y.super.f(a)
          }
          throw Err.make
        }

        Str h(Int a)
        {
          switch (a)
          {
            case 'C': return ((C)this).f(a)
            case 'S': return ((B)this).f(a)
            case 'B': return ((B)this).f(a)
            case 'A': return ((A)this).f(a)
            case 'X': return ((X)this).f(a)
            case 'Y': return ((Y)this).f(a)
          }
          throw Err.make
        }
      }
      ")

     // compiler.fpod.dump

     Type t := pod.types[5]
     verifyEq(t.name, "C")
     obj := t.make;

     ["f", "g", "h"].each |Str m|
     {
       verifyEq(t.method(m).callOn(obj, ['C']), "C.f")
       verifyEq(t.method(m).callOn(obj, ['S']), "B.f")
       verifyEq(t.method(m).callOn(obj, ['B']), "B.f")
       verifyEq(t.method(m).callOn(obj, ['A']), "A.f")
       verifyEq(t.method(m).callOn(obj, ['X']), "X.f")
       verifyEq(t.method(m).callOn(obj, ['Y']), "Y.f")
     }

     verifyEq(t.method("s").callOn(obj, ['Z']), "Z.s")
     verifyEq(t.method("s").callOn(obj, ['X']), "X.s")
  }

  Void testNamedSuperErrors()
  {
    verifyErrors("
      mixin X
      {
        virtual Str f() { return \"X.f\" }
      }

      mixin Y : X
      {
        //Str m00() { return super.f }  Unknown slot before CheckErrors
        Str m01() { return super.toStr }
        Str m02() { return Obj.super.toStr }
      }

      class A
      {
        virtual Str a() { return \"A.f\" }
      }

      class Foo
      {
        virtual Str f() { return \"Foo.f\" }
      }

      class B : A, Y
      {
        Str m00() { return Foo.super.f }
      }

     ", [
       10, 28, "Must use named 'super' inside mixin",
       11, 28, "Cannot use 'Obj.super' inside mixin (yeah I know - take it up with Sun)",
       26, 28, "Named super '$podName::Foo' not a super class of 'B'",
     ])
  }

}