//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   02 Aug 06  Andy Frank  Creation
//

using web

**
** FandocUri handles routing and creating URIs for Fandoc.
**
mixin FandocUri
{

//////////////////////////////////////////////////////////////////////////
// Accessors
//////////////////////////////////////////////////////////////////////////

  **
  ** The WebReq instance for this request.
  **
  abstract WebReq _req()

  **
  ** The WebRes instance for this request.
  **
  abstract WebRes _res()

//////////////////////////////////////////////////////////////////////////
// Routing
//////////////////////////////////////////////////////////////////////////

  **
  ** Route this request to the appropriate Weblet. If this
  ** request is not valid, return a 404 error response.
  **
  Void route()
  {
    path := _req.suffixUri.path

    // Index
    if (path.size == 0)
    {
      FandocIndex.make.get
      return
    }

    // Search
    if (path.size == 1 && path[0] == "search")
    {
      FandocSearch.make.get
      return
    }

    // Resources
    if (path.size == 2 && path[0] == "res")
    {
      serveRes(_req, _res)
      return
    }

    // Verify pod
    pod := Pod.find(path[0], false)
    if (pod == null) { _res.sendError(404); return }
    _req.stash["pod"] = pod

    // Pod
    if (path.size == 1)
    {
      FandocPod.make.get
      return
    }

    // Verify type
    type := pod.findType(path[1], false)
    if (type == null) { _res.sendError(404); return }
    _req.stash["type"] = type

    // Type
    if (path.size == 2)
    {
      FandocType.make.get
      return
    }

    // Src
    if (path.size == 3 && path[2] == "src")
    {
      FandocSrc.make.get
      return
    }

    _res.sendError(404)
  }

  **
  ** Serve the requested resource out of the fandoc pod.
  **
  Void serveRes(WebReq _req, WebRes _res)
  {
    try
    {
      // Try to set content type
      if (_req.suffixUri.ext.lower == "css")
        _res.headers["Content-Type"] = "text/css"
      else if (_req.suffixUri.ext.lower == "js")
        _res.headers["Content-Type"] = "text/javascript"

      // Pipe resource to outstream
      Pod.find("fandoc").files["/res/${_req.suffixUri.path[1]}".toUri].in.pipe(_res.out)
    }
    catch
    {
      // Assume if anything failed, its because the file
      // did not exist - TODO could probably be better
      _res.sendError(404)
    }
  }

//////////////////////////////////////////////////////////////////////////
// Hrefs
//////////////////////////////////////////////////////////////////////////

  **
  ** Return the Uri to the index page.
  **
  Uri indexUri()
  {
    return _req.prefixUri // + "/".toUri
  }

  **
  ** Return the Uri to this pod.
  **
  Uri podUri(Pod p)
  {
    return _req.prefixUri + "$p.name".toUri
  }

  **
  ** Return the Uri to this type.
  **
  Uri typeUri(Type t)
  {
    return _req.prefixUri + "$t.pod.name/$t.name".toUri
  }

  **
  ** Return the Uri to the source for this type.
  **
  Uri typeSrcUri(Type t)
  {
    return _req.prefixUri + "$t.pod.name/$t.name/src".toUri
  }

}