//
// Copyright (c) 2006, John Sublett
// Licensed under the Academic Free License version 3.0
//
// History:
//   31 July 2006  John Sublett  Creation
//

**
** TokenizerTest
**

class TokenizerTest : FantextTest
{
  **
  ** Test all token types in the TokenType enum.
  ** If a mismatch occurs when the input token
  ** type is "spaces" the input token is skipped.
  **
  Void testTypes()
  {
    types := TokenType.values
    in := InStream.makeForStr("*&`#-. \tabcde 1234<>()[]\"\n~")
    tokens := Tokenizer.make(in)

    i := 0
    Token t
    while ((t = tokens.next).ttype != TokenType.eof)
    {
      verify((t.ttype === types[i]) || (t.ttype === TokenType.spaces),
             "Expecting token type " + types[i] + " not " + t.ttype)
      if (t.ttype != types[i]) continue; // spaces
      i++
    }
  }

  **
  ** Test that the "spaces" tokens have the right length.
  **
  Void testSpaces()
  {
    for (i := 1; i <= 10; i++)
    {
      s := Str.spaces(i)
      in := InStream.makeForStr(s)
      tokens := Tokenizer.make(in)

      t := tokens.next
      verify((t.ttype === TokenType.spaces) && (t.lex.size === i))
      verify(tokens.next.ttype === TokenType.eof)
    }
  }
}