//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   27 Mar 06  Brian Frank  Creation
//
package fan.sys;

import java.io.*;
import sun.nio.cs.StreamEncoder;

/**
 * SysOutStream
 */
public class SysOutStream
  extends FanObj
  implements OutStream
{

//////////////////////////////////////////////////////////////////////////
// Constructor
//////////////////////////////////////////////////////////////////////////

  public SysOutStream(OutputStream out)
  {
    this.out = out;
  }

//////////////////////////////////////////////////////////////////////////
// Obj
//////////////////////////////////////////////////////////////////////////

  public Type type() { return Sys.SysOutStreamType; }

//////////////////////////////////////////////////////////////////////////
// OutStream
//////////////////////////////////////////////////////////////////////////

  public final OutStream write(Int b) { return w((int)b.val); }
  public OutStream w(int v)
  {
    try
    {
      out.write(v);
      return this;
    }
    catch (IOException e)
    {
      throw IOErr.make(e).val;
    }
  }

  public final OutStream writeBuf(Buf buf) { return writeBuf(buf, buf.remaining()); }
  public OutStream writeBuf(Buf buf, Int n)
  {
    try
    {
      int len = (int)n.val;
      if (buf.pos + len > buf.size) throw IOErr.make("Not enough bytes to write").val;
      out.write(buf.buf, buf.pos, len);
      buf.pos += len;
      return this;
    }
    catch (IOException e)
    {
      throw IOErr.make(e).val;
    }
  }

  public OutStream writeI2(Int x) { return OutStream$.writeI2(this, x); }

  public OutStream writeI4(Int x) { return OutStream$.writeI4(this, x); }

  public OutStream writeI8(Int x) { return OutStream$.writeI8(this, x); }

  public OutStream writeF4(Float x) { return OutStream$.writeF4(this, x); }

  public OutStream writeF8(Float x) { return OutStream$.writeF8(this, x); }

  public OutStream writeBool(Bool x) { return OutStream$.writeBool(this, x); }

  public OutStream writeUtf(Str x) { return OutStream$.writeUtf(this, x); }

  public OutStream writeChar(Int c)
  {
    if (charsetEncoder == null) charsetEncoder = charset.newEncoder();
    charsetEncoder.encode((char)c.val, this);
    return this;
  }

  public final OutStream writeChars(Str s) { return writeChars(s.val, 0, s.val.length()); }
  public final OutStream writeChars(Str s, Int off) { return writeChars(s.val, (int)off.val, s.val.length()-(int)off.val); }
  public final OutStream writeChars(Str s, Int off, Int len) { return writeChars(s.val, (int)off.val, (int)len.val); }
  public OutStream writeChars(String s, int off, int len)
  {
    if (charsetEncoder == null) charsetEncoder = charset.newEncoder();
    int end = off+len;
    for (int i=off; i<end; ++i)
      charsetEncoder.encode(s.charAt(i), this);
    return this;
  }

  public OutStream print(Obj obj) { return OutStream$.print(this, obj); }

  public OutStream printLine() { return OutStream$.printLine(this, Str.Empty); }
  public OutStream printLine(Obj obj) { return OutStream$.printLine(this, obj); }

  public OutStream writeProps(Map props) { return OutStream$.writeProps(this, props, Bool.True); }
  public OutStream writeProps(Map props, Bool close) { return OutStream$.writeProps(this, props, close); }

  public void flush()
  {
    try
    {
      out.flush();
    }
    catch (IOException e)
    {
      throw IOErr.make(e).val;
    }
  }

  public Bool close()
  {
    try
    {
      out.close();
      return Bool.True;
    }
    catch (IOException e)
    {
      return Bool.False;
    }
  }

  public Charset charset()
  {
    return charset;
  }

  public void charset(Charset charset)
  {
    this.charset = charset;
    this.charsetEncoder = null;
  }

//////////////////////////////////////////////////////////////////////////
// Java Conversion
//////////////////////////////////////////////////////////////////////////

  /**
   * Get a java.io.OutputStream for the specified output stream.
   */
  public static OutputStream java(OutStream out)
  {
    if (out instanceof SysOutStream)
      return ((SysOutStream)out).out;
    else
      return new JavaOutputStream(out);
  }

  static class JavaOutputStream extends OutputStream
  {
    JavaOutputStream(OutStream out) { this.out = out; }
    public void write(int b) { out.write(Int.make(b)); }
    public void write(byte[] b, int off, int len)
    {
      buf.buf = b;
      buf.pos = off;
      buf.size = b.length;
      out.writeBuf(buf, Int.make(len));
    }
    public void close() { out.close(); }
    OutStream out;
    Buf buf = new Buf();
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  OutputStream out;
  Charset charset = Charset.utf8();
  Charset.Encoder charsetEncoder;

}