//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   3 Jul 06  Brian Frank  Creation
//
package fan.sys;

import java.text.*;
import java.util.*;

/**
 * Time represents an absolute instance in time.  Fan time is normalized
 * as nanosecond ticks since 1 Jan 2000  GMT.
 */
public final class Time
  extends FanObj
{

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  public static Time now ()  { return new Time(System.currentTimeMillis()*1000000L); }

  public static Time make(Int ticks) { return make(ticks.val); }
  public static Time make(long ticks)
  {
    if (ticks == 0) return Zero;
    return new Time(ticks);
  }

  private Time(long ticks)
  {
    this.ticks = ticks;
  }

//////////////////////////////////////////////////////////////////////////
// Identity
//////////////////////////////////////////////////////////////////////////

  public final Bool equals(Obj obj)
  {
    if (obj instanceof Time)
    {
      return ticks == ((Time)obj).ticks ? Bool.True : Bool.False;
    }
    return Bool.False;
  }

  public Int compare(Obj obj)
  {
    long that = ((Time)obj).ticks;
    if (ticks < that) return Int.LT; return ticks  == that ? Int.EQ : Int.GT;
  }

  public int hashCode()
  {
    return (int)(ticks ^ (ticks >>> 32));
  }

  public Int hash()
  {
    return Int.make(ticks);
  }

  public Type type()
  {
    return Sys.TimeType;
  }

//////////////////////////////////////////////////////////////////////////
// Access
//////////////////////////////////////////////////////////////////////////

  public final Int ticks()
  {
    return Int.make(ticks);
  }

//////////////////////////////////////////////////////////////////////////
// Operators
//////////////////////////////////////////////////////////////////////////

  public Duration minus(Time time)
  {
    return Duration.make(ticks-time.ticks);
  }

  public Time plus(Duration duration)
  {
    long d = duration.ticks;
    if (d == 0) return this;
    return make(ticks+d);
  }

//////////////////////////////////////////////////////////////////////////
// Conversion
//////////////////////////////////////////////////////////////////////////

  public final Str toStr()
  {
    if (str == null)
      str = Str.make(format.format(java()));
    return str;
  }

//////////////////////////////////////////////////////////////////////////
// Java
//////////////////////////////////////////////////////////////////////////

  public static Time java(long millis) { if (millis <= 0) return null; return new Time(millis*1000000L); }
  public long java() { return ticks / 1000000L; }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  private static final DateFormat format = new SimpleDateFormat("d-MMM-yyyy HH:mm:ss zzz");
  private static final Time Zero = new Time(0);

  private final long ticks;
  private Str str;

}