//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   15 Sep 05  Brian Frank  Creation
//
package fanx.compiler;

import java.io.*;
import fanx.util.*;

/**
 * Log manages logging compiler messages.  The default
 * writes everything to standard output.
 */
public class Log
  extends PrintWriter
{

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  public Log(OutputStream out)
  {
    super(out);
  }

  public Log(Writer out)
  {
    super(out);
  }

  public Log()
  {
    super(System.out);
  }

//////////////////////////////////////////////////////////////////////////
// Methods
//////////////////////////////////////////////////////////////////////////

  public void indent()
  {
    indent++;
  }

  public void unindent()
  {
    indent--;
    if (indent < 0) indent = 0;
  }

  public void message(String msg, Object file)
  {
    if (silent) return;
    int n = indent*2;
    print(StrUtil.getSpaces(n));
    print(msg); n += msg.length();
    if (file != null)
    {
      print(' '); n++;
      if (n < 25) print(StrUtil.getSpaces(25-n));
      print("[" + file + "]");
    }
    println();
    flush();
  }

  public void message(String msg)
  {
    message(msg, null);
  }

  public void verbose(String msg, Object file)
  {
    if (verbose) message(msg, file);
  }

  public void verbose(String msg)
  {
    if (verbose) message(msg, null);
  }

  public void warning(String msg, Object file)
  {
    int oldIndent = indent;
    indent = 0;
    message("WARNING: " + msg, file);
    indent = oldIndent;
  }

  public void error(String msg, Object file, Throwable ex)
  {
    indent = 0;
    message("ERROR: " + msg, file);

    // dump CompileException only while verbose, all others always
    if (ex != null)
    {
      if (verbose || !(ex instanceof CompileException))
      {
        dump(ex);
      }
    }
  }

  public void error(CompileException err)
  {
    if (silent && !verbose) return;
    println(err.getMessage());
    flush();
    if (verbose) dump(err);
  }

  public void dump(Throwable ex)
  {
    if (ex == null) return;
    ex.printStackTrace();
    if (ex instanceof CompileException) dump(((CompileException)ex).cause);
    flush();
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  public boolean verbose;
  public boolean silent;
  public int indent;

}