//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   25 Dec 05  Brian Frank  Creation
//
package fanx.compiler;

import fan.sys.*;
import fanx.compiler.ast.*;

/**
 * ShimType is a place holder Type for a type we are currently compiling,
 * but haven't yet committed to the Sys namespace.  By subclassing Type
 * it may be used in place of type checks against existing libraries for
 * check within the compilation unit.
 */
public class ShimType
  extends Type
{

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  public ShimType(ShimPod pod, ClassDef def)
  {
    super(pod, def.name, def.flags, null);
    this.def = def;
  }

//////////////////////////////////////////////////////////////////////////
// Overrides
//////////////////////////////////////////////////////////////////////////

  public Type base()
  {
    if (base == null && !isMixin().val) throw new IllegalStateException();
    return base;
  }

  public List mixins()
  {
    if (mixins == null) throw new IllegalStateException();
    return new List(Sys.TypeType, mixins);
  }

  public void restubSlots()
  {
    Slot[] slots = new Slot[def.slots.size()];
    for (int i=0; i<slots.length; ++i)
      slots[i] = ((SlotDef)def.slots.get(i)).slot();
    stubSlots(slots);
  }

  public void stubSlots(Slot[] slots)
  {
    init = true;
    super.stubSlots(slots);
  }

  public Str toStr()
  {
    // if base is a MethodType, then this is a closure,
    // always return our method type signature in error messages
    if (base instanceof MethodType) return base.toStr();

    return super.toStr();
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  public ClassDef def;
  public Type base = Sys.ObjType;
  public Type[] mixins = new Type[0];
  public boolean init;

}