//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   15 Sep 05  Brian Frank  Creation
//
package fanx.compiler.ast;

import java.util.*;
import fan.sys.*;
import fanx.compiler.*;

/**
 * Block is a list of Statements
 */
public class Block
  extends Node
{

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  public Block(ILocation location)
  {
    super(location);
    statements = new ArrayList();
  }

//////////////////////////////////////////////////////////////////////////
// Access
//////////////////////////////////////////////////////////////////////////

  public boolean isEmpty()
  {
    return statements.isEmpty();
  }

  public Stmt first()
  {
    if (statements.size() == 0) return null;
    return (Stmt)statements.get(0);
  }

  public Stmt last()
  {
    if (statements.size() == 0) return null;
    return (Stmt)statements.get(statements.size()-1);
  }

  public boolean lastIsMethodExit()
  {
    Stmt last = last();
    if (last != null) return last.isMethodExit();
    return false;
  }

  public void add(Stmt stmt)
  {
    statements.add(stmt);
  }

  public void add(Expr expr)
  {
    statements.add(new ExprStmt(expr.location, expr));
  }

  public void add(int index, Block b)
  {
    for (int i=0; i<b.statements.size(); ++i)
      statements.add(index+i, b.statements.get(i));
  }

  public void add(int index, Stmt stmt)
  {
    statements.add(index, stmt);
  }

  public void add(int index, Expr expr)
  {
    statements.add(index, new ExprStmt(expr.location, expr));
  }

  public void add(Block b)
  {
    add(statements.size(), b);
  }

//////////////////////////////////////////////////////////////////////////
// Tree
//////////////////////////////////////////////////////////////////////////

  public Node walk(Visitor v)
  {
    v.enterBlock(this);
    statements = walk(v, statements);
    v.exitBlock(this);
    return v.visit(this);
  }

//////////////////////////////////////////////////////////////////////////
// Debug
//////////////////////////////////////////////////////////////////////////

  public String toString()
  {
    return "Block " + Location.format(this);
  }

  public void tree(AstWriter out)
  {
    out.w("Block").nl();
    out.indent();
    for (int i=0; i<statements.size(); ++i)
      ((Stmt)statements.get(i)).tree(out);
    out.unindent();
  }

  public void pretty(AstWriter out) { pretty(out, true); }
  public void pretty(AstWriter out, boolean braces)
  {
    if (braces) out.w("{").nl();
    if (braces) out.indent();
    for (int i=0; i<statements.size(); ++i)
      ((Stmt)statements.get(i)).pretty(out);
    if (braces) out.unindent();
    if (braces) out.w("}").nl();
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  public ArrayList statements;   // Stmt[]

}