//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   15 Sep 05  Brian Frank  Creation
//
package fanx.compiler.ast;

import java.util.*;
import fan.sys.*;
import fanx.compiler.*;

/**
 * CastExpr converts the target expression into the toType.  This AST node
 * covers the abstract notion of conversion generically - although at the
 * assembly level conversion may be: casts, auto-boxing/unboxing, and type
 * wrapping/unwrapping.
 */
public class CastExpr
  extends Expr
{

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  public CastExpr(TypeRef toType, Expr target)
  {
    super(toType.location);
    this.toType = toType;
    this.target = target;
  }

  public CastExpr(Type toType, Expr target)
  {
    super(target.location);
    this.type   = toType;
    this.toType = new TypeRef(location, toType);
    this.target = target;
  }

  public CastExpr(ILocation location)
  {
    super(location);
  }

//////////////////////////////////////////////////////////////////////////
// Expr
//////////////////////////////////////////////////////////////////////////

  public int id()
  {
    return CAST;
  }

//////////////////////////////////////////////////////////////////////////
// Tree
//////////////////////////////////////////////////////////////////////////

  protected void walkChildren(Visitor v)
  {
    toType = walk(v, toType);
    target = walk(v, target);
  }

//////////////////////////////////////////////////////////////////////////
// Debug
//////////////////////////////////////////////////////////////////////////

  public String toString()
  {
    return " ((" + toType + ")" + target + ")";
  }

  public void tree(AstWriter out)
  {
    out.w("CastExpr").nl();
    out.indent();
    out.w("toType: "); toType.tree(out);
    out.w("target: "); target.tree(out);
    out.unindent();
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  public TypeRef toType;
  public Expr target;

}