//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   15 Sep 05  Brian Frank  Creation
//
package fanx.compiler.ast;

import java.util.*;
import fanx.compiler.*;

/**
 * CompilationUnit models the top level compilation unit of a source file.
 */
public class CompilationUnit
  extends Node
{

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  public CompilationUnit(ILocation location, String podName)
  {
    super(location);
    this.podName = podName;
    this.classes = new ArrayList();
    this.imports = new ArrayList();
  }

//////////////////////////////////////////////////////////////////////////
// Methods
//////////////////////////////////////////////////////////////////////////

  public ClassDef topClass()
  {
    // first class is always the main top level class
    return (ClassDef)classes.get(0);
  }

//////////////////////////////////////////////////////////////////////////
// Tree
//////////////////////////////////////////////////////////////////////////

  public Node walk(Visitor v)
  {
    imports = walk(v, imports);
    classes = walk(v, classes);
    return v.visit(this);
  }

//////////////////////////////////////////////////////////////////////////
// Debug
//////////////////////////////////////////////////////////////////////////

  public void tree(AstWriter out)
  {
    out.nl();
    out.w("CompilationUnit").nl();
    if (imports.size() > 0)
    {
      out.nl();
      for (int i=0; i<imports.size(); ++i)
        ((Import)imports.get(i)).tree(out);
    }
    for (int i=0; i<classes.size(); ++i)
    {
      ((ClassDef)classes.get(i)).tree(out);
    }
  }

  public void pretty(AstWriter out)
  {
    out.nl();
    if (imports.size() > 0)
    {
      out.nl();
      for (int i=0; i<imports.size(); ++i)
        ((Import)imports.get(i)).pretty(out);
    }
    for (int i=0; i<classes.size(); ++i)
    {
      ((ClassDef)classes.get(i)).pretty(out);
    }
  }

  public String toString()
  {
    return Location.format(location);
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  public String podName;
  public ArrayList imports;  // Import[]
  public ArrayList classes;  // ClassRef[]

}