//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   10 Nov 05  Brian Frank  Creation
//
package fanx.compiler.ast;

import fan.sys.*;
import fanx.fcode.*;
import fanx.compiler.*;

/**
 * DefNode is the abstract base class for definition nodes such as ClassDef,
 * MethodDef, and FieldDef.  All definitions may be documented using a Java
 * style JavaDoc comment.
 */
public abstract class DefNode
  extends Node
{

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  public DefNode(ILocation location)
  {
    super(location);
  }

//////////////////////////////////////////////////////////////////////////
// Flags
//////////////////////////////////////////////////////////////////////////

  public boolean isAbstract()  { return (flags & FConst.Abstract) != 0;  }
  public boolean isOverride()  { return (flags & FConst.Override) != 0;  }
  public boolean isPrivate()   { return (flags & FConst.Private) != 0;   }
  public boolean isProtected() { return (flags & FConst.Protected) != 0; }
  public boolean isPublic()    { return (flags & FConst.Public) != 0;    }
  public boolean isStatic()    { return (flags & FConst.Static) != 0;    }
  public boolean isConst()     { return (flags & FConst.Const) != 0;    }
  public boolean isSynthetic() { return (flags & FConst.Synthetic) != 0;    }
  public boolean isVirtual()   { return (flags & FConst.Virtual) != 0;   }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  public String doc;        // documentation comment or null
  public int flags;         // class/slot flags

}