//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   15 Sep 05  Brian Frank  Creation
//
package fanx.compiler.ast;

import java.util.*;
import fanx.compiler.*;

/**
 * ListLiteralExpr supports creating a List populated by zero or more expressions.
 */
public class ListLiteralExpr
  extends Expr
{

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  public ListLiteralExpr(ILocation location, TypeRef explicitType)
  {
    super(location);
    this.explicitType = explicitType;
  }

//////////////////////////////////////////////////////////////////////////
// Expr
//////////////////////////////////////////////////////////////////////////

  public int id()
  {
    return LIST_LITERAL;
  }

//////////////////////////////////////////////////////////////////////////
// Tree
//////////////////////////////////////////////////////////////////////////

  protected void walkChildren(Visitor v)
  {
    explicitType = walk(v, explicitType);
    items = walk(v, items);
  }

//////////////////////////////////////////////////////////////////////////
// Debug
//////////////////////////////////////////////////////////////////////////

  public void tree(AstWriter out)
  {
    out.w("ListLiteralExpr ").w(this).w(" -> ").w(type).nl();
    out.indent();
    if (explicitType != null) { out.w("explicitType: "); explicitType.tree(out); }
    for (int i=0; i<items.size(); ++i)
    {
      out.w("items[" + i + "] ");
      ((Expr)items.get(i)).tree(out);
    }
    out.unindent();
  }

  public String toString()
  {
    StringBuilder s = new StringBuilder();
    if (explicitType != null) s.append(explicitType);
    s.append("[");
    for (int i=0; i<items.size(); ++i)
    {
      if (i > 0) s.append(", ");
      s.append(items.get(i));
    }
    s.append("]");
    return s.toString();
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  public TypeRef explicitType;               // for example: ExplicitType[...]
  public ArrayList items = new ArrayList();  // Expr[]

}