//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   15 Sep 05  Brian Frank  Creation
//
package fanx.compiler.ast;

import fan.sys.*;
import fanx.compiler.*;

/**
 * ListTypeRef is used to represent a parameterized List type.
 */
public class ListTypeRef
  extends TypeRef
{

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  public ListTypeRef(ILocation location, TypeRef of)
  {
    super(location, "sys::List");
    this.of = of;
  }

  public ListTypeRef(ILocation location, Type of)
  {
    super(location, "sys::List");
    this.of = new TypeRef(location, of);
  }

//////////////////////////////////////////////////////////////////////////
// Tree
//////////////////////////////////////////////////////////////////////////

  public Node walk(Visitor v)
  {
    of = walk(v, of);
    return v.visit(this);
  }

//////////////////////////////////////////////////////////////////////////
// Debug
//////////////////////////////////////////////////////////////////////////

  public String toString()
  {
    if (type != null) return type.toString();
    return of + "[]?";
  }

  public void tree(AstWriter out)
  {
    out.w("ListTypeRef " + toString()).nl();
  }

  public void pretty(AstWriter out)
  {
    out.w(toString());
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  public TypeRef of;

}