//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   15 Sep 05  Brian Frank  Creation
//
package fanx.compiler.ast;

import java.util.*;
import fan.sys.*;
import fanx.compiler.*;

/**
 * MethodTypeRef is used to represent a parameterized Method type.
 */
public class MethodTypeRef
  extends TypeRef
{

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  public MethodTypeRef(ILocation location)
  {
    super(location, "sys::Method");
  }

//////////////////////////////////////////////////////////////////////////
// MethodTypeRef
//////////////////////////////////////////////////////////////////////////

  public int paramCount()
  {
    return paramTypes.size();
  }

  public TypeRef paramType(int i)
  {
    return (TypeRef)paramTypes.get(i);
  }

  public MethodType toMethodType()
  {
    Type[] params = new Type[paramTypes.size()];
    for (int i=0; i<params.length; ++i)
    {
      params[i] = ((TypeRef)paramTypes.get(i)).type;
      if (params[i] == null) throw new IllegalStateException();
    }
    if (ret.type == null) throw new IllegalStateException();
    return new MethodType(params, ret.type);
  }

//////////////////////////////////////////////////////////////////////////
// Tree
//////////////////////////////////////////////////////////////////////////

  public Node walk(Visitor v)
  {
    paramTypes = walk(v, paramTypes);
    ret = walk(v, ret);
    return v.visit(this);
  }

//////////////////////////////////////////////////////////////////////////
// Debug
//////////////////////////////////////////////////////////////////////////

  public String toString()
  {
    if (type != null) return type.toString();
    StringBuilder s = new StringBuilder();
    s.append('|');
    for (int i=0; i<paramTypes.size(); ++i)
    {
      if (i > 0) s.append(", ");
      s.append(paramTypes.get(i) + " " + paramNames.get(i));
    }
    s.append(" -> ").append(ret);
    s.append('|');
    return s.toString();
  }

  public void tree(AstWriter out)
  {
    out.w("MethodTypeRef " + toString()).nl();
    out.indent();
    out.w("ret: "); ret.tree(out);
    out.w("params:").nl();
    for (int i=0; i<paramNames.size(); ++i)
    {
      out.w("  " + paramNames.get(i) + " ");
      ((TypeRef)paramTypes.get(i)).tree(out);
    }
    out.unindent();
  }

  public void pretty(AstWriter out)
  {
    out.w(toString());
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  public TypeRef ret;
  public ArrayList paramNames = new ArrayList();
  public ArrayList paramTypes = new ArrayList();

}