//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   15 Sep 05  Brian Frank  Creation
//
package fanx.compiler.ast;

import fanx.compiler.*;

/**
 * Stmt models a statement production.
 */
public abstract class Stmt
  extends Node
{

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  public Stmt(ILocation location)
  {
    super(location);
  }

//////////////////////////////////////////////////////////////////////////
// Overrides
//////////////////////////////////////////////////////////////////////////

  /**
   * Get a statement type id constant (defined below).
   */
  public abstract int id();

  /**
   * Return if statement when executed is guaranteed to exit
   * the method either with a return or throw.
   */
  public boolean isMethodExit()
  {
    return false;
  }

//////////////////////////////////////////////////////////////////////////
// Tree
//////////////////////////////////////////////////////////////////////////

  public final Node walk(Visitor v)
  {
    v.enterStmt(this);
    walkChildren(v);
    v.exitStmt(this);
    return v.visit(this);
  }

  protected void walkChildren(Visitor v)
  {
  }

  public void pretty(AstWriter out)
  {
    pretty(out, true);
  }

  public void pretty(AstWriter out, boolean endOfStmt)
  {
    pretty(out);
  }

//////////////////////////////////////////////////////////////////////////
// Id Constants
//////////////////////////////////////////////////////////////////////////

  public static final int NOP           = 0;
  public static final int EXPR          = 1;
  public static final int LOCAL_VAR_DEF = 2;
  public static final int IF            = 3;
  public static final int RETURN        = 4;
  public static final int THROW         = 5;
  public static final int FOR           = 6;
  public static final int WHILE         = 7;
  public static final int BREAK         = 8;
  public static final int CONTINUE      = 9;
  public static final int SWITCH        = 10;
  public static final int TRY           = 11;

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

}