//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   27 Jun 06  Brian Frank  Creation
//
package fanx.compiler.ast;

import java.util.*;
import fanx.compiler.*;

/**
 * Models the catch block of a try statement.
 */
public class TryCatch
  extends Node
{

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  public TryCatch(ILocation location)
  {
    super(location);
  }

//////////////////////////////////////////////////////////////////////////
// Tree
//////////////////////////////////////////////////////////////////////////

  public Node walk(Visitor v)
  {
    block = walk(v, block);
    errType = walk(v, errType);
    return v.visit(this);
  }

//////////////////////////////////////////////////////////////////////////
// Debug
//////////////////////////////////////////////////////////////////////////

  public String toString()
  {
    if (errVariable == null) return "catch";
    return "catch (" + errType + " " + errVariable + ")";
  }

  public void tree(AstWriter out)
  {
    out.w("TryCatch " + this).nl();
    out.indent();
    out.w("block: "); block.tree(out);
    out.unindent();
  }

  public void pretty(AstWriter out)
  {
    out.w(this).nl();
    block.pretty(out);
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  public TypeRef errType;      // Err type to catch in this block
  public String errVariable;   // name of err local variable
  public Block block;          // body of catch block

}