//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   15 Sep 05  Brian Frank  Creation
//
package fanx.compiler.ast;

import fan.sys.*;
import fanx.compiler.*;

/**
 * TypeRef indicates a type identifier such as the type of a field.
 * It is resolved into a Type during the resolve compilation phase.
 */
public class TypeRef
  extends Node
{

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  public TypeRef(ILocation location, String id)
  {
    super(location);
    this.id = id;
  }

  public TypeRef(ILocation location, Type type)
  {
    super(location);
    this.id   = type.name().val;
    this.type = type;
  }

//////////////////////////////////////////////////////////////////////////
// TypeRef
//////////////////////////////////////////////////////////////////////////

  public boolean isVoid()
  {
    if (type != null) return type.isVoid();
    return id.equals("Void");
  }

  public TypeRef toListOf()
  {
    return new ListTypeRef(location, this);
  }

//////////////////////////////////////////////////////////////////////////
// Tree
//////////////////////////////////////////////////////////////////////////

  public Node walk(Visitor v)
  {
    return v.visit(this);
  }

//////////////////////////////////////////////////////////////////////////
// Debug
//////////////////////////////////////////////////////////////////////////

  public String toString()
  {
    if (type != null) return type.toString();
    return id + "?";
  }

  public void tree(AstWriter out)
  {
    out.w("TypeRef " + toString()).nl();
  }

  public void pretty(AstWriter out)
  {
    out.w(toString());
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  public String id;        // unbound identifier
  public Type type;        // resolved Type or null if unresolved

}