//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   15 Sep 05  Brian Frank  Creation
//
package fanx.compiler.ast;

import java.util.*;
import fanx.compiler.*;
import fanx.compiler.parser.*;

/**
 * UnaryExpr is used for unary expressions including !, +.
 * Note that - is mapped to negate() as a shortcut method.
 */
public class UnaryExpr
  extends Expr
{

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  public UnaryExpr(ILocation location, int opToken, Expr operand)
  {
    super(location);
    this.opToken = opToken;
    switch (opToken)
    {
      case Token.BANG:  id = BOOL_NOT; break;
      case Token.PLUS:  id = POSITIVE; break;
      default: throw new IllegalStateException(Token.toString(opToken));
    }
    this.operand = operand;
  }

  public UnaryExpr(ILocation location, int id, int opToken, Expr operand)
  {
    super(location);
    this.id = id;
    this.opToken = opToken;
    this.operand = operand;
  }

  public UnaryExpr(ILocation location, int opToken)
  {
    this(location, opToken, null);
  }

//////////////////////////////////////////////////////////////////////////
// Expr
//////////////////////////////////////////////////////////////////////////

  public int id()
  {
    return id;
  }

//////////////////////////////////////////////////////////////////////////
// Tree
//////////////////////////////////////////////////////////////////////////

  protected void walkChildren(Visitor v)
  {
    operand = walk(v, operand);
  }

//////////////////////////////////////////////////////////////////////////
// Debug
//////////////////////////////////////////////////////////////////////////

  public String toString()
  {
    if (id == POST_INCREMENT || id == POST_DECREMENT)
      return operand + Token.toString(opToken);
    else
      return Token.toString(opToken) + operand;
  }

  public void tree(AstWriter out)
  {
    out.w("UnaryExpr ").w(Token.toString(opToken) + " " + leaveToString()).nl();
    out.indent();
    out.w("operand: "); operand.tree(out);
    out.unindent();
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  public int opToken;    // operator token type (Token.BANG, etc)
  public int id;         // NOT, etc
  public Expr operand;   // operand expression

}