//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   3 Sep 05  Brian Frank  Creation
//
package fanx.compiler.steps;

import java.io.*;
import java.util.*;
import fanx.util.*;
import fanx.compiler.*;
import fanx.compiler.Compiler;
import fanx.compiler.ast.*;
import fanx.compiler.parser.*;

/**
 * Parse all the source files in into memory as abstract syntax tree
 * compilation units containing one or more ClassDefs:
 *   Compiler.srcFiles -> Compiler.units & Compiler.classDefs
 */
public class ParseFiles
  extends CompilerStep
{

  public ParseFiles(Compiler compiler)
  {
    super(compiler);
  }

  public void run()
  {
    log("Compile", compiler.srcFiles.length + " files");

    // compile each source file into a compilation unit
    compiler.units = new CompilationUnit[compiler.srcFiles.length];
    for (int i=0; i<compiler.srcFiles.length; ++i)
      compiler.units[i] = parse(compiler.srcFiles[i]);

    // flatten the classes into the classDefs field
    ArrayList acc = new ArrayList();
    for (int i=0; i<compiler.units.length; ++i)
      for (int j=0; j<compiler.units[i].classes.size(); ++j)
        acc.add(compiler.units[i].classes.get(j));
    compiler.classDefs = (ClassDef[])acc.toArray(new ClassDef[acc.size()]);
    bombIfErr();
  }

  CompilationUnit parse(File file)
  {
    verbose("Parse", file);

    // read source file into memory
    char[] buf = null;
    try
    {
      buf = FileUtil.read(file);
    }
    catch (IOException e)
    {
      throw err("Cannot read source file", file, e);
    }

    // parse into compilation unit
    CompilationUnit unit = new Parser(compiler, buf, file.toString()).parse();

    return unit;
  }

}