//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   26 Dec 05  Brian Frank  Creation
//
package fanx.fcode;

import java.io.*;
import java.util.*;

/**
 * FSlot is the read/write fcode representation of sys::Slot.
 */
public class FSlot
  implements FConst
{

  public FFacet facet(FPod pod, String name)
  {
    return FFacet.facet(pod, facets, name);
  }

  protected void writeCommon(FStore.Output out) throws IOException
  {
    out.u2(name);
    out.u4(flags & FlagsMask);
  }

  protected void readCommon(FStore.Input in) throws IOException
  {
    name  = in.u2();
    flags = in.u4();
  }

  protected void writeFacets(FStore.Output out) throws IOException
  {
    if (facets == null)
    {
      out.u2(0);
    }
    else
    {
      out.u2(facets.length);
      for (int i=0; i<facets.length; ++i)
        facets[i].write(out);
    }
  }

  protected void readFacets(FStore.Input in) throws IOException
  {
    int len = in.u2();
    if (len == 0)
    {
      facets = FFacet.None;
    }
    else
    {
      facets = new FFacet[len];
      for (int i=0; i<len; ++i)
        facets[i] = new FFacet().read(in);
    }
  }

  public int name;          // str index
  public int flags;         // bitmask
  public FFacet[] facets;   // meta-data

}