//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   26 Dec 05  Brian Frank  Creation
//
package fanx.fcode;

import java.io.*;
import java.util.*;

/**
 * FType is the read/write fcode representation of sys::Type.
 */
public class FType
  implements FConst
{

//////////////////////////////////////////////////////////////////////////
// Constructor
//////////////////////////////////////////////////////////////////////////

  public FType(FPod pod)
  {
    this.pod = pod;
    this.facets = FFacet.None;
  }

//////////////////////////////////////////////////////////////////////////
// Lookup
//////////////////////////////////////////////////////////////////////////

  public FSlot slot(String name)
  {
    for (int i=0; i<fields.length; ++i)
      if (pod.name(fields[i].name).equals(name))
        return fields[i];
    for (int i=0; i<methods.length; ++i)
      if (pod.name(methods[i].name).equals(name))
        return methods[i];
    throw fan.sys.UnknownSlotErr.make(name).val;
  }

  public FFacet facet(String name)
  {
    return FFacet.facet(pod, facets, name);
  }

//////////////////////////////////////////////////////////////////////////
// Meta IO
//////////////////////////////////////////////////////////////////////////

  public void writeMeta(FStore.Output out) throws IOException
  {
    out.u2(self);
    out.u2(base);
    out.u2(mixins.length);
    for (int i=0; i<mixins.length; ++i) out.u2(mixins[i]);
    out.u4(flags & FlagsMask);
  }

  public FType readMeta(FStore.Input in) throws IOException
  {
    self   = in.u2();
    base   = in.u2();
    mixins = new int[in.u2()];
    for (int i=0; i<mixins.length; ++i) mixins[i] = in.u2();
    flags  = in.u4();
    return this;
  }

//////////////////////////////////////////////////////////////////////////
// Body IO
//////////////////////////////////////////////////////////////////////////

  public String filename()
  {
    return pod.name(pod.typeRef(self).typeName) + ".fcode";
  }

  public void write() throws IOException
  {
    FStore.Output out = pod.store.write(filename());

    out.u4(FCodeMagic);
    out.u4(FCodeVersion);

    out.u2(fields.length);
    for (int i=0; i<fields.length; ++i)
      fields[i].write(out);

    out.u2(methods.length);
    for (int i=0; i<methods.length; ++i)
      methods[i].write(out);

    out.u2(facets.length);
    for (int i=0; i<facets.length; ++i)
      facets[i].write(out);

    out.close();
  }

  public void read() throws IOException
  {
    read(pod.store.read(filename()));
  }

  public void read(FStore.Input in) throws IOException
  {
// TODO
int fieldCount = in.u2();
if (fieldCount == 0x0FC0)
{
  in.u2();
  in.u4();
  fieldCount = in.u2();
}
fields = new FField[fieldCount];

//    fields = new FField[in.u2()];
    for (int i=0; i<fields.length; ++i)
      fields[i] = new FField().read(in);

    methods = new FMethod[in.u2()];
    for (int i=0; i<methods.length; ++i)
      methods[i] = new FMethod().read(in);

    facets = new FFacet[in.u2()];
    for (int i=0; i<facets.length; ++i)
      facets[i] = new FFacet().read(in);

    hollow = false;
    in.close();
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  public boolean hollow = true; // have we only read meta-data
  public FPod pod;              // parent pod
  public int self;              // self typeRef index
  public int flags;             // bitmask
  public int base;              // base typeRef index
  public int[] mixins;          // mixin TypeRef indexes
  public FField[] fields;       // fields
  public FMethod[] methods;     // methods
  public FFacet[] facets;       // facets on type itself

}