//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   05  Brian Frank  Creation
//
package fanx.test;

import java.io.*;
import java.io.File;
import java.util.*;
import fanx.compiler.*;
import fanx.compiler.Compiler;
import fan.sys.*;
import fan.sys.Float;
import fanx.fcode.*;
import fanx.util.*;
import fanx.tools.*;
import fan.sys.*;

/**
 * FPodIOTest verifies reading/writing the fcode to zip file.
 */
public class FPodIOTest
  extends Test
{

  String src =
  "class Foo\n" +
  "{\n" +
  "  Bool b() { return true }\n" +
  "  Int  i() { return 3 }\n" +
  "  Float r() { return 7.0 }\n" +
  "  Str  s() { return \"hello\" }\n" +
  "  Obj  d() { return 66ns }\n" +
  "  Int add(Int x, Int y) { return x + y }\n" +
  "  const static foo := 23\n" +
  "}\n";

  public boolean skip()
  {
    // can't run compiler tests if dynamic class loading not supported
    return Sys.usePrecompiledOnly;
  }

  public void run()
    throws Exception
  {
    clean();
    generate();
    compile();
    read();
    verify();
    //clean();
  }

  void clean()
    throws Exception
  {
    FileUtil.delete(dir());
  }

  void generate()
    throws Exception
  {
    dir().mkdirs();
    PrintWriter out = new PrintWriter(new FileWriter(srcFile()));
    out.print(src);
    out.close();
  }

  void compile()
    throws Exception
  {
    // use fanc to test it can be done cleanly from cmd line
    String log = verbose ? "-v" : "-silent";
    new Fanc().run(new String[] { log, "-d ", ""+dir(), ""+dir() });
  }

  void read()
    throws Exception
  {
    pod = new FPod("scratch", FStore.openToRead(new java.util.zip.ZipFile(podFile())));
    pod.read();
    //pod.store.close();

    if (verbose) new FPrinter(pod).all();
  }

  void verify()
    throws Exception
  {
    Pod realPod = Pod.find(pod.podName, false, pod);
    Class cls = realPod.findType("Foo", true).emit();
    Object o = cls.newInstance();
    verify(get(cls, "foo"), Int.make(23));
    verify(invoke(o, "b"), Bool.True);
    verify(invoke(o, "i"), Int.make(3));
    verify(invoke(o, "r"), Float.make(7));
    verify(invoke(o, "s"), Str.make("hello"));
    verify(invoke(o, "d"), Duration.make(66));
    verify(invoke(o, "add", ints(2, 8)), Int.make(10));
  }

  File dir()     { return new File(Sys.HomeDir, "tmp/scratch");  }
  File srcFile() { return new File(dir(), "Scratch.fan");  }
  File podFile() { return new File(dir(), "scratch.pod");  }

  FPod pod;
}