//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   1 Jan 06  Brian Frank  Creation
//
package fanx.test;

import java.lang.reflect.*;
import fan.sys.*;
import fanx.util.*;
import fan.sys.*;

/**
 * LineNumTest
 */
public class LineNumTest
  extends CompileTest
{

  String src  =
    "class Foo : Test\n" +      // 1
    "{\n" +                           // 2
    "  Void a() { fail(null) }\n" +   // 3
    "  Void b(Bool x)\n" +            // 4
    "  {\n" +                         // 5
    "    if (x)\n" +                  // 6
    "      fail(null)\n" +            // 7
    "    else\n" +                    // 8
    "      fail(null)\n" +            // 9
    "  }\n" +
    "}\n";


  public void doRun()
    throws Exception
  {
    Class cls = compileToClass(src);
    Object o = make(cls);
    verify(o, "a", null, 3);
    verify(o, "b", bools(true), 7);
    verify(o, "b", bools(false), 9);
  }

  void verify(Object o, String method, Obj[] args, int expectedLine)
    throws Exception
  {
    Exception ex = null;
    try
    {
      invoke(o, method, args);
      verify(false);
    }
    catch (Exception e)
    {
      ex = e;
    }

    String trace = StrUtil.traceToString(ex);
    int file  = trace.indexOf("Memory:");
    int paren = trace.indexOf(")", file);
    int line  = Integer.parseInt(trace.substring(file+"Memory:".length(), paren));
    verify(line == expectedLine);
  }

}