//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   1 Jan 06  Brian Frank  Creation
//
package fanx.test;

import java.lang.reflect.*;
import fan.sys.*;
import fan.sys.Type;
import fanx.compiler.*;

/**
 * ParamTest
 */
public class ParamTest
  extends CompileTest
{

//////////////////////////////////////////////////////////////////////////
// Main
//////////////////////////////////////////////////////////////////////////

  public void doRun()
    throws Exception
  {
    verifyStatic();
    verifyInstance();
    verifyCtor();
    verifyStaticCall();
    verifyInstanceCall();
    verifyCtorCall();
  }

//////////////////////////////////////////////////////////////////////////
// Static
//////////////////////////////////////////////////////////////////////////

  String _static  =
    "class A\n" +
    "{\n" +
    "  static Int f(Int a := 3) { return a }\n" +
    "  static Int g(Int a := 1, Int b := 2) { return a+b }\n" +
    "  static Int h(Int a,Int b:= 2,Int c:=3, Int d:=4, Int e:=5, Int f:=6) { return a+b+c+d+e+f }\n" +
    "}\n";

  public void verifyStatic()
    throws Exception
  {
    Type type = compileToType(_static);
    Class cls = type.emit();

    verify(invoke(cls, "f", null), Int.make(3));
    verify(invoke(cls, "f", ints(-2)), Int.make(-2));

    verify(invoke(cls, "g", null), Int.make(3));
    verify(invoke(cls, "g", ints(5)), Int.make(7));
    verify(invoke(cls, "g", ints(-1, -2)), Int.make(-3));

    verify(invoke(cls, "h", ints(0)), Int.make(20));
    verify(invoke(cls, "h", ints(0, 0)), Int.make(18));
    verify(invoke(cls, "h", ints(0, 0, 0)), Int.make(15));
    verify(invoke(cls, "h", ints(0, 0, 0, 0)), Int.make(11));
    verify(invoke(cls, "h", ints(0, 0, 0, 0, 0)), Int.make(6));
    verify(invoke(cls, "h", ints(0, 0, 0, 0, 0, 0)), Int.make(0));
  }

//////////////////////////////////////////////////////////////////////////
// Instance
//////////////////////////////////////////////////////////////////////////

  String instance =
    "class A\n" +
    "{\n" +
    "  Int f(Int a := 3) { return a }\n" +
    "  Int g(Int a := 1, Int b := 2) { return a+b }\n" +
    "  Int h(Int a,Int b:= 2,Int c:=3, Int d:=4, Int e:=5, Int f:=6) { return a+b+c+d+e+f }\n" +
    "}\n";

  public void verifyInstance()
    throws Exception
  {
    Type type = compileToType(instance);
    Obj o = type.make();

    verify(invoke(o, "f", null), Int.make(3));
    verify(invoke(o, "f", ints(-2)), Int.make(-2));

    verify(invoke(o, "g", null), Int.make(3));
    verify(invoke(o, "g", ints(5)), Int.make(7));
    verify(invoke(o, "g", ints(-1, -2)), Int.make(-3));

    verify(invoke(o, "h", ints(0)), Int.make(20));
    verify(invoke(o, "h", ints(0, 0)), Int.make(18));
    verify(invoke(o, "h", ints(0, 0, 0)), Int.make(15));
    verify(invoke(o, "h", ints(0, 0, 0, 0)), Int.make(11));
    verify(invoke(o, "h", ints(0, 0, 0, 0, 0)), Int.make(6));
    verify(invoke(o, "h", ints(0, 0, 0, 0, 0, 0)), Int.make(0));
  }

//////////////////////////////////////////////////////////////////////////
// Constructor
//////////////////////////////////////////////////////////////////////////

  String ctor =
    "class A\n" +
    "{\n" +
    "  new f(Int a := 3) { x = a }\n" +
    "  new g(Int a := 1, Int b := 2) { x = a+b }\n" +
    "  new h(Int a,Int b:= 2,Int c:=3, Int d:=4, Int e:=5, Int f:=6) { x = a+b+c+d+e+f }\n" +
    "  Int x\n" +
    "}\n";

  public void verifyCtor()
    throws Exception
  {
    Class cls = compileToClass(ctor);
    Object o;

    o = invoke(cls, "f", null);  verify(get(o, "x"), Int.make(3));
    o = invoke(cls, "f", ints(-2));  verify(get(o, "x"), Int.make(-2));

    o = invoke(cls, "g", null);  verify(get(o, "x"), Int.make(3));
    o = invoke(cls, "g", ints(5));  verify(get(o, "x"), Int.make(7));
    o = invoke(cls, "g", ints(-1, -2));  verify(get(o, "x"), Int.make(-3));

    o = invoke(cls, "h", ints(0));  verify(get(o, "x"), Int.make(20));
    o = invoke(cls, "h", ints(0, 0));  verify(get(o, "x"), Int.make(18));
    o = invoke(cls, "h", ints(0, 0, 0));  verify(get(o, "x"), Int.make(15));
    o = invoke(cls, "h", ints(0, 0, 0, 0));  verify(get(o, "x"), Int.make(11));
    o = invoke(cls, "h", ints(0, 0, 0, 0, 0));  verify(get(o, "x"), Int.make(6));
    o = invoke(cls, "h", ints(0, 0, 0, 0, 0, 0));  verify(get(o, "x"), Int.make(0));
  }

//////////////////////////////////////////////////////////////////////////
// Static Call
//////////////////////////////////////////////////////////////////////////

  String staticCall  =
    "class A\n" +
    "{\n" +
    "  static Int f(Int a := 3) { return a }\n" +
    "  static Int c0() { return f() }\n" +
    "  static Int c1(Int a) { return f(a) }\n" +
    "}\n";

  public void verifyStaticCall()
    throws Exception
  {
    Type type = compileToType(staticCall);
    Class cls = type.emit();

    verify(invoke(cls, "c0", null), Int.make(3));
    verify(invoke(cls, "c1", ints(-2)), Int.make(-2));
  }

//////////////////////////////////////////////////////////////////////////
// Instance Call
//////////////////////////////////////////////////////////////////////////

  String instanceCall  =
    "class A\n" +
    "{\n" +
    "  Int f(Int a, Int b := 3) { return a+b }\n" +
    "  Int c0(Int a) { return f(a) }\n" +
    "  Int c1(Int a, Int b) { return f(a, b) }\n" +
    "}\n";

  public void verifyInstanceCall()
    throws Exception
  {
    Type type = compileToType(instanceCall);
    Obj o = type.make();

    verify(invoke(o, "c0", ints(-1)), Int.make(2));
    verify(invoke(o, "c1", ints(6, 7)), Int.make(13));
  }

//////////////////////////////////////////////////////////////////////////
// Ctor Call
//////////////////////////////////////////////////////////////////////////

  String ctorCall  =
    "class A\n" +
    "{\n" +
    "  static A c0() { return make() }\n" +
    "  static A c1(Int x) { return make(x) }\n" +
    "  private new make(Int x := 0) { this.x = x }\n" +
    "  Int x\n" +
    "}\n";

  public void verifyCtorCall()
    throws Exception
  {
    Type type = compileToType(ctorCall);
    Class cls = type.emit();
    Object o;

    o = invoke(cls, "c0", null);      verify(get(o, "x"), Int.make(0));
    o = invoke(cls, "c1", ints(77));  verify(get(o, "x"), Int.make(77));
  }

//////////////////////////////////////////////////////////////////////////
// TODO
//////////////////////////////////////////////////////////////////////////

  // TODO - error checking such as: static Void f(Foo a := this)

}