//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   27 Sep 06  Andy Frank  Creation
//

using System;
using Fanx.Fcode;

namespace Fan.Sys
{
  /// <summary>
  /// Facets encapsulate meta-data as a map of Str name/value pairs.
  /// </summary>
  public sealed class Facets : FanObj
  {

  //////////////////////////////////////////////////////////////////////////
  // FCode
  //////////////////////////////////////////////////////////////////////////

    public static Facets Make(FPod fpod, Str qname, Str slot, FFacet[] facets)
    {
      Facets r = Make();
      for (int i=0; i<facets.Length; i++)
      {
        FFacet f = facets[i];
        if ((f.flags & FConst.Reflective) != 0)
        {
          String name = fpod.Name(f.name);
          try
          {
            String val  = f.data.Utf();
            r.Set(Str.Make(name), Str.Make(val));
          }
          catch (Exception)
          {
            String spec = ""+qname;
            if (slot != null) spec += "." + slot;
            Console.WriteLine("ERROR: Invalid facet value: " + spec + "/" + name);
          }
        }
      }
      return r.RO();
    }

  //////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////

    public static Facets Make()
    {
      return new Facets(new Map(Sys.StrType, Sys.StrType));
    }

    public Facets(Map map)
    {
      this.map = map;
    }

  //////////////////////////////////////////////////////////////////////////
  // Identity
  //////////////////////////////////////////////////////////////////////////

    /*
    public Bool Equals(Obj that)
    {
      if (that instanceof Facets)
      {
        return map.equals(((Facets)that).map);
      }
      return Bool.False;
    }
    */

    public override Int Hash() { return map.Hash(); }

    public override Str ToStr() { return map.ToStr(); }

    public override Type Type() { return Sys.FacetsType; }

  //////////////////////////////////////////////////////////////////////////
  // Map
  //////////////////////////////////////////////////////////////////////////

    public Bool IsEmpty() { return map.IsEmpty(); }

    public Int Size() { return map.Size(); }

    //public List Keys() { return map.Keys(); }

  //////////////////////////////////////////////////////////////////////////
  // Getters
  //////////////////////////////////////////////////////////////////////////

    public Str Get(Str key) { return (Str)map.Get(key); }
    public Str Get(Str key, Str def)
    {
      Str val = (Str)map.Get(key);
      if (val == null) return def;
      return val;
    }

    public Bool GetBool(Str key) { return GetBool(key, null); }
    public Bool GetBool(Str key, Bool def)
    {
      Str val = (Str)map.Get(key);
      if (val == null) return def;
      return val.ToBool();
    }

    public Int GetInt(Str key) { return GetInt(key, null); }
    public Int GetInt(Str key, Int def)
    {
      Str val = (Str)map.Get(key);
      if (val == null) return def;
      return val.ToInt();
    }

    //public Float GetFloat(Str key) { return GetFloat(key, null); }
    //public Float GetFloat(Str key, Float def)
    //{
    //  Str val = (Str)map.Get(key);
    // if (val == null) return def;
    //  return val.ToFloat();
    //}

  //////////////////////////////////////////////////////////////////////////
  // Setters
  //////////////////////////////////////////////////////////////////////////

    public Facets Set(Str key, Str val)
    {
      map.Set(key, val);
      return this;
    }

    public Facets SetBool(Str key, Bool val)
    {
      return Set(key, val.ToStr());
    }

    public Facets SetInt(Str key, Int val)
    {
      return Set(key, val.ToStr());
    }

    public Facets SetFloat(Str key, Float val)
    {
      return Set(key, val.ToStr());
    }

    public Str Remove(Str key)
    {
      return (Str)map.Remove(key);
    }

    public void Clear()
    {
      map.Clear();
    }

  //////////////////////////////////////////////////////////////////////////
  // Readonly
  //////////////////////////////////////////////////////////////////////////

    public Bool IsRW()
    {
      return map.IsRW();
    }

    public Bool IsRO()
    {
      return map.IsRO();
    }

    public Facets RW()
    {
      if (IsRW().val) return this;
      return new Facets(map.RW());
    }

    public Facets RO()
    {
      if (IsRO().val) return this;
      return new Facets(map.RO());
    }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    private Map map;

  }
}