//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   27 Dec 06  Andy Frank  Creation
//

using System;
using System.Collections;

namespace Fan.Sys
{
  /// <summary>
  /// MapType is the GenericType for Maps: Foo:Bar -> K = Foo, V = Bar.
  /// </summary>
  public class MapType : GenericType
  {

  //////////////////////////////////////////////////////////////////////////
  // Constructor
  //////////////////////////////////////////////////////////////////////////

    public MapType(Type k, Type v) : base(Sys.MapType)
    {
      this.k = k;
      this.v = v;
    }

  //////////////////////////////////////////////////////////////////////////
  // Type
  //////////////////////////////////////////////////////////////////////////

    public override Int Hash()
    {
      return Signature().Hash();
    }

    public override Bool Equals(Obj obj)
    {
      if (obj is MapType)
      {
        MapType x = (MapType)obj;
        return Bool.Make(k.Equals(x.k).val && v.Equals(x.v).val);
      }
      return Bool.False;
    }

    public override Type Base()
    {
      return Sys.MapType;
    }

    public override Str Signature()
    {
      if (sig == null)
      {
        sig = Str.Make('[' + k.Signature().val + ':' + v.Signature().val + ']');
      }
      return sig;
    }

    public override bool Is(Type type)
    {
      if (type is MapType)
      {
        MapType t = (MapType)type;
        return k.Is(t.k) && v.Is(t.v);
      }
      return base.Is(type);
    }

    internal override Map MakeParams()
    {
      return new Map(Sys.StrType, Sys.TypeType)
        .Set(Str.ascii['K'], k)
        .Set(Str.ascii['V'], v)
        .Set(Str.ascii['M'], this).RO();
    }

  //////////////////////////////////////////////////////////////////////////
  // GenericType
  //////////////////////////////////////////////////////////////////////////

    public override Type GetRawType()
    {
      return Sys.MapType;
    }

    public override bool IsGenericParameter()
    {
      return v.IsGenericParameter() && k.IsGenericParameter();
    }

    protected override Type DoParameterize(Type t)
    {
      if (t == Sys.KType) return k;
      if (t == Sys.VType) return v;
      if (t == Sys.MType) return this;
      throw new InvalidOperationException(t.ToString());
    }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    public readonly Type k;
    public readonly Type v;
    private Str sig;

  }
}