//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   27 Sep 06  Andy Frank  Creation
//

using System;
using System.IO;

namespace Fan.Sys
{
  /// <summary>
  /// ReadonlyErr.
  /// </summary>
  public class ReadonlyErr : Err
  {

  //////////////////////////////////////////////////////////////////////////
  // C# Convenience
  //////////////////////////////////////////////////////////////////////////

    public new static ReadonlyErr Make(string msg)  { return Make(Str.Make(msg)); }

  //////////////////////////////////////////////////////////////////////////
  // Fan Constructors
  //////////////////////////////////////////////////////////////////////////

    public new static ReadonlyErr Make() { return Make((Str)null, (Err)null); }
    public new static ReadonlyErr Make(Str msg) { return Make(msg, null); }
    public new static ReadonlyErr Make(Str msg, Err cause)
    {
      ReadonlyErr err = new ReadonlyErr();
      Make_(err, msg, cause);
      return err;
    }

    public static void Make_(ReadonlyErr self) { Make_(self, null);  }
    public static void Make_(ReadonlyErr self, Str msg) { Make_(self, msg, null); }
    public static void Make_(ReadonlyErr self, Str msg, Err cause) { Err.Make_(self, msg, cause); }

  //////////////////////////////////////////////////////////////////////////
  // C# Constructors
  //////////////////////////////////////////////////////////////////////////

    public ReadonlyErr(Err.Val val) : base(val) {}
    public ReadonlyErr() : base(new ReadonlyErr.Val()) {}
    public ReadonlyErr(Exception actual) : base(new ReadonlyErr.Val(), actual) {}

  //////////////////////////////////////////////////////////////////////////
  // Identity
  //////////////////////////////////////////////////////////////////////////

    public override Type Type() { return Sys.ReadonlyErrType; }

  //////////////////////////////////////////////////////////////////////////
  // Val - C# Exception Type
  //////////////////////////////////////////////////////////////////////////

    public new class Val : Err.Val {}

  }
}