//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   27 Sep 06  Andy Frank  Creation
//

using System;
using Fanx.Fcode;

namespace Fan.Sys
{
  /// <summary>
  /// Slot models a member field or method of a Type.
  /// </summary>
  public abstract class Slot : FanObj
  {

  //////////////////////////////////////////////////////////////////////////
  // Constructor
  //////////////////////////////////////////////////////////////////////////

    public Slot(Type parent, Str name, int flags, Facets facets)
    {
      this.parent = parent;
      this.name   = name;
      this.qname  = parent == null ? name : Str.Make(parent.qname.val + "." + name.val);
      this.flags  = flags;
      //this.facets = facets == null ? Facets.Make().ro() : facets.ro();
      this.facets = null;
    }

  //////////////////////////////////////////////////////////////////////////
  // Methods
  //////////////////////////////////////////////////////////////////////////

    public override Type Type() { return Sys.SlotType; }

    public Type Parent()      { return parent; }
    public virtual Str Name() { return name; }
    public Str Qname()        { return qname; }
    public Bool IsField()     { return (this is Field) ? Bool.True : Bool.False; }
    public Bool IsMethod()    { return (this is Method) ? Bool.True : Bool.False; }
    public abstract Str Signature();

  //////////////////////////////////////////////////////////////////////////
  // Flags
  //////////////////////////////////////////////////////////////////////////

    public int Flags() { return flags; }
    public Bool IsAbstract()  { return Bool.Make(flags & FConst.Abstract); }
    public Bool IsCtor()      { return Bool.Make(flags & FConst.Ctor); }
    public Bool IsFinal()     { return Bool.Make(flags & FConst.Final); }
    public Bool IsInternal()  { return Bool.Make(flags & FConst.Internal); }
    public Bool IsNative()    { return Bool.Make(flags & FConst.Native); }
    public Bool IsOverride()  { return Bool.Make(flags & FConst.Override); }
    public Bool IsPrivate()   { return Bool.Make(flags & FConst.Private); }
    public Bool IsProtected() { return Bool.Make(flags & FConst.Protected); }
    public Bool IsPublic()    { return Bool.Make(flags & FConst.Public); }
    public Bool IsStatic()    { return Bool.Make(flags & FConst.Static); }
    public Bool IsSynthetic() { return Bool.Make(flags & FConst.Synthetic); }
    public Bool IsVirtual()   { return Bool.Make(flags & FConst.Virtual); }

  //////////////////////////////////////////////////////////////////////////
  // Facets
  //////////////////////////////////////////////////////////////////////////

    public Facets Facets()
    {
      return facets;
    }

    public Int LineNumber()
    {
      return Int.Make(0); //facets.getInt(Str.make(FConst.LineNumber));
    }

  //////////////////////////////////////////////////////////////////////////
  // Documentation
  //////////////////////////////////////////////////////////////////////////

    public Str Doc()
    {
      //parent.doc();  // ensure parent has loaded documentation
      //return doc;
      return Str.Make("todo");
    }

  //////////////////////////////////////////////////////////////////////////
  // Conversion
  //////////////////////////////////////////////////////////////////////////

    public override Str ToStr() { return qname; }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    internal readonly Type parent;
    internal readonly Str name;
    internal readonly Str qname;
    internal readonly int flags;
    internal readonly Facets facets;
    public Str doc;

  }
}