//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   9 Nov 06  Andy Frank  Creation
//

using System;
using System.Text;

namespace Fan.Sys
{
  /// <summary>
  /// StrBuf mutable random-access sequence of integer characters.
  /// </summary>
  public class StrBuf : FanObj
  {

  //////////////////////////////////////////////////////////////////////////
  // Construction
  //////////////////////////////////////////////////////////////////////////

    /// <summary>
    /// Create with initial capacity of 16.
    /// </summary>
    public static StrBuf Make()
    {
      return new StrBuf(new StringBuilder(16));
    }

    /// <summary>
    /// Create with specified capacity.
    /// </summary>
    public static StrBuf MakeCapacity(Int capacity)
    {
      return new StrBuf(new StringBuilder((int)capacity.val));
    }

    public StrBuf(StringBuilder sb)
    {
      this.sb = sb;
    }

  //////////////////////////////////////////////////////////////////////////
  // Methods
  //////////////////////////////////////////////////////////////////////////

    /*
    public Bool equals(Obj obj)
    {
      return val.equals(((Str)obj).val) ? Bool.True : Bool.False;
    }
    */

    public Bool IsEmpty()
    {
      return sb.Length == 0 ? Bool.True : Bool.False;
    }

    public Int Size()
    {
      return Int.Pos(sb.Length);
    }

    public StrBuf Add(Obj x)
    {
      string s = (x == null) ? "null" : x.ToStr().val;
      sb.Append(s);
      return this;
    }

    public StrBuf AddChar(Int ch)
    {
      sb.Append((char)ch.val);
      return this;
    }

    public StrBuf Insert(Int index, Obj x)
    {
      string s = (x == null) ? "null" : x.ToStr().val;
      sb.Insert((int)index.val, s);
      return this;
    }

    public StrBuf Set(Int index, Obj x)
    {
      string s = (x == null) ? "null" : x.ToStr().val;
      sb.Remove((int)index.val, s.Length);
      sb.Insert((int)index.val, s);
      return this;
    }

    public StrBuf Grow(Int size)
    {
      sb.EnsureCapacity((int)size.val);
      return this;
    }

    public StrBuf Clear()
    {
      sb.Length = 0;
      return this;
    }

    public override Str ToStr()
    {
      return Str.Make(sb.ToString());
    }

    public override Type Type()
    {
      return Sys.StrBufType;
    }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    StringBuilder sb;
  }
}