//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   20 Dec 06  Andy Frank  Creation
//

using System;
using System.Collections;
using Fanx.Util;

namespace Fan.Sys
{
  /// <summary>
  /// Test is the base class of unit tests.
  /// </summary>
  public class Test : FanObj
  {

  //////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////

    //public static void make$(Test t) {}

  //////////////////////////////////////////////////////////////////////////
  // Obj Overrides
  //////////////////////////////////////////////////////////////////////////

    public override Type Type()
    {
      return Sys.TestType;
    }

  //////////////////////////////////////////////////////////////////////////
  // Lifecycle
  //////////////////////////////////////////////////////////////////////////

    public Str Id()
    {
      //if (id == null)
      //{
      //  Str qname = Type().Qname();
      //  Integer n = (Integer)idMap.Get(qname);
      //  if (n == null) n = Integer.valueOf(0);
      //  id = Str.make(qname.val + n);
      //  idMap[qname] = Integer.valueOf(n.intValue()+1));
      //}
      return id;
    }

    public void Setup() {}

    public void Teardown() {}

  //////////////////////////////////////////////////////////////////////////
  // Verify
  //////////////////////////////////////////////////////////////////////////

    public void Verify(Bool cond) { Verify(cond, null); }
    public void Verify(Bool cond, Str msg)
    {
      if (!cond.val) Fail(msg);
      verifyCount++;
    }

    public void VerifyFalse(Bool cond) { VerifyFalse(cond, null); }
    public void VerifyFalse(Bool cond, Str msg)
    {
      if (cond.val) Fail(msg);
      verifyCount++;
    }

    public void VerifyEq(Obj expected, Obj actual) { VerifyEq(expected, actual, null); }
    public void VerifyEq(Obj expected, Obj actual, Str msg)
    {
      if (!OpUtil.CompareEQ(expected, actual).val)
      {
        if (msg == null) msg = Str.Make(expected + " != " + actual);
        Fail(msg);
      }
      if (expected != null && actual != null)
      {
        if (expected.Hash().val != actual.Hash().val)
        {
          Fail(Str.Make("Equal but different hash codes: " + expected + " ?= " + actual));
        }
      }
      verifyCount++;
    }

    public void VerifyNotEq(Obj expected, Obj actual) { VerifyNotEq(expected, actual, null); }
    public void VerifyNotEq(Obj expected, Obj actual, Str msg)
    {
      if (!OpUtil.CompareNE(expected, actual).val)
      {
        if (msg == null) msg = Str.Make(expected + " == " + actual);
        Fail(msg);
      }
      verifyCount++;
    }

    public void VerifySame(Obj expected, Obj actual) { VerifySame(expected, actual, null); }
    public void VerifySame(Obj expected, Obj actual, Str msg)
    {
      if (!OpUtil.CompareSame(expected, actual).val)
      {
        if (msg == null) msg = Str.Make(expected + " !== " + actual);
        Fail(msg);
      }
      verifyCount++;
    }

    public void VerifyNotSame(Obj expected, Obj actual) { VerifyNotSame(expected, actual, null); }
    public void VerifyNotSame(Obj expected, Obj actual, Str msg)
    {
      if (OpUtil.CompareSame(expected, actual).val)
      {
        if (msg == null) msg = Str.Make(expected + " === " + actual);
        Fail(msg);
      }
      verifyCount++;
    }

    public void VerifyErr(Type errType, Method c)
    {
      try
      {
        c.Call0();
      }
      catch(Err.Val e)
      {
        if (verbose) System.Console.WriteLine("  verifyErr: " + e);
        if (e.Err().Type() == errType) { verifyCount++; return; }
        Fail(Str.Make(e.Err().Type() + " thrown, expected " + errType));
      }
      catch (System.Exception e)
      {
        if (verbose) System.Console.WriteLine("  verifyErr: " + e);
        Err err = Fan.Sys.Err.Make(e);
        if (err.Type() == errType) { verifyCount++; return; }
        Fail(Str.Make(e.ToString() + " thrown, expected " + errType));
      }
      Fail(Str.Make("No err thrown, expected " + errType));
    }

    public void Fail() { Fail(null); }
    public void Fail(Str msg)
    {
      throw Err(msg);
    }

    private Exception Err(Str msg)
    {
      if (msg == null)
        return Fan.Sys.Err.Make("Test failed").val;
      else
        return Fan.Sys.Err.Make("Test failed: " + msg.val).val;
    }

  //////////////////////////////////////////////////////////////////////////
  // Utils
  //////////////////////////////////////////////////////////////////////////

    public File TempDir()
    {
      //if (tempDir == null)
      //{
      //  tempDir = Sys.HomeDir().Plus(Uri.Parse("tmp/test/"));
      //  tempDir.Delete();
      //  tempDir.Create();
      //}
      return tempDir;
    }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    internal static readonly Hashtable idMap = new Hashtable(); // qname -> Integer

    public int verifyCount;
    public static bool verbose;
    File tempDir = null;
    Str id = null;

  }
}