//
// Copyright (c) 2007, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   16 Jan 07  Brian Frank  Creation
//

namespace Fan.Sys
{
  /// <summary>
  /// Time represents an absolute instance in time.  Fan time is normalized
  /// as nanosecond ticks since 1 Jan 2000  GMT.
  /// </summary>
  public sealed class Time : FanObj
  {

  //////////////////////////////////////////////////////////////////////////
  // Construction
  //////////////////////////////////////////////////////////////////////////

    public static Time Now ()  { return new Time(0); } //new Time(System.currentTimeMillis()*1000000L); }

    public static Time Make(Int ticks) { return Make(ticks.val); }
    public static Time Make(long ticks)
    {
      if (ticks == 0) return Zero;
      return new Time(ticks);
    }

    private Time(long ticks)
    {
      this.ticks = ticks;
    }

  //////////////////////////////////////////////////////////////////////////
  // Identity
  //////////////////////////////////////////////////////////////////////////

    public override Bool Equals(Obj obj)
    {
      if (obj is Time)
      {
        return ticks == ((Time)obj).ticks ? Bool.True : Bool.False;
      }
      return Bool.False;
    }

    public override Int Compare(Obj obj)
    {
      long that = ((Time)obj).ticks;
      if (ticks < that) return Int.LT; return ticks  == that ? Int.EQ : Int.GT;
    }

    public override int GetHashCode()
    {
      //return (int)(ticks ^ (ticks >>> 32));
      return (int)(ticks ^ (ticks >> 32));
    }

    public override Int Hash()
    {
      return Int.Make(ticks);
    }

    public override Type Type()
    {
      return Sys.TimeType;
    }

  //////////////////////////////////////////////////////////////////////////
  // Access
  //////////////////////////////////////////////////////////////////////////

    public Int Ticks()
    {
      return Int.Make(ticks);
    }

  //////////////////////////////////////////////////////////////////////////
  // Operators
  //////////////////////////////////////////////////////////////////////////

    public Duration Minus(Time time)
    {
      return Duration.Make(ticks-time.ticks);
    }

    public Time Plus(Duration duration)
    {
      long d = duration.ticks;
      if (d == 0) return this;
      return Make(ticks+d);
    }

  //////////////////////////////////////////////////////////////////////////
  // Conversion
  //////////////////////////////////////////////////////////////////////////

    public override Str ToStr()
    {
      if (str == null)
        str = Str.Make("todo"); //format.Format(java()));
      return str;
    }

  //////////////////////////////////////////////////////////////////////////
  // C#
  //////////////////////////////////////////////////////////////////////////

    //public static Time java(long millis) { if (millis <= 0) return null; return new Time(millis*1000000L); }
    //public long java() { return ticks / 1000000L; }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    //private static const DateFormat format = new SimpleDateFormat("d-MMM-yyyy HH:mm:ss zzz");
    private static readonly Time Zero = new Time(0);

    private readonly long ticks;
    private Str str;

  }
}