//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   19 Oct 06  Andy Frank  Creation
//

using System;
using URI = System.Uri;

namespace Fan.Sys
{
  /// <summary>
  /// Uri is used to immutably represent a Universal Resource Identifier.
  /// </summary>
  public sealed class Uri : FanObj
  {

  //////////////////////////////////////////////////////////////////////////
  // Construction
  //////////////////////////////////////////////////////////////////////////

    public static Uri Parse(string s) { return Parse(Str.Make(s)); }
    public static Uri Parse(Str s)
    {
      try
      {
        return new Uri(s);
      }
      catch (Exception)
      {
        throw ArgErr.Make("Invalid Uri: `" + s + "`").val;
      }
    }

    public Uri(Str s) : this(s, new URI(s.val, UriKind.RelativeOrAbsolute)) {}

    public Uri(URI uri) : this(Str.Make(uri.ToString()), uri) {}

    private Uri(Str s, URI u)
    {
      str      = s;
      uri      = u;
      scheme   = null; //Lower(u.getScheme());
      host     = null; //Str.make(u.getHost());
      pathStr  = null; //Str.make(u.getPath());
      path     = null; //ParsePath(pathStr);
      queryStr = null; //Str.make(u.getQuery());
      query    = null; //ParseQuery(queryStr);
    }
    /*
    internal static Str Lower(String s)
    {
      if (s == null) return null;
      return Str.make(Str.lower(s));
    }

    internal static List parsePath(Str pathStr)
    {
      if (pathStr == null) return null;
      List list = new List(Sys.StrType);

      try
      {
        String p = pathStr.val;
        int start = 0, len = p.length();
        for (int i=0; i<len; ++i)
        {
          int ch = p.charAt(i);
          if (ch != '/') continue;
          if (start < i)
            list.add(Str.make(p.substring(start, i)));
          start = i+1;
        }
        if (start < len)
          list.add(Str.make(p.substring(start, len)));
      }
      catch(Exception e)
      {
        // don't let internal error bring down whole uri
        e.printStackTrace();
      }

      return list.trim().ro();
    }

    internal static Map parseQuery(Str queryStr)
    {
      if (queryStr == null) return null;
      Map map = new Map(Sys.StrType, Sys.StrType);

      try
      {
        String q = queryStr.val;
        int start = 0, eq = 0, len = q.length();
        for (int i=0; i<len; ++i)
        {
          int ch = q.charAt(i);
          if (ch == '=') eq = i;
          if (ch != '&' && ch != ';') continue;

          if (start < i)
          {
            if (start == eq)
              map.set(Str.make(q.substring(start, i)), null);
            else
              map.set(Str.make(q.substring(start, eq)), Str.make(q.substring(eq+1, i)));
          }

          start = eq = i+1;
        }

        if (start < len)
        {
          if (start == eq)
            map.set(Str.make(q.substring(start, len)), null);
          else
            map.set(Str.make(q.substring(start, eq)), Str.make(q.substring(eq+1, len)));
        }
      }
      catch(Exception e)
      {
        // don't let internal error bring down whole uri
        e.printStackTrace();
      }

      return map.ro();
    }
    */
  //////////////////////////////////////////////////////////////////////////
  // Identity
  //////////////////////////////////////////////////////////////////////////

    public override bool Equals(Object obj)
    {
      return str.val == (obj as Uri).str.val;
    }

    public override Bool Equals(Obj obj)
    {
      if (obj is Uri)
      {
        return str.Equals((obj as Uri).str);
      }
      return Bool.False;
    }

    public override int GetHashCode()
    {
      return str.GetHashCode();
    }

    public override Int Hash()
    {
      return str.Hash();
    }

    public override Str ToStr()
    {
      return str;
    }

    public override Type Type()
    {
      return Sys.UriType;
    }

  //////////////////////////////////////////////////////////////////////////
  // Components
  //////////////////////////////////////////////////////////////////////////

    public Bool IsAbs()
    {
      return scheme != null ? Bool.True : Bool.False;
    }

    public Bool IsRel()
    {
      return scheme == null ? Bool.True : Bool.False;
    }

    public Bool IsDir()
    {
      //if (pathStr != null)
      //{
      //  String p = pathStr.val;
      //  int len = p.Length;
      //  if (len > 0 && p[len-1] == '/')
      //    return Bool.True;
      //}
      return Bool.False;
    }

    //public Str Scheme()
    //{
    //  return scheme;
    //}

    //public Str Body()
    //{
    //  if (scheme == null) return str;
    //  return Str.Make(str.val.Substring(scheme.val.length()+1));
    //}

    /*
    public Str auth()
    {
      if (host == null) return null;
      int port = uri.getPort();
      String userInfo = uri.getUserInfo();
      if (port < 0)
      {
        if (userInfo == null) return host;
        else return Str.make(userInfo + '@' + host.val);
      }
      else
      {
        if (userInfo == null) return Str.make(host.val + ':' + port);
        else return Str.make(userInfo + '@' + host.val + ':' + port);
      }
    }
    */

    public Str Host()
    {
      return host;
    }

    public Str UserInfo()
    {
      return null; //Str.make(uri.getUserInfo());
    }

    public Int Port()
    {
      int port = 80; //uri.getPort();
      return port < 0 ? null : Int.Make(port);
    }

    public List Path()
    {
      return path;
    }

    public Str PathStr()
    {
      return pathStr;
    }

    public Str Name()
    {
      //if (path == null || path.sz() == 0) return null;
      //return (Str)path.last();
      return null;
    }

    public Str Basename()
    {
      //Str name = name();
      //if (name == null) return null;
      //String n = name.val;
      //int dot = n.lastIndexOf('.');
      //if (dot < 2)
      //{
      //  if (dot < 0) return name;
      //  if (n.equals(".")) return name;
      //  if (n.equals("..")) return name;
      //}
      //return Str.make(n.substring(0, dot));
      return null;
    }

    public Str Ext()
    {
      //Str name = name();
      //if (name == null) return null;
      //String n = name.val;
      //int dot = n.lastIndexOf('.');
      //if (dot < 2)
      //{
      //  if (dot < 0) return null;
      //  if (n.equals(".")) return null;
      //  if (n.equals("..")) return null;
      //}
      //return Str.make(n.substring(dot+1));
      return null;
    }

    public Map Query()
    {
      return null; //query;
    }

    public Str QueryStr()
    {
      return null; //queryStr;
    }

    public Str Frag()
    {
      return null; //Str.make(uri.getFragment());
    }

  //////////////////////////////////////////////////////////////////////////
  // Manipulation
  //////////////////////////////////////////////////////////////////////////

    /*
    public Uri normalize()
    {
      URI norm = uri.normalize();
      if (norm == uri) return this;
      return new Uri(norm);
    }

    public Uri plus(Uri join)
    {
      return new Uri(uri.resolve(join.uri));
    }

    public Uri minus(Uri base)
    {
      return new Uri(base.uri.relativize(uri));
    }

    public Uri tail() { return tail(Int.One); }
    public Uri tail(Int levels)
    {
      int lvl = (int)levels.val;
      if (path == null || path.sz() < lvl)
        throw ArgErr.make("Cannot call tail(" + lvl + ") on `" + str + "`").val;
      StringBuilder s = new StringBuilder();
      for (int i=lvl; i<path.sz(); ++i)
      {
        if (i > lvl) s.append('/');
        s.append(path.get(i));
      }
      return parse(Str.make(s.toString()));
    }
    */

  //////////////////////////////////////////////////////////////////////////
  // Resolution
  //////////////////////////////////////////////////////////////////////////

    /*
    public Obj Get() { return Get(null); }
    public Obj Get(Obj baseUri)
    {
      if (scheme != null)
      {
        throw UnsupportedErr.Make("Absolute Uris not supported yet: " + str).val;
      }

      if (baseUri == null)
        throw ArgErr.Make("Must specify base for `" + str + "`").val;

      return baseUri.trapUri(this);
    }
    */

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    internal URI uri;
    internal Str str;
    internal Str scheme;
    internal Str host;
    internal List path;
    internal Str pathStr;
    internal Map query;
    internal Str queryStr;

  }
}