//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   4 Oct 06  Andy Frank  Creation
//

using System.Collections;
using Fan.Sys;
using Fanx.Fcode;

namespace Fanx.Emit
{
  /// <summary>
  /// FClassEmit emits a normal class type.
  /// </summary>
  public class FClassEmit : FTypeEmit
  {

  //////////////////////////////////////////////////////////////////////////
  // Constructor
  //////////////////////////////////////////////////////////////////////////

    public FClassEmit(Emitter emitter, Type parent, FType type)
      : base(emitter, parent, type) {}

  //////////////////////////////////////////////////////////////////////////
  // Overrides
  //////////////////////////////////////////////////////////////////////////

    protected override string Base()
    {
      // if the base is a generic instance, then this must be a closure
      // method type (since we can't subclass List or Map).  We subclass
      // from one of the canned Method.Indirect inner classes.
      FTypeRef refer = pod.TypeRef(type.baset);
      if (refer.IsGenericInstance())
      {
        this.methodType = (MethodType)Sys.FindType(refer.sig, true);
        int paramCount = methodType.pars.Length;
        if (paramCount  > Method.MaxIndirectParams)
          return "Fan.Sys.Method/IndirectX";
        else
          return "Fan.Sys.Method/Indirect" + paramCount;
      }
      else
      {
        string baset = Nname(type.baset);
        if (baset == "Fan.Sys.Obj") return "Fan.Sys.FanObj";
        return baset;
      }
    }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

  }
}