//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   16 Oct 05  Andy Frank  Creation
//

using System;
using PERWAPI;
using Fan.Sys;
using Fanx.Fcode;
using Fanx.Util;

namespace Fanx.Emit
{

  /// <summary>
  /// FPodEmit translates FPod fcode to IL as a class called <podName>.$Pod.
  /// The pod class itself defines all the constants used by it's types.
  /// </summary>
  public class FPodEmit
  {

  //////////////////////////////////////////////////////////////////////////
  // Emit
  //////////////////////////////////////////////////////////////////////////

    public static void Emit(Emitter emitter, FPod pod)
    {
      //FPodEmit emit = new FPodEmit(pod);

      TypeAttr tattr = TypeAttr.Public | TypeAttr.Sealed;
      FieldAttr fattr = FieldAttr.Public | FieldAttr.Static;

      emitter.EmitClass("System.Object", NameUtil.Upper("Fan." + pod.podName + ".$Pod"), tattr);

      // generate constant fields other types will reference, we don't
      // initialize them, rather we do that later via reflection
      for (int i=0; i<pod.ints.Size(); i++)
        emitter.EmitField("I" + i, "Fan.Sys.Int", fattr);
      for (int i=0; i<pod.floats.Size(); i++)
        emitter.EmitField("F" + i, "Fan.Sys.Float", fattr);
      for (int i=0; i<pod.strs.Size(); i++)
        emitter.EmitField("S" + i, "Fan.Sys.Str", fattr);
      for (int i=0; i<pod.durations.Size(); i++)
        emitter.EmitField("D" + i, "Fan.Sys.Duration", fattr);
      for (int i=0; i<pod.uris.Size(); i++)
        emitter.EmitField("U" + i, "Fan.Sys.Uri", fattr);
    }

  //////////////////////////////////////////////////////////////////////////
  // Load
  //////////////////////////////////////////////////////////////////////////

    public static System.Type Load(System.Reflection.Assembly assembly, FPod pod)
    {
      System.Type type = null;
      string name = NameUtil.Upper("fan." + pod.podName + ".$Pod");

      if (Sys.usePrecompiledOnly)
        type = System.Type.GetType(name);
      else
        type = assembly.GetType(name);

      InitFields(pod, type);
      return type;
    }

    private static void InitFields(FPod pod, System.Type type)
    {
      for (int i=0; i<pod.ints.Size(); i++)
        type.GetField("I"+i).SetValue(null, pod.ints.Get(i));
      for (int i=0; i<pod.floats.Size(); i++)
        type.GetField("F"+i).SetValue(null, pod.floats.Get(i));
      for (int i=0; i<pod.strs.Size(); i++)
        type.GetField("S"+i).SetValue(null, pod.strs.Get(i));
      for (int i=0; i<pod.durations.Size(); i++)
        type.GetField("D"+i).SetValue(null, pod.durations.Get(i));
      for (int i=0; i<pod.uris.Size(); i++)
        type.GetField("U"+i).SetValue(null, pod.uris.Get(i));
    }
  }
}