//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   2 Oct 06  Andy Frank  Creation
//

using System.Collections;
using System.Reflection;
using Fan.Sys;
using Fanx.Fcode;

namespace Fanx.Emit
{
  /// <summary>
  /// FTypeEmit translates FType fcode to Java bytecode.
  /// </summary>
  public abstract class FTypeEmit : Emitter
  {

  //////////////////////////////////////////////////////////////////////////
  // Factory
  //////////////////////////////////////////////////////////////////////////

    public static System.Type EmitAndLoad(FType ftype)
    {
      /*
      FTypeEmit[] emitted = Emit(parent, ftype);
      System.Type[] types = new System.Type[emitted.Length];
      for (int i=0; i<emitted.Length; ++i)
      {
        FTypeEmit e = emitted[i];
System.Console.WriteLine("looking for... " + e.className);
        types[i] = EmitPod(ftype.pod, parent).GetType(e.className);
        //types[i] = FanClassLoader.loadClass(e.className.replace('/', '.'), e.classFile);
      }
      return types[0];
      */

      string className = ftype.pod.Nname(ftype.selft);
      return EmitPod(ftype.pod).GetType(className);
    }

    private static Assembly EmitPod(FPod pod)
    {
      string podName = pod.podName;
      Assembly assembly = (Assembly)assemblies[podName];
      if (assembly == null)
      {
        Emitter emitter = new Emitter(podName);

        // emit the pod class itself (which declares all constants)
        //FPodEmit.EmitAndLoad(emitter, pod);
        FPodEmit.Emit(emitter, pod);

        // emit all the rest of the types in this pod
        for (int i=0; i<pod.types.Length; i++)
        {
          FType ftype = pod.types[i];
          FTypeRef tref = ftype.pod.TypeRef(ftype.selft);
          Type parent = Sys.FindType(pod.Name(tref.podName), pod.Name(tref.typeName), true);

          // make sure we have reflected to setup slots
          parent.Reflect();

          new FClassEmit(emitter, parent, ftype).Emit();
        }

        byte[] buf = emitter.Commit();
        assembly = Assembly.Load(buf);
        assemblies[podName] = assembly;

        // load $Pod type
        FPodEmit.Load(assembly, pod);
      }
      return assembly;
    }

    /*
    public static FTypeEmit[] Emit(Type parent, FType ftype)
    {
      // route based on type
      if ((ftype.flags & Mixin) != 0)
      {
        // interface
        FMixinInterfaceEmit iemit = new FMixinInterfaceEmit(parent, ftype);
        iemit.emit();

        // body class
        FMixinBodyEmit bemit = new FMixinBodyEmit(parent, ftype);
        bemit.emit();

        return new FTypeEmit[] { iemit, bemit };
      }
      else if (parent.is(Sys.ErrType))
      {
        FErrEmit emitErr = new FErrEmit(parent, ftype);
        emitErr.emit();

        FErrValEmit emitErrVal = new FErrValEmit(parent, ftype);
        emitErrVal.emit();

        return new FTypeEmit[] { emitErr, emitErrVal };
      }
      else
      {
        FClassEmit emit = new FClassEmit(parent, ftype);
        emit.Emit();
        return new FTypeEmit[] { emit };
      }
    }
    */

  //////////////////////////////////////////////////////////////////////////
  // Constructor
  //////////////////////////////////////////////////////////////////////////

    protected FTypeEmit(Emitter emitter, Type parent, FType type)
    {
      this.emitter = emitter;
      this.parent  = parent;
      this.pod     = type.pod;
      this.type    = type;
    }

    /// <summary>
    /// Initialize an Emitter to generate a class with given name,
    /// super class, interfaces, and class level access flags.
    /// <summary>
    public void Init(string thisClass, string baseClass, string[] interfaces, int flags)
    {
      this.className = thisClass;
      this.baseClassName = baseClass;
      //this.cp.add(new CpDummy());  // dummy entry since constant pool starts at 1
      //this.thisClassIndex  = cls(thisClass);
      //this.superClassIndex = cls(superClass);
      //this.interfaces = new int[interfaces.length];
      //for (int i=0; i<interfaces.length; ++i)
      //  this.interfaces[i] = cls(interfaces[i]);
      //this.flags = flags;
    }
  //////////////////////////////////////////////////////////////////////////
  // Emit
  //////////////////////////////////////////////////////////////////////////

    /// <summary>
    /// Emit to IL assembly.
    /// </summary>
    public void Emit()
    {
      Init(Nname(type.selft), Base(), null, 0);
      this.selfName = className;
      //preview();
      EmitType();

      // Create ctor emit objects first so we can reference them
      // .ctor
      ctor = emitter.classDef.AddMethod(
        PERWAPI.MethAttr.Public | PERWAPI.MethAttr.HideBySig | PERWAPI.MethAttr.SpecialRTSpecialName,
        PERWAPI.ImplAttr.IL, ".ctor", emitter.FindType("System.Void"), new PERWAPI.Param[0]);
      ctor.AddCallConv(PERWAPI.CallConv.Instance);
      emitter.AddToMethodMap(selfName, ".ctor", new string[0], ctor);
      // .cctor
      cctor = emitter.classDef.AddMethod(
        PERWAPI.MethAttr.Private | PERWAPI.MethAttr.Static | PERWAPI.MethAttr.HideBySig | PERWAPI.MethAttr.SpecialRTSpecialName,
        PERWAPI.ImplAttr.IL, ".cctor", emitter.FindType("System.Void"), new PERWAPI.Param[0]);
      emitter.AddToMethodMap(selfName, ".cctor", new string[0], cctor);

      // emit class body
      for (int i=0; i<type.fields.Length;  i++)  Emit(type.fields[i]);
      for (int i=0; i<type.methods.Length; i++)  Emit(type.methods[i]);
      for (int i=0; i<type.facets.Length;  i++)  Emit(type.facets[i]);
      //for (int i=0; i<parent.mixins().sz(); ++i) emitMixinRouters((Type)parent.mixins().get(i));
      if (!hasInstanceInit) EmitInstanceInit(null);
      if (!hasStaticInit) EmitStaticInit(null);
      EmitTypeConstFields();
      //return classFile = pack();
    }

    /*
    /// <summary>
    /// Emit to bytecode classfile.
    /// </summary>
    public Box Emit()
    {
      init(jname(type.self), base(), mixins(), jflags(type.flags));
      this.selfName = className;
      preview();
      emitType();
      for (int i=0; i<type.fields.length; ++i)   emit(type.fields[i]);
      for (int i=0; i<type.methods.length; ++i)  emit(type.methods[i]);
      for (int i=0; i<type.facets.length; ++i)   emit(type.facets[i]);
      for (int i=0; i<parent.mixins().sz(); ++i) emitMixinRouters((Type)parent.mixins().get(i));
      if (!hasInstanceInit) emitInstanceInit(null);
      if (!hasStaticInit) emitStaticInit(null);
      emitTypeConstFields();
      return classFile = pack();
    }
    */

    protected abstract string Base();

    /*
    internal string[] Mixins()
    {
      String[] mixins = new String[type.mixins.length];
      for (int i=0; i<mixins.length; ++i)
        mixins[i] = jname(type.mixins[i]);
      return mixins;
    }

    private void Preview()
    {
      this.isNative = (type.flags & FConst.Native) != 0;
      if (!this.isNative)
      {
        for (int i=0; i<type.methods.length; ++i)
          if ((type.methods[i].flags & Native) != 0) { this.isNative = true; break; }
      }
    }
    */

    private void EmitType()
    {
      emitter.EmitClass(Base(), className, PERWAPI.TypeAttr.Public);

      // generate private static Type $Type; set in clinit
      typeField = emitter.classDef.AddField(
        PERWAPI.FieldAttr.Private | PERWAPI.FieldAttr.Static,
        "$Type", emitter.FindType("Fan.Sys.Type"));

      // generate Type() instance method
      PERWAPI.MethodDef m = emitter.classDef.AddMethod(
        PERWAPI.MethAttr.Public | PERWAPI.MethAttr.Virtual, PERWAPI.ImplAttr.IL,
        "Type", emitter.FindType("Fan.Sys.Type"), new PERWAPI.Param[0]);
      m.AddCallConv(PERWAPI.CallConv.Instance);
      emitter.AddToMethodMap(className, "Type", new string[0], m);

      PERWAPI.CILInstructions code = m.CreateCodeBuffer();
      code.FieldInst(PERWAPI.FieldOp.ldsfld, typeField);
      code.Inst(PERWAPI.Op.ret);

      /*
      // generate peer field if native
      if (isNative)
        peerField = emitField("peer", "L" + className + "Peer;", EmitConst.PUBLIC);
      */
    }

    /// <summary>
    /// Emit a facet.
    /// </summary>
    private void Emit(FFacet f)
    {
      //String name = pod.name(f.name);
      //if (name.equals(SourceFile))
      //{
      //  AttrEmit attr = emitAttr("SourceFile");
      //  attr.info.u2(utf(f.data.utf()));
      //}
    }

    /// <summary>
    /// Emit a field.
    /// </summary>
    protected void Emit(FField f)
    {
      emitter.EmitField(Name(f.name), Nname(f.type), FieldFlags(f.flags));
    }

    /// <summary>
    /// Emit a method.
    /// </summary>
    private void Emit(FMethod m)
    {
      string name = Name(m.name);
      bool isNative = (m.flags & FConst.Native) != 0;
      bool isCtor   = (m.flags & FConst.Ctor)   != 0;

      // static$init -> .cctor
      // instance$init -> .ctor
      if (name == "static$init")   { EmitStaticInit(m); return; }
      if (name == "instance$init") { EmitInstanceInit(m); return; }

      // handle native/constructor/normal method
      if (isNative)
      {
        //new FMethodEmit(this, m).emitNative();
        throw new System.Exception("(FTypeEmit.Emit() -> Don't handle this yet!!!!");
      }
      else if (isCtor)
      {
        new FMethodEmit(this, m).EmitCtor();
      }
      else
      {
        new FMethodEmit(this, m).EmitStandard();
      }
    }

    protected void EmitInstanceInit(FMethod m)
    {
      hasInstanceInit = true;
      PERWAPI.CILInstructions code = ctor.CreateCodeBuffer();

      //code.Inst(Op.ldarg_0);
      //code.Inst(...)  call    instance void [mscorlib]System.Object::.ctor()

      //MethodEmit me = EmitMethod("<init>", "()V", EmitConst.PUBLIC);
      // initalize code to call super
      //CodeEmit code = me.emitCode();
      //code.op(ALOAD_0);

      // if closure, push MethodType static field
      //if (methodType != null)
      //{
      //  code.op2(GETSTATIC, typeField.ref());
      //  code.op2(CHECKCAST, cls("fan/sys/MethodType"));
      //  code.op2(INVOKESPECIAL, method(superClassName +".<init>(Lfan/sys/MethodType;)V"));
      //}
      //else
      //{
      //  code.op2(INVOKESPECIAL, method(superClassName +".<init>()V"));
      //}

      // make peer
      //if (isNative)
      //{
      //  code.op(ALOAD_0);  // for putfield
      //  code.op(DUP);      // for arg to make
      //  code.op2(INVOKESTATIC, method(selfName + "Peer.make(L" + className + ";)L" + className + "Peer;"));
      //  code.op2(PUTFIELD, peerField.ref());
      //}

      if (m == null)
        code.Inst(PERWAPI.Op.ret);
      else
        new FCodeEmit(this, m, code).Emit();
    }

    internal void EmitStaticInit(FMethod m)
    {
      hasStaticInit = true;
      PERWAPI.CILInstructions code = cctor.CreateCodeBuffer();

      // set $Type field with type (if we this is a closure,
      // then the MethodType will be the type exposed)
      if (!parent.IsMixin().val)
      {
        Type t = parent;
        if (parent.Base() is MethodType) t = parent.Base();

        code.ldstr(t.Signature().val);
        PERWAPI.Method findType = emitter.FindMethod("Fan.Sys.Sys", "FindType",
          new string[] { "System.String" }, "Fan.Sys.Type");
        code.MethInst(PERWAPI.MethodOp.call, findType);
        code.FieldInst(PERWAPI.FieldOp.stsfld, typeField);
      }

      if (m == null)
        code.Inst(PERWAPI.Op.ret);
      else
        new FCodeEmit(this, m, code).Emit();
    }

    internal void EmitTypeConstFields()
    {
      // if during the emitting of all the methods we ran across a non-sys
      // LoadType opcode, then we need to generate a static field called
      // type${pod}${name} we can use to cache the type once it is looked up
//      if (typeLiteralFields == null) return;
//      Iterator it = typeLiteralFields.values().iterator();
//      while (it.hasNext())
//      {
//        String fieldName = (String)it.next();
//        emitField(fieldName, "Lfan/sys/Type;", EmitConst.PRIVATE|EmitConst.STATIC);
//      }
    }

    /*
    private void EmitMixinRouters(Type type)
    {
      // generate router method for each concrete instance method
      List methods = type.methods();
      for (int i=0; i<methods.sz(); ++i)
      {
        Method m = (Method)methods.get(i);

        // only emit router for non-abstract instance methods
        if (m.isStatic().val || m.isAbstract().val) continue;

        // only emit the router unless this is the exact one I inherit
        String name = m.name().val;
        if (parent.slot(name, true).parent() != type) continue;

        // do it
        new FMethodEmit(this).emitMixinRouter(m);
      }
    }
    */

  //////////////////////////////////////////////////////////////////////////
  // Utils
  //////////////////////////////////////////////////////////////////////////

    /// <summary>
    /// Map Fan flags to .NET field flags.  Note we emit protected as
    /// public and internal/private as package-private so that we don't
    /// need to deal with scope issues for accessors like closures and
    /// helper classes.
    /// </summary>
    internal static PERWAPI.FieldAttr FieldFlags(int fflags)
    {
      PERWAPI.FieldAttr nflags = 0;
      if ((fflags & FConst.Private)   != 0) nflags |= PERWAPI.FieldAttr.Assembly;
      if ((fflags & FConst.Protected) != 0) nflags |= PERWAPI.FieldAttr.Public;
      if ((fflags & FConst.Public)    != 0) nflags |= PERWAPI.FieldAttr.Public;
      if ((fflags & FConst.Internal)  != 0) nflags |= PERWAPI.FieldAttr.Assembly;
      if ((fflags & FConst.Static)    != 0) nflags |= PERWAPI.FieldAttr.Static;
      return nflags;
    }

    /// <summary>
    /// Given a Fan qname index, map to a .NET type name: sys/Bool
    /// </summary>
    internal string Nname(int index)
    {
      return pod.Nname(index);
    }

    /// <summary>
    /// Map a simple name index to it's string value
    /// </summary>
    internal string Name(int index)
    {
      return pod.Name(index);
    }

    /*
    /// <summary>
    /// Map a simple name index to it's string value
    /// </summary>
    internal string Name(int index)
    {
      return pod.Name(index);
    }

    /// <summary>
    /// Get method ref to sys::Sys.type(String, bool)
    /// </summary>
    internal int SysFindType()
    {
      if (sysFindType == 0)
        sysFindType = method("fan/sys/Sys.findType(Ljava/lang/String;Z)Lfan/sys/Type;");
      return sysFindType ;
    }
    private int sysFindType;
    */

  //////////////////////////////////////////////////////////////////////////
  // Cached CpInfo
  //////////////////////////////////////////////////////////////////////////

    internal PERWAPI.Method CompareEQ, CompareEQz;
    internal PERWAPI.Method CompareNE, CompareNEz;
    internal PERWAPI.Method Compare;
    internal PERWAPI.Method CompareLT, CompareLTz;
    internal PERWAPI.Method CompareLE, CompareLEz;
    internal PERWAPI.Method CompareGE, CompareGEz;
    internal PERWAPI.Method CompareGT, CompareGTz;
    internal PERWAPI.Method CompareSame;
    //internal PERWAPI.Method CompareNotSame;
    internal PERWAPI.Method CompareNull;
    internal PERWAPI.Method CompareNotNull;
    //internal int IsViaType;
    //internal int AsViaType;
    internal PERWAPI.Method BoolMake;
    internal PERWAPI.Field BoolTrue;
    internal PERWAPI.Field BoolFalse;
    //internal PERWAPI.Field BoolVal;
    //internal int IntVal;
    internal PERWAPI.Method ErrMake;
    internal PERWAPI.Field ErrVal;

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    public string className;
    public string baseClassName;

    internal PERWAPI.MethodDef ctor;   // cached .ctor emit object
    internal PERWAPI.MethodDef cctor;  // cahced .cctor emit object

    //public Box classFile;
    internal Emitter emitter;                // The emitter for the assembly for this type
    internal Type parent;
    internal FPod pod;
    internal FType type;
    internal string selfName;              // type name to use as self (for mixin body - this is interface)
    internal PERWAPI.Field typeField;      // private static final Type $Type
    //internal FieldEmit peerField;          // public static final TypePeer peer
    internal bool hasInstanceInit;         // true if we already emitted <init>
    internal bool hasStaticInit;           // true if we already emitted <clinit>
    internal MethodType methodType;        // if type is a closure
    //internal Hashtable typeLiteralFields;  // signature Strings we need to turn into cached fields
    internal bool isNative = false;        // do we have any native methods requiring a peer

    private static Hashtable assemblies = new Hashtable();  // assembly cache

  }
}