//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   23 Dec 06  Andy Frank  Creation
//

using Fan.Sys;

namespace Fanx.Test
{
  /// <summary>
  /// CallTest.
  /// </summary>
  public class CallTest : CompileTest
  {

  //////////////////////////////////////////////////////////////////////////
  // Main
  //////////////////////////////////////////////////////////////////////////

    public override void Run()
    {
      VerifyReflectCalls();
    }

  //////////////////////////////////////////////////////////////////////////
  // Reflect Calls
  //////////////////////////////////////////////////////////////////////////

    private void VerifyReflectCalls()
    {
      string src = "class ReflectCallTest\n";
      src += "{\n";
      src += "  Int i0() { return 0 }\n";
      src += "  Int i1(Int a) { return a }\n";
      src += "  Int i2(Int a, Int b) { return a+b }\n";
      src += "  Int i3(Int a, Int b, Int c) { return a+b+c }\n";
      src += "  Int i4(Int a, Int b, Int c, Int d) { return a+b+c+d }\n";
      src += "  Int i5(Int a, Int b, Int c, Int d, Int e) { return a+b+c+d+e }\n";
      src += "  Int i6(Int a, Int b, Int c, Int d, Int e, Int f) { return a+b+c+d+e+f }\n";
      src += "  Int i7(Int a, Int b, Int c, Int d, Int e, Int f, Int g) { return a+b+c+d+e+f+g }\n";
      src += "  Int i8(Int a, Int b, Int c, Int d, Int e, Int f, Int g, Int h) { return a+b+c+d+e+f+g+h }\n";
      src += "  Int i9(Int a, Int b, Int c, Int d, Int e, Int f, Int g, Int h, Int i) { return a+b+c+d+e+f+g+h+i }\n";
      src += "  Int i10(Int a, Int b, Int c, Int d, Int e, Int f, Int g, Int h, Int i, Int j) { return a+b+c+d+e+f+g+h+i+j }\n";
      src += "  Int i11(Int a, Int b, Int c, Int d, Int e, Int f, Int g, Int h, Int i, Int j, Int k) { return a+b+c+d+e+f+g+h+i+j+k }\n";
      src += "  Int i12(Int a, Int b, Int c, Int d, Int e, Int f, Int g, Int h, Int i, Int j, Int k, Int l) { return a+b+c+d+e+f+g+h+i+j+k+l }\n";
      src += "  static Int s0() { return 0 }\n";
      src += "  static Int s1(Int a) { return a }\n";
      src += "  static Int s2(Int a, Int b) { return a+b }\n";
      src += "  static Int s3(Int a, Int b, Int c) { return a+b+c }\n";
      src += "  static Int s4(Int a, Int b, Int c, Int d) { return a+b+c+d }\n";
      src += "  static Int s5(Int a, Int b, Int c, Int d, Int e) { return a+b+c+d+e }\n";
      src += "  static Int s6(Int a, Int b, Int c, Int d, Int e, Int f) { return a+b+c+d+e+f }\n";
      src += "  static Int s7(Int a, Int b, Int c, Int d, Int e, Int f, Int g) { return a+b+c+d+e+f+g }\n";
      src += "  static Int s8(Int a, Int b, Int c, Int d, Int e, Int f, Int g, Int h) { return a+b+c+d+e+f+g+h }\n";
      src += "  static Int s9(Int a, Int b, Int c, Int d, Int e, Int f, Int g, Int h, Int i) { return a+b+c+d+e+f+g+h+i }\n";
      src += "  static Int s10(Int a, Int b, Int c, Int d, Int e, Int f, Int g, Int h, Int i, Int j) { return a+b+c+d+e+f+g+h+i+j }\n";
      src += "  static Int s11(Int a, Int b, Int c, Int d, Int e, Int f, Int g, Int h, Int i, Int j, Int k) { return a+b+c+d+e+f+g+h+i+j+k }\n";
      src += "  static Int s12(Int a, Int b, Int c, Int d, Int e, Int f, Int g, Int h, Int i, Int j, Int k, Int l) { return a+b+c+d+e+f+g+h+i+j+k+l }\n";
      src += "}\n";

      Type t = CompileToFanType(src);
      for (int i=0; i<12; i++)
        VerifyReflectCalls(t, i);
    }

    private void VerifyReflectCalls(Type t, int n)
    {
      Obj obj = t.Make();
      Method ms = t.Method("s" + n, true);
      Method mi = t.Method("i" + n, true);

  //System.Console.WriteLine("-- " + ms);
  //System.Console.WriteLine("   " + mi);

      // anything less than n should throw Err
      for (int i=0; i<n; i++)
      {
        Err err = null;
        try { CallList(obj, ms, i); } catch (Err.Val e) { err = e.Err(); }
  //System.out.println("  " + err);
        Verify(err != null);
        Verify(err.Message().val.StartsWith("Too few arguments: " + i));

        err = null;
        try { CallList(obj, mi, i); } catch (Err.Val e) { err = e.Err(); }
  //System.out.println("  " + err);
        Verify(err != null);
        Verify(err.Message().val.StartsWith("Too few arguments: " + (i+1)));

        if (i <= Method.MaxIndirectParams)
        {
          err = null;
          try { CallIndirect(ms, i); } catch (Err.Val e) { err = e.Err(); }
  //System.out.println("  " + err);
          Verify(err != null);
          Verify(err.Message().val.StartsWith("Too few arguments: " + i));
        }

        if (i+1 <= Method.MaxIndirectParams)
        {
          err = null;
          try { CallIndirect(obj, mi, i); } catch (Err.Val e) { err = e.Err(); }
    //System.out.println("  " + err);
          Verify(err != null);
          Verify(err.Message().val.StartsWith("Too few arguments: " + (i+1)));
        }
      }

      // compute result 0->0, 1->1, 2->3, 3->7, etc
      int result = 0;
      for (int i=0; i<=n; i++) result += i;

      // anything n and over should be ok
      for (int i=n; i<12; i++)
      {
  //System.out.println("  verify " + i + " -> " + n);
        Verify(CallList(obj, ms, i), Int.Make(result));
        Verify(CallList(obj, mi, i), Int.Make(result));
        if (i <= Method.MaxIndirectParams)
          Verify(CallIndirect(ms, i), Int.Make(result));
        if (i+1 <= Method.MaxIndirectParams)
          Verify(CallIndirect(obj, mi, i), Int.Make(result));
      }
    }

    Int CallList(Obj obj, Method m, int argn)
    {
      List list = new List(Sys.ObjType);
      if (!m.IsStatic().val) list.Add(obj);
      for (int i=0; i<argn; i++) list.Add(Int.Make(1+i));
      return (Int)m.Call(list);
    }

    Int CallIndirect(Method m, int argn)
    {
      switch (argn)
      {
        case 0:  return (Int)m.Call0();
        case 1:  return (Int)m.Call1(Int.Make(1));
        case 2:  return (Int)m.Call2(Int.Make(1), Int.Make(2));
        case 3:  return (Int)m.Call3(Int.Make(1), Int.Make(2), Int.Make(3));
        case 4:  return (Int)m.Call4(Int.Make(1), Int.Make(2), Int.Make(3), Int.Make(4));
        case 5:  return (Int)m.Call5(Int.Make(1), Int.Make(2), Int.Make(3), Int.Make(4), Int.Make(5));
        case 6:  return (Int)m.Call6(Int.Make(1), Int.Make(2), Int.Make(3), Int.Make(4), Int.Make(5), Int.Make(6));
        case 7:  return (Int)m.Call7(Int.Make(1), Int.Make(2), Int.Make(3), Int.Make(4), Int.Make(5), Int.Make(6), Int.Make(7));
        case 8:  return (Int)m.Call8(Int.Make(1), Int.Make(2), Int.Make(3), Int.Make(4), Int.Make(5), Int.Make(6), Int.Make(7), Int.Make(8));
        //case 9:  return (Int)m.Call9(Int.Make(1), Int.Make(2), Int.Make(3), Int.Make(4), Int.Make(5), Int.Make(6), Int.Make(7), Int.Make(8), Int.Make(9));
        //case 10: return (Int)m.Call10(Int.Make(1), Int.Make(2), Int.Make(3), Int.Make(4), Int.Make(5), Int.Make(6), Int.Make(7), Int.Make(8), Int.Make(9), Int.Make(10));
        //case 11: return (Int)m.Call11(Int.Make(1), Int.Make(2), Int.Make(3), Int.Make(4), Int.Make(5), Int.Make(6), Int.Make(7), Int.Make(8), Int.Make(9), Int.Make(10), Int.Make(11));
        //case 12: return (Int)m.Call12(Int.Make(1), Int.Make(2), Int.Make(3), Int.Make(4), Int.Make(5), Int.Make(6), Int.Make(7), Int.Make(8), Int.Make(9), Int.Make(10), Int.Make(11), Int.Make(12));
        default: Fail(); break;
      }
      return null;
    }

    Int CallIndirect(Obj obj, Method m, int argn)
    {
      switch (argn)
      {
        case 0:  return (Int)m.Call1(obj);
        case 1:  return (Int)m.Call2(obj, Int.Make(1));
        case 2:  return (Int)m.Call3(obj, Int.Make(1), Int.Make(2));
        case 3:  return (Int)m.Call4(obj, Int.Make(1), Int.Make(2), Int.Make(3));
        case 4:  return (Int)m.Call5(obj, Int.Make(1), Int.Make(2), Int.Make(3), Int.Make(4));
        case 5:  return (Int)m.Call6(obj, Int.Make(1), Int.Make(2), Int.Make(3), Int.Make(4), Int.Make(5));
        case 6:  return (Int)m.Call7(obj, Int.Make(1), Int.Make(2), Int.Make(3), Int.Make(4), Int.Make(5), Int.Make(6));
        case 7:  return (Int)m.Call8(obj, Int.Make(1), Int.Make(2), Int.Make(3), Int.Make(4), Int.Make(5), Int.Make(6), Int.Make(7));
        //case 8:  return (Int)m.Call9(obj, Int.Make(1), Int.Make(2), Int.Make(3), Int.Make(4), Int.Make(5), Int.Make(6), Int.Make(7), Int.Make(8));
        //case 9:  return (Int)m.Call10(obj, Int.Make(1), Int.Make(2), Int.Make(3), Int.Make(4), Int.Make(5), Int.Make(6), Int.Make(7), Int.Make(8), Int.Make(9));
        //case 10: return (Int)m.Call11(obj, Int.Make(1), Int.Make(2), Int.Make(3), Int.Make(4), Int.Make(5), Int.Make(6), Int.Make(7), Int.Make(8), Int.Make(9), Int.Make(10));
        //case 11: return (Int)m.Call12(obj, Int.Make(1), Int.Make(2), Int.Make(3), Int.Make(4), Int.Make(5), Int.Make(6), Int.Make(7), Int.Make(8), Int.Make(9), Int.Make(10), Int.Make(11));
        default: Fail(); break;
      }
      return null;
    }

  }
}