//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   19 Oct 06  Andy Frank  Creation
//

using Object = System.Object;
using Fan.Sys;

namespace Fanx.Test
{
  /// <summary>
  /// ExprTest
  /// </summary>
  public class ExprTest : CompileTest
  {

  //////////////////////////////////////////////////////////////////////////
  // Main
  //////////////////////////////////////////////////////////////////////////

    public override void Run()
    {
      VerifyCond();
      VerifyEquality();
      VerifyComparision();
      VerifyNegate();
      VerifyMultiply();
      VerifyDivide();
      VerifyMod();
      VerifyBitwise();
      VerifyAdd();
      VerifySub();
      VerifyOrderAndPrecedence();
      VerifyIncrementDecrement();
      VerifyPlusMinusCombos();
      VerifyStrAdd();
      VerifyLocalVar();
      VerifyFieldGet();
      VerifyFieldSet();
      VerifyCalls();
      VerifyParens();
      VerifyCasts();
    }

  //////////////////////////////////////////////////////////////////////////
  // || and &&
  //////////////////////////////////////////////////////////////////////////

    void VerifyCond()
    {
      Verify("Bool f(Bool a, Bool b) { return a || b;  }", tt, or(tt));
      Verify("Bool f(Bool a, Bool b) { return a || b;  }", ft, or(ft));
      Verify("Bool f(Bool a, Bool b) { return a || b;  }", tf, or(tf));
      Verify("Bool f(Bool a, Bool b) { return a || b;  }", ff, or(ff));

      Verify("Bool f(Bool a, Bool b, Bool c) { return a || b || c;  }", ttt, or(ttt));
      Verify("Bool f(Bool a, Bool b, Bool c) { return a || b || c;  }", ftt, or(ftt));
      Verify("Bool f(Bool a, Bool b, Bool c) { return a || b || c;  }", tft, or(tft));
      Verify("Bool f(Bool a, Bool b, Bool c) { return a || b || c;  }", ttf, or(ttf));
      Verify("Bool f(Bool a, Bool b, Bool c) { return a || b || c;  }", fft, or(fft));
      Verify("Bool f(Bool a, Bool b, Bool c) { return a || b || c;  }", ftf, or(ftf));
      Verify("Bool f(Bool a, Bool b, Bool c) { return a || b || c;  }", tff, or(tff));
      Verify("Bool f(Bool a, Bool b, Bool c) { return a || b || c;  }", fff, or(fff));

      Verify("Bool f(Bool a, Bool b) { return a && b;  }", tt, and(tt));
      Verify("Bool f(Bool a, Bool b) { return a && b;  }", ft, and(ft));
      Verify("Bool f(Bool a, Bool b) { return a && b;  }", tf, and(tf));
      Verify("Bool f(Bool a, Bool b) { return a && b;  }", ff, and(ff));
      Verify("Bool f(Bool a, Bool b, Bool c) { return a && b && c;  }", ttt, and(ttt));
      Verify("Bool f(Bool a, Bool b, Bool c) { return a && b && c;  }", ftt, and(ftt));
      Verify("Bool f(Bool a, Bool b, Bool c) { return a && b && c;  }", tft, and(tft));
      Verify("Bool f(Bool a, Bool b, Bool c) { return a && b && c;  }", ttf, and(ttf));
      Verify("Bool f(Bool a, Bool b, Bool c) { return a && b && c;  }", fft, and(fft));
      Verify("Bool f(Bool a, Bool b, Bool c) { return a && b && c;  }", ftf, and(ftf));
      Verify("Bool f(Bool a, Bool b, Bool c) { return a && b && c;  }", tff, and(tff));
      Verify("Bool f(Bool a, Bool b, Bool c) { return a && b && c;  }", fff, and(fff));

      Verify("Bool f(Bool a, Bool b, Bool c) { return a || b && c;  }", ttt, MakeBool(true ||true &&true));
      Verify("Bool f(Bool a, Bool b, Bool c) { return a || b && c;  }", ftt, MakeBool(false||true &&true));
      Verify("Bool f(Bool a, Bool b, Bool c) { return a || b && c;  }", tft, MakeBool(true ||false&&true));
      Verify("Bool f(Bool a, Bool b, Bool c) { return a || b && c;  }", ttf, MakeBool(true ||true &&false));
      Verify("Bool f(Bool a, Bool b, Bool c) { return a || b && c;  }", fft, MakeBool(false||false&&true));
      Verify("Bool f(Bool a, Bool b, Bool c) { return a || b && c;  }", tff, MakeBool(true ||false&&false));
      Verify("Bool f(Bool a, Bool b, Bool c) { return a || b && c;  }", fff, MakeBool(false||false&&false));

      Verify("Bool f(Bool a, Bool b, Bool c) { return a && b || c; }",  ttt, MakeBool(true &&true ||true));
      Verify("Bool f(Bool a, Bool b, Bool c) { return a && b || c; }",  ftt, MakeBool(false&&true ||true));
      Verify("Bool f(Bool a, Bool b, Bool c) { return a && b || c; }",  tft, MakeBool(true &&false||true));
      Verify("Bool f(Bool a, Bool b, Bool c) { return a && b || c; }",  ttf, MakeBool(true &&true ||false));
      Verify("Bool f(Bool a, Bool b, Bool c) { return a && b || c; }",  fft, MakeBool(false&&false||true));
      Verify("Bool f(Bool a, Bool b, Bool c) { return a && b || c; }",  ftf, MakeBool(false&&true ||false));
      Verify("Bool f(Bool a, Bool b, Bool c) { return a && b || c; }",  tff, MakeBool(true &&false||false));
      Verify("Bool f(Bool a, Bool b, Bool c) { return a && b || c; }",  fff, MakeBool(false&&false||false));

      Verify("Bool f(Bool a) { return !a; }",  MakeBools(true), MakeBool(false));
      Verify("Bool f(Bool a) { return !a; }",  MakeBools(false), MakeBool(true));
      Verify("Bool f() { return !true; }",     MakeBool(false));
      Verify("Bool f() { return !false; }",    MakeBool(true));

      Verify("Bool f(Bool a, Bool b) { return !a && !b; }", tt, MakeBool(false));
      Verify("Bool f(Bool a, Bool b) { return !a && !b; }", ff, MakeBool(true));
    }

    Bool or(Bool[] b)
    {
      bool r = b[0].val;
      for (int i=1; i<b.Length; i++) r = r || b[i].val;
      return Bool.Make(r);
    }

    Bool and(Bool[] b)
    {
      bool r = b[0].val;
      for (int i=1; i<b.Length; i++) r = r && b[i].val;
      return Bool.Make(r);
    }

  //////////////////////////////////////////////////////////////////////////
  // == and !=
  //////////////////////////////////////////////////////////////////////////

    void VerifyEquality()
    {
      //
      // bool
      //
      Verify("static Bool f(Bool a, Bool b) { return a == b; }", tt, Bool.True);
      Verify("static Bool f(Bool a, Bool b) { return a == b; }", ft, Bool.False);
      Verify("static Bool f(Bool a, Bool b) { return a == b; }", tf, Bool.False);
      Verify("static Bool f(Bool a, Bool b) { return a == b; }", ff, Bool.True);

      Verify("static Bool f(Bool a, Bool b) { return a != b; }", tt, Bool.False);
      Verify("static Bool f(Bool a, Bool b) { return a != b; }", ft, Bool.True);
      Verify("static Bool f(Bool a, Bool b) { return a != b; }", tf, Bool.True);
      Verify("static Bool f(Bool a, Bool b) { return a != b; }", ff, Bool.False);

      Verify("static Bool f(Bool a, Bool b) { return a == b; }", nn, Bool.True);
      Verify("static Bool f(Bool a, Bool b) { return a == b; }", tn, Bool.False);
      Verify("static Bool f(Bool a, Bool b) { return a == b; }", nt, Bool.False);

      Verify("static Bool f(Bool a, Bool b) { return a != b; }", nn, Bool.False);
      Verify("static Bool f(Bool a, Bool b) { return a != b; }", tn, Bool.True);
      Verify("static Bool f(Bool a, Bool b) { return a != b; }", nt, Bool.True);

      Verify("static Bool f(Bool a, Bool b, Bool c) { return a == b == c; }", ttt, Bool.Make(true  == true  == true));
      Verify("static Bool f(Bool a, Bool b, Bool c) { return a == b == c; }", ftt, Bool.Make(false == true  == true));
      Verify("static Bool f(Bool a, Bool b, Bool c) { return a == b == c; }", tft, Bool.Make(true  == false == true));
      Verify("static Bool f(Bool a, Bool b, Bool c) { return a == b == c; }", ttf, Bool.Make(true  == true  == false));
      Verify("static Bool f(Bool a, Bool b, Bool c) { return a == b == c; }", fft, Bool.Make(false == false == true));
      Verify("static Bool f(Bool a, Bool b, Bool c) { return a == b == c; }", tff, Bool.Make(true  == false == false));
      Verify("static Bool f(Bool a, Bool b, Bool c) { return a == b == c; }", ftf, Bool.Make(false == true  == false));
      Verify("static Bool f(Bool a, Bool b, Bool c) { return a == b == c; }", fff, Bool.Make(false == false == false));

      Verify("static Bool f(Bool a, Bool b, Bool c) { return a != b != c; }", ttt, Bool.Make(true  != true  != true));
      Verify("static Bool f(Bool a, Bool b, Bool c) { return a != b != c; }", ftt, Bool.Make(false != true  != true));
      Verify("static Bool f(Bool a, Bool b, Bool c) { return a != b != c; }", tft, Bool.Make(true  != false != true));
      Verify("static Bool f(Bool a, Bool b, Bool c) { return a != b != c; }", ttf, Bool.Make(true  != true  != false));
      Verify("static Bool f(Bool a, Bool b, Bool c) { return a != b != c; }", fft, Bool.Make(false != false != true));
      Verify("static Bool f(Bool a, Bool b, Bool c) { return a != b != c; }", tff, Bool.Make(true  != false != false));
      Verify("static Bool f(Bool a, Bool b, Bool c) { return a != b != c; }", ftf, Bool.Make(false != true  != false));
      Verify("static Bool f(Bool a, Bool b, Bool c) { return a != b != c; }", fff, Bool.Make(false != false != false));

      //
      // int
      //
      Verify("static Bool f(Int a, Int b) { return a == b; }", MakeInts(0, 0),   Bool.True);
      Verify("static Bool f(Int a, Int b) { return a == b; }", MakeInts(1, 0),   Bool.False);
      Verify("static Bool f(Int a, Int b) { return a == b; }", MakeInts(0, 1),   Bool.False);
      Verify("static Bool f(Int a, Int b) { return a == b; }", MakeInts(1, 1),   Bool.True);
      Verify("static Bool f(Int a, Int b) { return a == b; }", MakeInts(-1, -1), Bool.True);
      Verify("static Bool f(Int a, Int b) { return a == b; }", MakeInts(-1, -2), Bool.False);
      Verify("static Bool f(Int a, Int b) { return a != b; }", MakeInts(0, 0),   Bool.False);
      Verify("static Bool f(Int a, Int b) { return a != b; }", MakeInts(1, 0),   Bool.True);
      Verify("static Bool f(Int a, Int b) { return a != b; }", MakeInts(0, 1),   Bool.True);
      Verify("static Bool f(Int a, Int b) { return a != b; }", MakeInts(1, 1),   Bool.False);
      Verify("static Bool f(Int a, Int b) { return a != b; }", MakeInts(77, -3), Bool.True);
      Verify("static Bool f(Int a, Int b) { return a == b; }", new Obj[] { null,null  },     Bool.True);
      Verify("static Bool f(Int a, Int b) { return a == b; }", new Obj[] { Int.Zero, null }, Bool.False);
      Verify("static Bool f(Int a, Int b) { return a == b; }", new Obj[] { null, Int.Zero }, Bool.False);
      Verify("static Bool f(Int a, Int b) { return a != b; }", new Obj[] { null,null  },     Bool.False);
      Verify("static Bool f(Int a, Int b) { return a != b; }", new Obj[] { Int.Zero, null }, Bool.True);
      Verify("static Bool f(Int a, Int b) { return a != b; }", new Obj[] { null, Int.Zero }, Bool.True);

      //
      // floats
      //
      Verify("static Bool f(Float a, Float b) { return a == b; }", MakeFloats(0, 0),   Bool.True);
      Verify("static Bool f(Float a, Float b) { return a == b; }", MakeFloats(1, 0),   Bool.False);
      Verify("static Bool f(Float a, Float b) { return a == b; }", MakeFloats(0, 1),   Bool.False);
      Verify("static Bool f(Float a, Float b) { return a == b; }", MakeFloats(1, 1),   Bool.True);
      Verify("static Bool f(Float a, Float b) { return a == b; }", MakeFloats(-1, -1), Bool.True);
      Verify("static Bool f(Float a, Float b) { return a == b; }", MakeFloats(-1, -2), Bool.False);
      Verify("static Bool f(Float a, Float b) { return a != b; }", MakeFloats(0, 0),   Bool.False);
      Verify("static Bool f(Float a, Float b) { return a != b; }", MakeFloats(1, 0),   Bool.True);
      Verify("static Bool f(Float a, Float b) { return a != b; }", MakeFloats(0, 1),   Bool.True);
      Verify("static Bool f(Float a, Float b) { return a != b; }", MakeFloats(1, 1),   Bool.False);
      Verify("static Bool f(Float a, Float b) { return a != b; }", MakeFloats(77, -3), Bool.True);
      Verify("static Bool f(Float a, Float b) { return a == b; }", new Obj[] { null,null  },     Bool.True);
      Verify("static Bool f(Float a, Float b) { return a == b; }", new Obj[] { Float.Zero, null }, Bool.False);
      Verify("static Bool f(Float a, Float b) { return a == b; }", new Obj[] { null, Float.Zero }, Bool.False);
      Verify("static Bool f(Float a, Float b) { return a != b; }", new Obj[] { null,null  },     Bool.False);
      Verify("static Bool f(Float a, Float b) { return a != b; }", new Obj[] { Float.Zero, null }, Bool.True);
      Verify("static Bool f(Float a, Float b) { return a != b; }", new Obj[] { null, Float.Zero }, Bool.True);

      //
      // str
      //
      Verify("static Bool f(Str a, Str b) { return a == b; }", MakeStrs(null, null), Bool.True);
      Verify("static Bool f(Str a, Str b) { return a == b; }", MakeStrs("a",  null), Bool.False);
      Verify("static Bool f(Str a, Str b) { return a == b; }", MakeStrs(null, "a"),  Bool.False);
      Verify("static Bool f(Str a, Str b) { return a == b; }", MakeStrs("a", "a"),   Bool.True);
      Verify("static Bool f(Str a, Str b) { return a == b; }", MakeStrs("a", "b"),   Bool.False);
      Verify("static Bool f(Str a, Str b) { return a != b; }", MakeStrs(null, null), Bool.False);
      Verify("static Bool f(Str a, Str b) { return a != b; }", MakeStrs("a",  null), Bool.True);
      Verify("static Bool f(Str a, Str b) { return a != b; }", MakeStrs(null, "a"),  Bool.True);
      Verify("static Bool f(Str a, Str b) { return a != b; }", MakeStrs("a", "a"),   Bool.False);
      Verify("static Bool f(Str a, Str b) { return a != b; }", MakeStrs("a", "b"),   Bool.True);

      //
      // Duration
      //
      Verify("static Bool f(Duration a, Duration b) { return a == b; }", MakeDurs(0, 0),  Bool.True);
      Verify("static Bool f(Duration a, Duration b) { return a == b; }", MakeDurs(20, 0), Bool.False);
      Verify("static Bool f(Duration a, Duration b) { return a != b; }", MakeDurs(0, 0),  Bool.False);
      Verify("static Bool f(Duration a, Duration b) { return a != b; }", MakeDurs(20, 0), Bool.True);

      //
      // same ===
      //
      Verify("static Bool f() { return null === null; }",        Bool.True);
      Verify("static Bool f() { return 5 === null; }",           Bool.False);
      Verify("static Bool f() { return null === 5; }",           Bool.False);
      Verify("static Bool f() { return 5 === 5; }",              Bool.True);
      Verify("static Bool f() { return \"x\" === \"x\"; }",      Bool.True);
      Verify("static Bool f() { return \"x\" === \"y\"; }",      Bool.False);
      Verify("static Bool f() { return !(\"x\" === \"y\"); }",   Bool.True);
      Verify("static Bool f(Int x, Int y) { return x === y; }",  MakeInts(256, 256), Bool.True);
      Verify("static Bool f(Int x, Int y) { return x === y; }",  MakeInts(2568888, 2568888), Bool.False);

      // auto-cast matrix
      /*
      Verify("Bool f(Int a, Float b) { return a == b; }", new Object[] { Int.make(4), Float.make(4) }, Bool.True);
      Verify("Bool f(Int a, Float b) { return a != b; }", new Object[] { Int.make(4), Float.make(4) }, Bool.False);
      Verify("Bool f(Float a, Int b) { return a == b; }", new Object[] { Float.make(-99), Int.make(-99) }, Bool.True);
      Verify("Bool f(Float a, Int b) { return a != b; }", new Object[] { Float.make(-99), Int.make(-99) }, Bool.False);
      */
    }

  //////////////////////////////////////////////////////////////////////////
  // < <= > >= <=>
  //////////////////////////////////////////////////////////////////////////

    void VerifyComparision()
    {
      //
      // MakeInts
      //
      Verify("static Bool f(Int a, Int b) { return a < b; }", MakeInts(0, 0),  Bool.Make(0 < 0));
      Verify("static Bool f(Int a, Int b) { return a < b; }", MakeInts(1, 0),  Bool.Make(1 < 0));
      Verify("static Bool f(Int a, Int b) { return a < b; }", MakeInts(0, 1),  Bool.Make(0 < 1));
      Verify("static Bool f(Int a, Int b) { return a < b; }", MakeInts(1, 1),  Bool.Make(1 < 1));
      Verify("static Bool f(Int a, Int b) { return a < b; }", new Obj[] { null, null },      Bool.False);
      Verify("static Bool f(Int a, Int b) { return a < b; }", new Obj[] { Int.Zero, null },  Bool.False);
      Verify("static Bool f(Int a, Int b) { return a < b; }", new Obj[] { null, Int.Zero },  Bool.True);

      Verify("static Bool f(Int a, Int b) { return a <= b; }", MakeInts(0,  0),   Bool.Make(0  <= 0));
      Verify("static Bool f(Int a, Int b) { return a <= b; }", MakeInts(-1, 0),   Bool.Make(-1 <= 0));
      Verify("static Bool f(Int a, Int b) { return a <= b; }", MakeInts(0,  -1),  Bool.Make(0  <= -1));
      Verify("static Bool f(Int a, Int b) { return a <= b; }", MakeInts(-1, -1),  Bool.Make(-1 <= -1));
      Verify("static Bool f(Int a, Int b) { return a <= b; }", new Obj[] { null, null },      Bool.True);
      Verify("static Bool f(Int a, Int b) { return a <= b; }", new Obj[] { Int.Zero, null },  Bool.False);
      Verify("static Bool f(Int a, Int b) { return a <= b; }", new Obj[] { null, Int.Zero },  Bool.True);

      Verify("static Bool f(Int a, Int b) { return a > b; }", MakeInts(4, 4),  Bool.Make(4 > 4));
      Verify("static Bool f(Int a, Int b) { return a > b; }", MakeInts(7, 4),  Bool.Make(7 > 4));
      Verify("static Bool f(Int a, Int b) { return a > b; }", MakeInts(4, 7),  Bool.Make(4 > 7));
      Verify("static Bool f(Int a, Int b) { return a > b; }", MakeInts(7, 7),  Bool.Make(7 > 7));
      Verify("static Bool f(Int a, Int b) { return a > b; }", new Obj[] { null, null },      Bool.False);
      Verify("static Bool f(Int a, Int b) { return a > b; }", new Obj[] { Int.Zero, null },  Bool.True);
      Verify("static Bool f(Int a, Int b) { return a > b; }", new Obj[] { null, Int.Zero },  Bool.False);

      Verify("static Bool f(Int a, Int b) { return a >= b; }", MakeInts(-2, -2),  Bool.Make(-2 >= -2));
      Verify("static Bool f(Int a, Int b) { return a >= b; }", MakeInts(+2, -2),  Bool.Make(+2 >= -2));
      Verify("static Bool f(Int a, Int b) { return a >= b; }", MakeInts(-2, +2),  Bool.Make(-2 >= +2));
      Verify("static Bool f(Int a, Int b) { return a >= b; }", MakeInts(+2, +2),  Bool.Make(+2 >= +2));
      Verify("static Bool f(Int a, Int b) { return a >= b; }", new Obj[] { null, null },      Bool.True);
      Verify("static Bool f(Int a, Int b) { return a >= b; }", new Obj[] { Int.Zero, null },  Bool.True);
      Verify("static Bool f(Int a, Int b) { return a >= b; }", new Obj[] { null, Int.Zero },  Bool.False);

      Verify("static Int f(Int a, Int b) { return a <=> b; }", MakeInts(3, 2),  Int.Make(1));
      Verify("static Int f(Int a, Int b) { return a <=> b; }", MakeInts(3, 3),  Int.Make(0));
      Verify("static Int f(Int a, Int b) { return a <=> b; }", MakeInts(2, 3),  Int.Make(-1));
      Verify("static Int f(Int a, Int b) { return a <=> b; }", new Obj[] { null, null },      Int.Make(0));
      Verify("static Int f(Int a, Int b) { return a <=> b; }", new Obj[] { Int.Zero, null },  Int.Make(1));
      Verify("static Int f(Int a, Int b) { return a <=> b; }", new Obj[] { null, Int.Zero },  Int.Make(-1));

      //
      // MakeFloats
      //
      Verify("static Bool f(Float a, Float b) { return a < b; }", MakeFloats(0, 0),  Bool.Make(0 < 0));
      Verify("static Bool f(Float a, Float b) { return a < b; }", MakeFloats(1, 0),  Bool.Make(1 < 0));
      Verify("static Bool f(Float a, Float b) { return a < b; }", MakeFloats(0, 1),  Bool.Make(0 < 1));
      Verify("static Bool f(Float a, Float b) { return a < b; }", MakeFloats(1, 1),  Bool.Make(1 < 1));

      Verify("static Bool f(Float a, Float b) { return a <= b; }", MakeFloats(0,  0),   Bool.Make(0  <= 0));
      Verify("static Bool f(Float a, Float b) { return a <= b; }", MakeFloats(-1, 0),   Bool.Make(-1 <= 0));
      Verify("static Bool f(Float a, Float b) { return a <= b; }", MakeFloats(0,  -1),  Bool.Make(0  <= -1));
      Verify("static Bool f(Float a, Float b) { return a <= b; }", MakeFloats(-1, -1),  Bool.Make(-1 <= -1));

      Verify("static Bool f(Float a, Float b) { return a > b; }", MakeFloats(4, 4),  Bool.Make(4 > 4));
      Verify("static Bool f(Float a, Float b) { return a > b; }", MakeFloats(7, 4),  Bool.Make(7 > 4));
      Verify("static Bool f(Float a, Float b) { return a > b; }", MakeFloats(4, 7),  Bool.Make(4 > 7));
      Verify("static Bool f(Float a, Float b) { return a > b; }", MakeFloats(7, 7),  Bool.Make(7 > 7));

      Verify("static Bool f(Float a, Float b) { return a >= b; }", MakeFloats(-2, -2),  Bool.Make(-2 >= -2));
      Verify("static Bool f(Float a, Float b) { return a >= b; }", MakeFloats(+2, -2),  Bool.Make(+2 >= -2));
      Verify("static Bool f(Float a, Float b) { return a >= b; }", MakeFloats(-2, +2),  Bool.Make(-2 >= +2));
      Verify("static Bool f(Float a, Float b) { return a >= b; }", MakeFloats(+2, +2),  Bool.Make(+2 >= +2));

      Verify("static Int f(Float a, Float b) { return a <=> b; }", MakeFloats(3, 2),  Int.Make(1));
      Verify("static Int f(Float a, Float b) { return a <=> b; }", MakeFloats(3, 3),  Int.Make(0));
      Verify("static Int f(Float a, Float b) { return a <=> b; }", MakeFloats(2, 3),  Int.Make(-1));

      //
      // bool
      //
      Verify("static Bool f(Bool a, Bool b) { return a < b; }", ft,  Bool.True);
      Verify("static Bool f(Bool a, Bool b) { return a < b; }", ff,  Bool.False);
      Verify("static Bool f(Bool a, Bool b) { return a < b; }", tf,  Bool.False);

      Verify("static Bool f(Bool a, Bool b) { return a <= b; }", ft,  Bool.True);
      Verify("static Bool f(Bool a, Bool b) { return a <= b; }", ff,  Bool.True);
      Verify("static Bool f(Bool a, Bool b) { return a <= b; }", tf,  Bool.False);

      Verify("static Bool f(Bool a, Bool b) { return a > b; }", ft,  Bool.False);
      Verify("static Bool f(Bool a, Bool b) { return a > b; }", ff,  Bool.False);
      Verify("static Bool f(Bool a, Bool b) { return a > b; }", tf,  Bool.True);

      Verify("static Bool f(Bool a, Bool b) { return a >= b; }", ft,  Bool.False);
      Verify("static Bool f(Bool a, Bool b) { return a >= b; }", ff,  Bool.True);
      Verify("static Bool f(Bool a, Bool b) { return a >= b; }", tf,  Bool.True);

      Verify("static Int f(Bool a, Bool b) { return a <=> b; }", tf,  Int.Make(1));
      Verify("static Int f(Bool a, Bool b) { return a <=> b; }", tt,  Int.Make(0));
      Verify("static Int f(Bool a, Bool b) { return a <=> b; }", ft,  Int.Make(-1));

      //
      // str
      //
      Verify("static Bool f(Str a, Str b) { return a < b; }", MakeStrs("a", "b"),  Bool.True);
      Verify("static Bool f(Str a, Str b) { return a < b; }", MakeStrs("a", "a"),  Bool.False);
      Verify("static Bool f(Str a, Str b) { return a < b; }", MakeStrs("b", "a"),  Bool.False);

      Verify("static Bool f(Str a, Str b) { return a <= b; }", MakeStrs("a", "b"),  Bool.True);
      Verify("static Bool f(Str a, Str b) { return a <= b; }", MakeStrs("a", "a"),  Bool.True);
      Verify("static Bool f(Str a, Str b) { return a <= b; }", MakeStrs("b", "a"),  Bool.False);

      Verify("static Bool f(Str a, Str b) { return a > b; }", MakeStrs("a", "b"),  Bool.False);
      Verify("static Bool f(Str a, Str b) { return a > b; }", MakeStrs("a", "a"),  Bool.False);
      Verify("static Bool f(Str a, Str b) { return a > b; }", MakeStrs("b", "a"),  Bool.True);

      Verify("static Bool f(Str a, Str b) { return a >= b; }", MakeStrs("a", "b"),  Bool.False);
      Verify("static Bool f(Str a, Str b) { return a >= b; }", MakeStrs("a", "a"),  Bool.True);
      Verify("static Bool f(Str a, Str b) { return a >= b; }", MakeStrs("b", "a"),  Bool.True);

      Verify("static Int f(Str a, Str b) { return a <=> b; }", MakeStrs("a", "b"),  Int.Make(-1));
      Verify("static Int f(Str a, Str b) { return a <=> b; }", MakeStrs("a", "a"),  Int.Make(0));
      Verify("static Int f(Str a, Str b) { return a <=> b; }", MakeStrs("b", "a"),  Int.Make(1));

      Verify("static Int f(Str a, Str b) { return a <=> b; }", MakeStrs(null, null), Int.Make(0));
      Verify("static Int f(Str a, Str b) { return a <=> b; }", MakeStrs(null, "a"),  Int.Make(-1));
      Verify("static Int f(Str a, Str b) { return a <=> b; }", MakeStrs("b", null),  Int.Make(1));

      //
      // Duration
      //
      Verify("static Bool f(Duration a, Duration b) { return a < b; }", MakeDurs(3, 9),  Bool.True);
      Verify("static Bool f(Duration a, Duration b) { return a < b; }", MakeDurs(3, 3),  Bool.False);
      Verify("static Bool f(Duration a, Duration b) { return a < b; }", MakeDurs(9, 3),  Bool.False);

      Verify("static Bool f(Duration a, Duration b) { return a <= b; }", MakeDurs(3, 9),  Bool.True);
      Verify("static Bool f(Duration a, Duration b) { return a <= b; }", MakeDurs(3, 3),  Bool.True);
      Verify("static Bool f(Duration a, Duration b) { return a <= b; }", MakeDurs(9, 3),  Bool.False);

      Verify("static Bool f(Duration a, Duration b) { return a > b; }", MakeDurs(3, 9),  Bool.False);
      Verify("static Bool f(Duration a, Duration b) { return a > b; }", MakeDurs(3, 3),  Bool.False);
      Verify("static Bool f(Duration a, Duration b) { return a > b; }", MakeDurs(9, 3),  Bool.True);

      Verify("static Bool f(Duration a, Duration b) { return a >= b; }", MakeDurs(3, 9),  Bool.False);
      Verify("static Bool f(Duration a, Duration b) { return a >= b; }", MakeDurs(3, 3),  Bool.True);
      Verify("static Bool f(Duration a, Duration b) { return a >= b; }", MakeDurs(9, 3),  Bool.True);

      Verify("static Int f(Duration a, Duration b) { return a <=> b; }", MakeDurs(3, 9),  Int.Make(-1));
      Verify("static Int f(Duration a, Duration b) { return a <=> b; }", MakeDurs(3, 3),  Int.Make(0));
      Verify("static Int f(Duration a, Duration b) { return a <=> b; }", MakeDurs(9, 3),  Int.Make(1));

      //
      // auto-cast matrix
      //
      /* TODO
      verify("Bool f(Int a, Float b) { return a < b; }",  new Object[] { Int.make(-6), Float.make(-7)  }, Bool.False);
      verify("Bool f(Int a, Float b) { return a <= b; }", new Object[] { Int.make(-6), Float.make(-7)  }, Bool.False);
      verify("Bool f(Int a, Float b) { return a > b; }",  new Object[] { Int.make(-6), Float.make(-7)  }, Bool.True);
      verify("Bool f(Int a, Float b) { return a >= b; }", new Object[] { Int.make(-6), Float.make(-7)  }, Bool.True);

      verify("Bool f(Float a, Int b) { return a < b; }",  new Object[] { Float.make(99), Int.make(-99) }, Bool.False);
      verify("Bool f(Float a, Int b) { return a <= b; }", new Object[] { Float.make(99), Int.make(-99) }, Bool.False);
      verify("Bool f(Float a, Int b) { return a > b; }",  new Object[] { Float.make(99), Int.make(-99) }, Bool.True);
      verify("Bool f(Float a, Int b) { return a >= b; }", new Object[] { Float.make(99), Int.make(-99) }, Bool.True);

      verify("Int f(Float a, Int b) { return a <=> b; }", new Object[] { Float.make(9), Int.make(-9) }, Int.make(1));
      verify("Int f(Float a, Int b) { return a <=> b; }", new Object[] { Float.make(9), Int.make(9) },  Int.make(0));
      verify("Int f(Float a, Int b) { return a <=> b; }", new Object[] { Float.make(9), Int.make(99) }, Int.make(-1));
      verify("Int f(Float a, Int b) { return a <=> b; }", new Object[] { Float.make(9), Int.make(-9) }, Int.make(1));
      verify("Int f(Float a, Int b) { return a <=> b; }", new Object[] { Float.make(9), Int.make(9) },  Int.make(0));
      verify("Int f(Float a, Int b) { return a <=> b; }", new Object[] { Float.make(9), Int.make(99) }, Int.make(-1));
      */
    }

  //////////////////////////////////////////////////////////////////////////
  // Negate
  //////////////////////////////////////////////////////////////////////////

    void VerifyNegate()
    {
      // MakeInts
      Verify("static Int f(Int a) { return -a; }", MakeInts(0), Int.Make(0));
      Verify("static Int f(Int a) { return -a; }", MakeInts(1), Int.Make(-1));
      Verify("static Int f(Int a) { return -a; }", MakeInts(-1), Int.Make(1));
      Verify("static Int f(Int a) { return -(a); }", MakeInts(8), Int.Make(-8));
      Verify("static Int f(Int a) { return -(-a); }", MakeInts(8), Int.Make(8));

      // MakeFloats
      Verify("static Float f(Float a) { return -a; }", MakeFloats(1), Float.Make(-1));
      Verify("static Float f(Float a) { return -a; }", MakeFloats(-1), Float.Make(1));
      Verify("static Float f(Float a) { return -(a); }", MakeFloats(8), Float.Make(-8));
      Verify("static Float f(Float a) { return -(-a); }", MakeFloats(8), Float.Make(8));

      // MakeDursation
      Verify("static Duration f(Duration a) { return -a; }", MakeDurs(1), Duration.Make(-1));
      Verify("static Duration f(Duration a) { return -a; }", MakeDurs(-1), Duration.Make(1));
      Verify("static Duration f(Duration a) { return -(a); }", MakeDurs(8), Duration.Make(-8));
      Verify("static Duration f(Duration a) { return -(-a); }", MakeDurs(8), Duration.Make(8));
    }

  //////////////////////////////////////////////////////////////////////////
  // Multiply
  //////////////////////////////////////////////////////////////////////////

    void VerifyMultiply()
    {
      Object o;

      // MakeInts
      Verify("Int f(Int a, Int b) { return a * b; }", MakeInts(0, 0), Int.Make(0));
      Verify("Int f(Int a, Int b) { return a * b; }", MakeInts(0x3146443defL, 7), Int.Make(0x3146443defL*7));
      Verify("Int f(Int a, Int b) { return a * b; }", MakeInts(-3, 3), Int.Make(-3*3));
      Verify("Int f(Int a, Int b) { return a * 7; }", MakeInts(-3, 3), Int.Make(-3*7));
      Verify("Int f() { Int x := 3; x *= 6; return x; }", Int.Make(18));
      members = "Int x := 2; Int y;";
      Verify("Int f() { x *= 3; return x; }", Int.Make(6));
      Verify("Int f() { return x *= -3; }",   Int.Make(-6));
      o = Verify("Int f() { return y = x *= -3; }",   Int.Make(-6));
        Verify(Get(o, "x").Equals(Int.Make(-6)));
        Verify(Get(o, "y").Equals(Int.Make(-6)));
      members = "static Int x := 2;";
      Verify("static Int f() { return x *= 4; }", Int.Make(8));

      // MakeFloats
      Verify("Float f(Float a, Float b) { return a * b; }", MakeFloats(0, 0), Float.Make(0));
      Verify("Float f(Float a, Float b) { return a * b; }", MakeFloats(2, 5), Float.Make(10));
      Verify("Float f(Float a, Float b) { return a * b; }", MakeFloats(-3.32, 66.44), Float.Make(-3.32*66.44));
      Verify("Float f(Float a, Float b) { return a * 0.8; }", MakeFloats(-3.32, 66.44), Float.Make(-3.32*0.8));
      Verify("Float f() { x := 3.0; x *= 6.0; return x; }", Float.Make(18));
      members = "Float x := 2.0; Float y;";
      Verify("Float f() { x *= 3.0; return x; }", Float.Make(6));
      Verify("Float f() { return x *= -3.0; }",   Float.Make(-6));
      o = Verify("Float f() { return y = x *= -3f; }", Float.Make(-6));
        Verify(Get(o, "x").Equals(Float.Make(-6)));
        Verify(Get(o, "y").Equals(Float.Make(-6)));
      members = "static Float x := 0.0;";
      Verify("static Float f() { return x *= 4f; }", Float.Make(0));

      // MakeDursation
      members = "x := 2ns; \n Duration y;";
      Verify("Duration f(Duration a, Float b) { return a * b; }", new Object[] { Duration.Make(4), Float.Make(3) }, Duration.Make(12));
      Verify("Duration f(Duration a) { return a * 6.0; }", new Object[] { Duration.Make(4) }, Duration.Make(24));
      Verify("Duration f() { x *= 3.0; return x; }", Duration.Make(6));
      Verify("Duration f() { return x *= -3.0; }",   Duration.Make(-6));

      Verify("Duration f() { x := 7ns; x *= 2.0; return x; }", Duration.Make(14));
      o = Verify("Duration f() { return y = x *= -3f; }", Duration.Make(-6));
        Verify(Get(o, "x").Equals(Duration.Make(-6)));
        Verify(Get(o, "y").Equals(Duration.Make(-6)));
      members = "static Duration x := 1ms;";
      Verify("static Duration f() { return x *= 4f; }", Duration.Make(4000000));

      // auto-cast matrix
      /* TODO
      Verify("Float f(Int a, Float b) { return a * b; }", new Object[] { Int.Make(10), Float.Make(-0.004) }, Float.Make(10L*-0.004));
      Verify("Float f(Float a, Int b) { return a * b; }", new Object[] { Float.Make(1.2), Int.Make(0xabcdL) }, Float.Make(1.2*0xabcdL));
      */
    }

  //////////////////////////////////////////////////////////////////////////
  // Divide
  //////////////////////////////////////////////////////////////////////////

    void VerifyDivide()
    {
      // Ints
      Verify("Int f(Int a, Int b) { return a / b; }", MakeInts(0, 1), Int.Make(0));
      Verify("Int f(Int a, Int b) { return a / b; }", MakeInts(0x3146443defL, 7), Int.Make(0x3146443defL/7));
      Verify("Int f(Int a, Int b) { return a / b; }", MakeInts(-3, 3), Int.Make(-3/3));
      Verify("Int f(Int a, Int b) { return a / 7; }", MakeInts(-3, 3), Int.Make(-3/7));
      Verify("Int f() { Int x := 14; x /= 2; return x; }", Int.Make(7));
      members = "Int x := 8;";
      Verify("Int f() { x /= 4; return x; }", Int.Make(2));
      Verify("Int f() { return x /= 4; }",    Int.Make(2));

      // floats
      Verify("Float f(Float a, Float b) { return a / b; }", MakeFloats(0, 1), Float.Make(0));
      Verify("Float f(Float a, Float b) { return a / b; }", MakeFloats(20, 5), Float.Make(20d/5d));
      Verify("Float f(Float a, Float b) { return a / b; }", MakeFloats(-3.32, 66.44), Float.Make(-3.32/66.44));
      Verify("Float f(Float a, Float b) { return a / 0.8; }", MakeFloats(-3.32, 66.44), Float.Make(-3.32/0.8));
      Verify("Float f() { Float x := 14.0; x /= 2.0; return x; }", Float.Make(7));
      members = "Float x := 8F;";
      Verify("Float f() { x /= 4f; return x; }", Float.Make(2));
      Verify("Float f() { return x /= 4f; }",    Float.Make(2));

      // duration
      Verify("Duration f(Duration a, Float b) { return a / b; }", new Object[] { Duration.Make(100), Float.Make(4)}, Duration.Make(25));
      Verify("Duration f(Duration a) { return a / 4.0; }", new Object[] { Duration.Make(100), }, Duration.Make(25));
      Verify("Duration f() { x := 7ns; x /= 2.0; return x; }", Duration.Make(3));
      members = "Duration x := 8ns;";
      Verify("Duration f() { x /= 4.0; return x; }", Duration.Make(2));
      Verify("Duration f() { return x /= 4f; }",   Duration.Make(2));

      // auto-cast matrix
      /* TODO
      Verify("Float f(Int a, Float b) { return a / b; }", new Object[] { Int.Make(10), Float.Make(-0.004) }, Float.Make(10L/-0.004));
      Verify("Float f(Float a, Int b) { return a / b; }", new Object[] { Float.Make(1.2), Int.Make(0xabcdL) }, Float.Make(1.2/0xabcdL));
      */
    }

  //////////////////////////////////////////////////////////////////////////
  // Mod
  //////////////////////////////////////////////////////////////////////////

    void VerifyMod()
    {
      // MakeInts
      Verify("Int f(Int a, Int b) { return a % b; }", MakeInts(3, 2), Int.Make(1));
      Verify("Int f(Int a, Int b) { return a % b; }", MakeInts(0x3146443defL, 7), Int.Make(0x3146443defL%7));
      Verify("Int f(Int a, Int b) { return a % b; }", MakeInts(-3, 3), Int.Make(-3%3));
      Verify("Int f(Int a, Int b) { return a % 7; }", MakeInts(-5, 3), Int.Make(-5%7));
      Verify("Int f() { Int x := 15; x %= 3; return x; }", Int.Make(0));
      members = "Int x := 9;";
      Verify("Int f() { x %= 4; return x; }", Int.Make(1));
      Verify("Int f() { return x %= 4; }",    Int.Make(1));

      // MakeFloats
      Verify("Float f(Float a, Float b) { return a % b; }", MakeFloats(9, 4), Float.Make(1));
      Verify("Float f(Float a, Float b) { return a % b; }", MakeFloats(20, 5), Float.Make(20d%5d));
      Verify("Float f(Float a, Float b) { return a % b; }", MakeFloats(-3.32, 66.44), Float.Make(-3.32%66.44));
      Verify("Float f(Float a, Float b) { return a % 0.8; }", MakeFloats(-3.32, 66.44), Float.Make(-3.32%0.8));
      Verify("Float f() { Float x := 15.0; x %= 3.0; return x; }", Float.Make(0));
      members = "Float x := 9.0;";
      Verify("Float f() { x %= 4.0; return x; }", Float.Make(1));
      Verify("Float f() { return x %= 4.0; }",    Float.Make(1));

      // duration
      Verify("Duration f(Duration a, Float b) { return a % b; }", new Object[] { Duration.Make(13), Float.Make(4)}, Duration.Make(1));
      Verify("Duration f(Duration a) { return a % 4.0; }", new Object[] { Duration.Make(13), }, Duration.Make(1));
      Verify("Duration f() { x := 7ns; x %= 5.0; return x; }", Duration.Make(2));
      members = "Duration x := 10ns;";
      Verify("Duration f() { x %= 4.0; return x; }", Duration.Make(2));
      Verify("Duration f() { return x %= 4f; }",   Duration.Make(2));

      // auto-cast matrix
      /* TODO
      Verify("Float f(Int a, Float b) { return a % b; }", new Object[] { Int.Make(10), Float.Make(-0.004) }, Float.Make(10L%-0.004));
      Verify("Float f(Float a, Int b) { return a % b; }", new Object[] { Float.Make(1.2), Int.Make(0xabcdL) }, Float.Make(1.2%0xabcdL));
      */
    }

  //////////////////////////////////////////////////////////////////////////
  // Bitwise
  //////////////////////////////////////////////////////////////////////////

    void VerifyBitwise()
    {
      // and, or, xor, not
      long[] az = { 0, 0x1, 0x1, 0x1, 0xabcd, 0xabcdef987654321L };
      long[] bz = { 0, 0x1, 0x2, 0x3, 0x1234, 0x333333333333333L };

      for (int i=0; i<az.Length; i++)
      {
        long a = az[i];
        long b = bz[i];
        long c = ~a << 3;
        Verify("Int f(Int a) { return ~a; }", MakeInts(a), Int.Make(~a));
        Verify("Int f(Int b) { return ~b; }", MakeInts(b), Int.Make(~b));
        Verify("Int f(Int a, Int b) { return a & b; }", MakeInts(a, b), Int.Make(a & b));
        Verify("Int f(Int a, Int b) { return a | b; }", MakeInts(a, b), Int.Make(a | b));
        Verify("Int f(Int a, Int b) { return a ^ b; }", MakeInts(a, b), Int.Make(a ^ b));
        Verify("Int f(Int a, Int b, Int c) { return a & b & c; }", MakeInts(a, b, c), Int.Make(a & b & c));
        Verify("Int f(Int a, Int b, Int c) { return a | b | c; }", MakeInts(a, b, c), Int.Make(a | b | c));
        Verify("Int f(Int a, Int b, Int c) { return a ^ b ^ c; }", MakeInts(a, b, c), Int.Make(a ^ b ^ c));
        Verify("Int f(Int a, Int b, Int c) { return a & b | c; }", MakeInts(a, b, c), Int.Make(a & b | c));
        Verify("Int f(Int a, Int b, Int c) { return a | b & c; }", MakeInts(a, b, c), Int.Make(a | b & c));
        Verify("Int f(Int a, Int b, Int c) { return a & b ^ c; }", MakeInts(a, b, c), Int.Make(a & b ^ c));
        Verify("Int f(Int a, Int b, Int c) { return a ^ b & c; }", MakeInts(a, b, c), Int.Make(a ^ b & c));
        // note: Fan puts | and ^ at same precedence, so left to right
        Verify("Int f(Int a, Int b, Int c) { return a | b ^ c; }", MakeInts(a, b, c), Int.Make((a | b) ^ c));
        Verify("Int f(Int a, Int b, Int c) { return a ^ b | c; }", MakeInts(a, b, c), Int.Make((a ^ b) | c));
        // Verify equality lower precedence than bitwise (different than Java/C#)
        Verify("Bool f(Int a, Int b) { return a & b == 0 }", MakeInts(0x2, 0x4), Bool.True);
        Verify("Bool f(Int a, Int b) { return a & b == 2 }", MakeInts(0x2, 0x3), Bool.True);
        Verify("Bool f(Int a, Int b) { return a & b != 2 }", MakeInts(0x2, 0x3), Bool.False);
        Verify("Bool f(Int a, Int b) { return a | b == 3 }", MakeInts(0x2, 0x1), Bool.True);
        Verify("Bool f(Int a, Int b) { return a ^ b == 1 }", MakeInts(0x2, 0x3), Bool.True);
        // Verify comparision lower precedence than bitwise (different than Java/C#)
        Verify("Bool f(Int a, Int b) { return a & b >= 2 }", MakeInts(0x2, 0x3), Bool.True);
        Verify("Bool f(Int a, Int b) { return a | b < 2 }",  MakeInts(0x2, 0x1), Bool.False);
        Verify("Bool f(Int a, Int b) { return a ^ b <= 5 }", MakeInts(0x2, 0x3), Bool.True);
      }

      // shift
      for (int i=0; i<65; i++)
      {
        Verify("Int f(Int a, Int b) { return a << b; }", MakeInts(1L, i), Int.Make(1L<<i));
        Verify("Int f(Int a, Int b) { return a << b; }", MakeInts(0x123456789abcdef1L, i), Int.Make(0x123456789abcdef1L<<i));
//ULONG?        Verify("Int f(Int a, Int b) { return a >> b; }", MakeInts(0x8000000000000000L, i), Int.Make(0x8000000000000000L>>i));
        Verify("Int f(Int a, Int b) { return a >> b; }", MakeInts(0x7000000000000000L, i), Int.Make(0x7000000000000000L>>i));
        Verify("Int f(Int a, Int b) { return a >> b; }", MakeInts(0x7edcba987654321fL, i), Int.Make(0x7edcba987654321fL>>i));
      }
      Verify("Int f(Int a, Int b, Int c) { return a << b << c; }", MakeInts(1, 2, 3), Int.Make(1L<<2<<3));
      Verify("Int f(Int a, Int b, Int c) { return a >> b >> c; }", MakeInts(0x800000, 2, 3), Int.Make(0x800000>>2>>3));
      Verify("Int f(Int a, Int b, Int c) { return a - b - c; }", MakeInts(1, 2, 3), Int.Make(1-2-3));

      Object o;

      // local var assignment, pop
      Verify("Int f() { x := 0xf; x &= 0xa3; return x }", Int.Make(0x3));
      Verify("Int f() { x := 0xf; x |= 0xa3; return x }", Int.Make(0xaf));
      Verify("Int f() { x := 0xf; x ^= 0xa3; return x }", Int.Make(0xf ^ 0xa3));
      Verify("Int f() { x := 0xf; x <<= 3; return x }", Int.Make(0xf << 3));
      Verify("Int f() { x := 0xf; x >>= 2; return x }", Int.Make(0xf >> 2));

      // local var assignment, leave
      Verify("Int f() { x := 0xf; return x &= 0xa3 }", Int.Make(0x3));
      Verify("Int f() { x := 0xf; return x |= 0xa3 }", Int.Make(0xaf));
      Verify("Int f() { x := 0xf; return x ^= 0xa3 }", Int.Make(0xf ^ 0xa3));
      Verify("Int f() { x := 0xf; return x <<= 3 }", Int.Make(0xf << 3));
      Verify("Int f() { x := 0xf; return x >>= 2 }", Int.Make(0xf >> 2));

      // field var assignment, pop
      members = "Int x := 0xf;";
      o = Verify("Int f() { x &= 0xa3; return x }", Int.Make(0x3));
        Verify(Get(o, "x").Equals(Int.Make(0x3)));
      o = Verify("Int f() { x |= 0xa3; return x }", Int.Make(0xaf));
        Verify(Get(o, "x").Equals(Int.Make(0xaf)));
      o = Verify("Int f() { x ^= 0xa3; return x }", Int.Make(0xf ^ 0xa3));
        Verify(Get(o, "x").Equals(Int.Make(0xf ^ 0xa3)));
      o = Verify("Int f() { x <<= 3; return x }", Int.Make(0xf << 3));
        Verify(Get(o, "x").Equals(Int.Make(0xf << 3)));
      o = Verify("Int f() { x >>= 2; return x }", Int.Make(0xf >> 2));
        Verify(Get(o, "x").Equals(Int.Make(0xf >> 2)));

      // field var assignment, leave
      members = "Int x := 0xf;";
      o = Verify("Int f() { return x &= 0xa3; }", Int.Make(0x3));
        Verify(Get(o, "x").Equals(Int.Make(0x3)));
      o = Verify("Int f() { return x |= 0xa3; }", Int.Make(0xaf));
        Verify(Get(o, "x").Equals(Int.Make(0xaf)));
      o = Verify("Int f() { return x ^= 0xa3; }", Int.Make(0xf ^ 0xa3));
        Verify(Get(o, "x").Equals(Int.Make(0xf ^ 0xa3)));
      o = Verify("Int f() { return x <<= 3; }", Int.Make(0xf << 3));
        Verify(Get(o, "x").Equals(Int.Make(0xf << 3)));
      o = Verify("Int f() { return x >>= 2; }", Int.Make(0xf >> 2));
        Verify(Get(o, "x").Equals(Int.Make(0xf >> 2)));
    }

  //////////////////////////////////////////////////////////////////////////
  // Add
  //////////////////////////////////////////////////////////////////////////

    void VerifyAdd()
    {
      Object o;

      // MakeInts
      Verify("Int f(Int a, Int b) { return a + b; }", MakeInts(0, 1), Int.Make(1));
      Verify("Int f(Int a, Int b) { return a + b; }", MakeInts(0x3146443defL, 7), Int.Make(0x3146443defL+7));
      Verify("Int f(Int a, Int b) { return a + b; }", MakeInts(-3, 3), Int.Make(-3+3));
      Verify("Int f(Int a, Int b) { return a + 7; }", MakeInts(-3, 3), Int.Make(-3+7));
      Verify("Int f() { x := 5; x += 3; return x; }", Int.Make(8));
      members = "Int x := 5; Int y;";
      Verify("Int f() { x += -3; return x; }", Int.Make(2));
      o = Verify("Int f() { return y = x += -3; }", Int.Make(2));
        Verify(Get(o, "x").Equals(Int.Make(2)));
        Verify(Get(o, "y").Equals(Int.Make(2)));

      // MakeFloats
      Verify("Float f(Float a, Float b) { return a + b; }", MakeFloats(0, 1), Float.Make(1));
      Verify("Float f(Float a, Float b) { return a + b; }", MakeFloats(20, 5), Float.Make(20d+5d));
      Verify("Float f(Float a, Float b) { return a + b; }", MakeFloats(-3.32, 66.44), Float.Make(-3.32+66.44));
      Verify("Float f(Float a, Float b) { return a + 0.8; }", MakeFloats(-3.32, 66.44), Float.Make(-3.32+0.8));
      Verify("Float f() { x := 5.0; x+=5.0; return x; }", Float.Make(10));
      members = "Float x := 5.0; Float y;";
      Verify("Float f() { x += -3.0; return x; }", Float.Make(2));
      o = Verify("Float f() { return y = x += -3.0; }", Float.Make(2));
        Verify(Get(o, "x").Equals(Float.Make(2)));
        Verify(Get(o, "y").Equals(Float.Make(2)));

      // duration
      Verify("Duration f(Duration a, Duration b) { return a + b; }", MakeDurs(0, 1), Duration.Make(1));
      Verify("Duration f(Duration a, Duration b) { return a + b; }", MakeDurs(20, 5), Duration.Make(20+5));
      Verify("Duration f(Duration a, Duration b) { return a + b; }", MakeDurs(-3, 7), Duration.Make(4));
      Verify("Duration f(Duration a, Duration b) { return a + 6ns; }", MakeDurs(7, 99), Duration.Make(13));
      Verify("Duration f() { x := 3ns; x += 5ns; return x; }", Duration.Make(8));
      members = "Duration x := 5ns; Duration y;";
      Verify("Duration f() { x += -3ns; return x; }", Duration.Make(2));
      o = Verify("Duration f() { return y = x += -3ns; }", Duration.Make(2));
        Verify(Get(o, "x").Equals(Duration.Make(2)));
        Verify(Get(o, "y").Equals(Duration.Make(2)));

      // auto-cast matrix
      /* TODO
      Verify("Float f(Int a, Float b) { return a + b; }", new Object[] { Int.Make(10), Float.Make(-0.004) }, Float.Make(10L+-0.004));
      Verify("Float f(Float a, Int b) { return a + b; }", new Object[] { Float.Make(1.2), Int.Make(0xabcdL) }, Float.Make(1.2+0xabcdL));
      */
    }

  //////////////////////////////////////////////////////////////////////////
  // Sub
  //////////////////////////////////////////////////////////////////////////

    void VerifySub()
    {
      Object o;

      // MakeInts
      Verify("Int f(Int a, Int b) { return a - b; }", MakeInts(0, 1), Int.Make(-1));
      Verify("Int f(Int a, Int b) { return a - b; }", MakeInts(0x3146443defL, 7), Int.Make(0x3146443defL-7));
      Verify("Int f(Int a, Int b) { return a - b; }", MakeInts(-3, 3), Int.Make(-3-3));
      Verify("Int f(Int a, Int b) { return a - 7; }", MakeInts(-3, 3), Int.Make(-3-7));
      Verify("Int f(Int a, Int b) { return 7 - b; }", MakeInts(-3, 3), Int.Make(7-3));
      Verify("Int f() { x := 5; x -= 7; return x; }", Int.Make(-2));
      members = "Int x := 9;";
      Verify("Int f() { x -= 4; return x; }", Int.Make(5));
      Verify("Int f() { return x -= 4; }", Int.Make(5));

      // MakeFloats
      Verify("Float f(Float a, Float b) { return a - b; }", MakeFloats(0, 1), Float.Make(-1));
      Verify("Float f(Float a, Float b) { return a - b; }", MakeFloats(20, 5), Float.Make(20d-5d));
      Verify("Float f(Float a, Float b) { return a - b; }", MakeFloats(-3.32, 66.44), Float.Make(-3.32-66.44));
      Verify("Float f(Float a, Float b) { return a - 0.8; }", MakeFloats(-3.32, 66.44), Float.Make(-3.32-0.8));
      Verify("Float f() { x := 5.0; x -= 6.0; return x; }", Float.Make(-1));

      // duration
      Verify("Duration f(Duration a, Duration b) { return a - b; }", MakeDurs(4, 6), Duration.Make(-2));
      Verify("Duration f(Duration a, Duration b) { return a - b; }", MakeDurs(20, 5), Duration.Make(20-5));
      Verify("Duration f(Duration a, Duration b) { return a - b; }", MakeDurs(-3, 7), Duration.Make(-10));
      Verify("Duration f(Duration a, Duration b) { return a - 6ns; }", MakeDurs(8, 99), Duration.Make(2));
      Verify("Duration f() { x := 3ns; x -= 5ns; return x; }", Duration.Make(-2));
      members = "Duration x := 5ns; Duration y;";
      Verify("Duration f() { x -= 3ns; return x; }", Duration.Make(2));
      o = Verify("Duration f() { return y = x -= 3ns; }", Duration.Make(2));
        Verify(Get(o, "x").Equals(Duration.Make(2)));
        Verify(Get(o, "y").Equals(Duration.Make(2)));

      // auto-cast matrix
      /* TODO
      Verify("Float f(Int a, Float b) { return a - b; }", new Object[] { Int.Make(10), Float.Make(-0.004) }, Float.Make(10L-(-0.004)));
      members = "Float x := 9;";
      Verify("Float f() { x -= 4; return x; }", Float.Make(5));
      Verify("Float f() { return x -= 4; }", Float.Make(5));
      Verify("Float f(Float a, Int b) { return a - b; }", new Object[] { Float.Make(1.2), Int.Make(0xabcdL) }, Float.Make(1.2-0xabcdL));
      */
    }

  //////////////////////////////////////////////////////////////////////////
  // IncrementDecrement
  //////////////////////////////////////////////////////////////////////////

    void VerifyOrderAndPrecedence()
    {
      Verify("Int f(Int a, Int b, Int c) { return a - b - c; }", MakeInts(1, 2, 3), Int.Make(1-2-3));
      Verify("Int f(Int a, Int b, Int c, Int d) { return a - b - c - d; }", MakeInts(1, 2, 3, 4), Int.Make(1-2-3-4));
      Verify("Int f(Int a, Int b, Int c, Int d) { return a / b / c / d; }", MakeInts(40, 2, 3, 2), Int.Make(40/2/3/2));
      Verify("Int f(Int a, Int b, Int c, Int d) { return a * b / c % d; }", MakeInts(40, 2, 3, 2), Int.Make(40*2/3%2));
      Verify("Int f(Int a, Int b, Int c, Int d) { return a - b * c - d; }", MakeInts(1, 2, 3, 4), Int.Make(1-2*3-4));
      Verify("Int f(Int a, Int b, Int c, Int d) { return a - b / c - d; }", MakeInts(40, 10, 5, 3), Int.Make(40-10/5-3));
      Verify("Int f(Int a, Int b, Int c, Int d) { return a / b - c - d; }", MakeInts(40, 10, 5, 3), Int.Make(40/10-5-3));
      Verify("Int f(Int a, Int b, Int c, Int d) { return a / b - c * d; }", MakeInts(40, 10, 5, 3), Int.Make(40/10-5*3));
      Verify("Int f(Int a, Int b, Int c) { return a << b << c; }", MakeInts(1, 2, 3), Int.Make(1L<<2<<3));
    }

  //////////////////////////////////////////////////////////////////////////
  // IncrementDecrement
  //////////////////////////////////////////////////////////////////////////

    void VerifyIncrementDecrement()
    {
      Object o;
      imports = "";

      // int
      members = "Int x := 0;";
      o = Verify("Int f() { return ++x; }", Int.Make(1)); Verify(Get(o, "x").Equals(Int.Make(1)));
      o = Verify("Int f() { return x++; }", Int.Make(0)); Verify(Get(o, "x").Equals(Int.Make(1)));
      o = Verify("Int f() { return --x; }", Int.Make(-1)); Verify(Get(o, "x").Equals(Int.Make(-1)));
      o = Verify("Int f() { return x--; }", Int.Make(0)); Verify(Get(o, "x").Equals(Int.Make(-1)));
      members = "static Int x := 0;";
      o = Verify("static Int f() { return ++x; }", Int.Make(1)); Verify(Get(o, "x").Equals(Int.Make(1)));
      o = Verify("static Int f() { return x++; }", Int.Make(0)); Verify(Get(o, "x").Equals(Int.Make(1)));
      o = Verify("static Int f() { return --x; }", Int.Make(-1)); Verify(Get(o, "x").Equals(Int.Make(-1)));
      o = Verify("static Int f() { return x--; }", Int.Make(0)); Verify(Get(o, "x").Equals(Int.Make(-1)));
      members = "";
      o = Verify("Int f(Int y) { return ++y; }", MakeInts(3), Int.Make(4));
      o = Verify("Int f(Int y) { return y++; }", MakeInts(3), Int.Make(3));
      o = Verify("Int f(Int y) { return --y; }", MakeInts(3), Int.Make(2));
      o = Verify("Int f(Int y) { return y--; }", MakeInts(3), Int.Make(3));
      members = "Int x := 0;";
      o = Verify("Int f(Int y) { return x = ++y; }", MakeInts(3), Int.Make(4)); Verify(Get(o, "x").Equals(Int.Make(4)));
      o = Verify("Int f(Int y) { return x = y++; }", MakeInts(3), Int.Make(3)); Verify(Get(o, "x").Equals(Int.Make(3)));
      o = Verify("Int f(Int y) { return x = --y; }", MakeInts(3), Int.Make(2)); Verify(Get(o, "x").Equals(Int.Make(2)));
      o = Verify("Int f(Int y) { return x = y--; }", MakeInts(3), Int.Make(3)); Verify(Get(o, "x").Equals(Int.Make(3)));

      // MakeFloats
      members = "Float x := 0.0;";
      o = Verify("Float f() { return ++x; }", Float.Make(1)); Verify(Get(o, "x").Equals(Float.Make(1)));
      o = Verify("Float f() { return x++; }", Float.Make(0)); Verify(Get(o, "x").Equals(Float.Make(1)));
      o = Verify("Float f() { return --x; }", Float.Make(-1)); Verify(Get(o, "x").Equals(Float.Make(-1)));
      o = Verify("Float f() { return x--; }", Float.Make(0)); Verify(Get(o, "x").Equals(Float.Make(-1)));
      members = "static Float x := 0.0;";
      o = Verify("static Float f() { return ++x; }", Float.Make(1)); Verify(Get(o, "x").Equals(Float.Make(1)));
      o = Verify("static Float f() { return x++; }", Float.Make(0)); Verify(Get(o, "x").Equals(Float.Make(1)));
      o = Verify("static Float f() { return --x; }", Float.Make(-1)); Verify(Get(o, "x").Equals(Float.Make(-1)));
      o = Verify("static Float f() { return x--; }", Float.Make(0)); Verify(Get(o, "x").Equals(Float.Make(-1)));
      members = "";
      o = Verify("Float f(Float y) { return ++y; }", MakeFloats(3), Float.Make(4));
      o = Verify("Float f(Float y) { return y++; }", MakeFloats(3), Float.Make(3));
      o = Verify("Float f(Float y) { return --y; }", MakeFloats(3), Float.Make(2));
      o = Verify("Float f(Float y) { return y--; }", MakeFloats(3), Float.Make(3));
      members = "Float x := 0.0;";
      o = Verify("Float f(Float y) { return x = ++y; }", MakeFloats(3), Float.Make(4)); Verify(Get(o, "x").Equals(Float.Make(4)));
      o = Verify("Float f(Float y) { return x = y++; }", MakeFloats(3), Float.Make(3)); Verify(Get(o, "x").Equals(Float.Make(3)));
      o = Verify("Float f(Float y) { return x = --y; }", MakeFloats(3), Float.Make(2)); Verify(Get(o, "x").Equals(Float.Make(2)));
      o = Verify("Float f(Float y) { return x = y--; }", MakeFloats(3), Float.Make(3)); Verify(Get(o, "x").Equals(Float.Make(3)));
    }

  //////////////////////////////////////////////////////////////////////////
  // Plus Minus Combos (as Unary and Binary)
  //////////////////////////////////////////////////////////////////////////

    void VerifyPlusMinusCombos()
    {
      Verify("Int f(Int a, Int b) { return a-3;  }",     MakeInts(2, 3), Int.Make(-1));
      Verify("Int f(Int a, Int b) { return a - 3;  }",   MakeInts(2, 3), Int.Make(-1));
      Verify("Int f(Int a, Int b) { return a - -3;  }",  MakeInts(2, 3), Int.Make(5));
      Verify("Int f(Int a, Int b) { return a - - 3;  }", MakeInts(2, 3), Int.Make(5));
      Verify("Int f(Int a, Int b) { return a+3;  }",     MakeInts(2, 3), Int.Make(5));
      Verify("Int f(Int a, Int b) { return a + 3;  }",   MakeInts(2, 3), Int.Make(5));
      Verify("Int f(Int a, Int b) { return a + +3;  }",  MakeInts(2, 3), Int.Make(5));
      Verify("Int f(Int a, Int b) { return a + + 3;  }", MakeInts(2, 3), Int.Make(5));

      Verify("Int f(Int a, Int b) { return a-b;  }",     MakeInts(2, 3), Int.Make(-1));
      Verify("Int f(Int a, Int b) { return a - b;  }",   MakeInts(2, 3), Int.Make(-1));
      Verify("Int f(Int a, Int b) { return a - -b;  }",  MakeInts(2, 3), Int.Make(5));
      Verify("Int f(Int a, Int b) { return a - - b;  }", MakeInts(2, 3), Int.Make(5));
      Verify("Int f(Int a, Int b) { return a+b;  }",     MakeInts(2, 3), Int.Make(5));
      Verify("Int f(Int a, Int b) { return a + b;  }",   MakeInts(2, 3), Int.Make(5));
      Verify("Int f(Int a, Int b) { return a + +b;  }",  MakeInts(2, 3), Int.Make(5));
      Verify("Int f(Int a, Int b) { return a + + b;  }", MakeInts(2, 3), Int.Make(5));
    }

  //////////////////////////////////////////////////////////////////////////
  // StrAdd
  //////////////////////////////////////////////////////////////////////////

    void VerifyStrAdd()
    {
      // str
      Verify("Str f(Str a, Str b) { return a + b; }", MakeStrs("a", "b"), Str.Make("ab"));
      Verify("Str f(Str a, Str b) { return a + b; }", MakeStrs(null, null), Str.Make("nullnull"));
      Verify("Str f(Str a, Str b) { return a + b; }", MakeStrs(null, "b"),  Str.Make("nullb"));
      Verify("Str f(Str a, Str b) { return a + b; }", MakeStrs("a", null),  Str.Make("anull"));

      // bool
      Verify("Str f(Str a, Bool b) { return a + b; }", new Obj[] {Str.Make("a"), Bool.Make(true)}, Str.Make("atrue"));
      Verify("Str f(Bool a, Str b) { return a + b; }", new Obj[] {Bool.Make(false), Str.Make("a")}, Str.Make("falsea"));
      Verify("Str f() { return \"foo\" + true; }", Str.Make("footrue"));
      Verify("Str f() { return false + \" foo\"; }", Str.Make("false foo"));

      // int
      Verify("Str f(Str a, Int b) { return a + b; }", new Obj[] { Str.Make("a"), Int.Make(3) }, Str.Make("a3"));
      Verify("Str f(Int a, Str b) { return a + b; }", new Obj[] { Int.Make(-99), Str.Make("a") }, Str.Make("-99a"));
      Verify("Str f() { return \"foo \" + 77; }", Str.Make("foo 77"));
      Verify("Str f() { return 0 + \" foo\"; }", Str.Make("0 foo"));

      // double
      Verify("Str f(Str a, Float b) { return a + b; }", new Obj[] { Str.Make("a"), Float.Make(3) }, Str.Make("a3.0"));
      Verify("Str f(Float a, Str b) { return a + b; }", new Obj[] { Float.Make(-99), Str.Make("a") }, Str.Make("-99.0a"));
      Verify("Str f() { return \"foo \" + 77.0; }", Str.Make("foo 77.0"));
      Verify("Str f() { return 0.0 + \" foo\"; }", Str.Make("0.0 foo"));

      // mix
      Verify("Str f(Int a, Str b, Bool c) { return a + b + c; }", new Obj[] { Int.Make(3), Str.Make(" wow "), Bool.Make(true) }, Str.Make("3 wow true"));
      Verify("Str f(Int a, Str b, Bool c) { return \"w\" + a + \"x\" + b + \"y\" + c + \"z\"; }", new Object[] { Int.Make(3), Str.Make(" wow "), Bool.Make(true) }, Str.Make("w3x wow ytruez"));
    }

  //////////////////////////////////////////////////////////////////////////
  // LocalVar
  //////////////////////////////////////////////////////////////////////////

    void VerifyLocalVar()
    {
      imports = "";
      members = "";

      // this
//COMPILER      VerifyErr("static Obj f() { return this; }", "Cannot access 'this' in static context");

      // local definitions
      Verify("Int f() { Int x; x = 7; return x; }", Int.Make(7));
      /* TODO is this allowed?
      Verify("Int f() { sys::Int x; x = 7; return x; }", Int.Make(7));
      */
      Verify("Int f(Int a) { Int x := 7; return x; }", MakeInts(1), Int.Make(7));
      Verify("Str f() { Str x := \"hello\"; return x; }", Str.Make("hello"));
      Verify("Int f() { Str a := \"hello\"; Int b := 66; return b; }", Int.Make(66));

      Verify("static Bool f(Bool x, Bool y) { x = true; y = x; return y; }", ff, Bool.True);
      Verify("static Bool f(Bool x, Bool y) { return x = y = true; }", ff, Bool.True);

      Verify("static Int f(Int x) { x = 17; return x; }", MakeInts(88), Int.Make(17));
      Verify("static Int f(Int x) { return  x = 17; }", MakeInts(88), Int.Make(17));
      Verify("static Str f(Str x) { x = \"yeah\"; return x; }", MakeStrs("not"), Str.Make("yeah"));
      Verify("static Str f(Str x) { return x = \"yeah\"; }", MakeStrs("not"), Str.Make("yeah"));

      Verify("Bool f(Bool x, Bool y) { x = true; y = x; return y; }", ff, Bool.True);
      Verify("Int f(Int x) { x = 17; return x; }",     MakeInts(88), Int.Make(17));
      Verify("Int f(Int x) { return x = 17; }",        MakeInts(88), Int.Make(17));
      Verify("Float f(Float x) { x = 17.0; return x; }", MakeFloats(88), Float.Make(17));
      Verify("Float f(Float x) { return x = 17.0; }",    MakeFloats(88), Float.Make(17));
      Verify("Str f(Str x) { x = \"yeah\"; return x; }", MakeStrs("not"), Str.Make("yeah"));
      Verify("Obj f(Obj x) { x = \"yeah\"; return x; }", MakeStrs("not"), Str.Make("yeah"));
//COMPILER      VerifyErr("Int f(Int x) { x = \"no way\"; return x; }", "Type 'sys::Str' is not assignable to 'sys::Int'");

      // auto-cast matrix: int * x
//COMPILER      VerifyErr("Int f(Int a, Float b) { a = b; return a; }", "Type 'sys::Float' is not assignable to 'sys::Int'");
//COMPILER      VerifyErr("Int f(Int a, Float b) { a = 6.0; return a; }", "Type 'sys::Float' is not assignable to 'sys::Int'");

      // auto-cast matrix: Float * x
      /*
      Verify("Float f(Float a, Int b) { a = b; return a; }", new Object[] { Float.Make(0), Int.Make(6) }, Float.Make(6));
      Verify("Float f(Float a, Int b) { a = 6; return a; }", new Object[] { Float.Make(0), Int.Make(6) }, Float.Make(6));
      Verify("Float f(Float a, Int b) { return a = b; }", new Object[] { Float.Make(0), Int.Make(6) }, Float.Make(6));
      */
    }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    void VerifyFieldGet()
    {
      imports = "";

      members = "static x := 0;";
      Verify("static Int f() { return x; }", Int.Make(0));

      members = "static Int x";
      Verify("Int f() { return x; }", null);

      members = "Int x := 6;";
      Verify("Int f() { return x; }", Int.Make(6));

      members = "x := 4;";
      Verify("Int f() { return this.x; }", Int.Make(4));

      members = "Int x;";
      //VerifyErr("static Int f() { return x; }", "Cannot access instance field '" + nextPodName + "::" + nextClassName + ".x' in static context");
//COMPILER      VerifyErr("static Int f() { return this.x }", "Cannot access 'this' in static context");

      members = "Float x";
      Verify("Float f() { return x; }", null);

      members = "x := 0.0";
      Verify("Float f() { return x; }", Float.Make(0));
    }

    void VerifyFieldSet()
    {
      Object o;

      imports = "";
      members = "static Int x; static Int y;";
      Verify("static Int f() { x = 919; return x; }", Int.Make(919));
      Verify("static Int f() { return x = 919; }",    Int.Make(919));
      o = Verify("static Int f() { return x = y = 7; }", Int.Make(7));
        Verify(Get(o, "x").Equals(Int.Make(7)));
        Verify(Get(o, "y").Equals(Int.Make(7)));

      imports = "";
      members = "Int x; Int y;";
      Verify("Int f() { x = 69; return x; }", Int.Make(69));

      Verify("Int f() { return x = 69; }", Int.Make(69));
      o = Verify("Int f() { return x = y = 7; }", Int.Make(7));
        Verify(Get(o, "x").Equals(Int.Make(7)));
        Verify(Get(o, "y").Equals(Int.Make(7)));

      imports = "";
      members = "Int x; static Int y;";
      Verify("Int f() { this.x = 99; return this.x; }", Int.Make(99));
      Verify("Int f() { return this.x = 99; }", Int.Make(99));
      o = Verify("Int f() { return x = y = 7; }", Int.Make(7));
        Verify(Get(o, "x").Equals(Int.Make(7)));
        Verify(Get(o, "y").Equals(Int.Make(7)));

      imports = "";
      members = "Int x;";
      //VerifyErr("static Int f() { x = 5; return x; }", "Cannot access instance field '" + nextPodName + "::" + nextClassName + ".x' in static context");
//COMPILER      VerifyErr("static Void f() { this.x = 4; }", "Cannot access 'this' in static context");

      imports = "";
      members = "Float x; Float y;";
      Verify("Float f() { x = -4.88; return x; }", Float.Make(-4.88));
      Verify("Float f() { return x = -4.88; }", Float.Make(-4.88));
      Verify("Float f() { return x = 88.0; }", Float.Make(88));
      o = Verify("Float f() { return x = y = 7.0; }", Float.Make(7));
        Verify(Get(o, "x").Equals(Float.Make(7)));
        Verify(Get(o, "y").Equals(Float.Make(7)));
    }

  //////////////////////////////////////////////////////////////////////////
  // Calls
  //////////////////////////////////////////////////////////////////////////

    void VerifyCalls()
    {
      members = "static Int x() { return 0; }";
      Verify("static Int f() { return x(); }", Int.Make(0));
      Verify("static Int f() { return x }", Int.Make(0));

      members = "Int x() { return 77; }";
      Verify("Int f() { return x(); }", Int.Make(77));
      Verify("Int f() { return x }", Int.Make(77));

      members = "Int x() { return 6; }";
      Verify("Int f() { return this.x(); }", Int.Make(6));
      Verify("Int f() { return this.x; }", Int.Make(6));

      members = "static Str x() { return \"hello\"; }";
      Verify("Str f() { return x(); }", Str.Make("hello"));
      Verify("Int f() { return x.size }", Int.Make(5));
//COMPILER      VerifyErr("Int f() { return x.compare }", "Invalid args compare() for compare(sys::Obj)");

      Verify("Str f(Str q) { q = x(); return q; }", MakeStrs("foo"), Str.Make("hello"));
      Verify("Str f() { x(); return x(); }", Str.Make("hello"));
      Verify("Str f() { x; return x }", Str.Make("hello"));

//COMPILER      members = "Str x() { return \"hello\"; }";
//COMPILER      VerifyErr("static Str f() { return x(); }",         "Cannot call instance method '" + nextPodName + "::" + nextClassName + ".x' in static context");
//COMPILER      VerifyErr("static Str f() { return x }",            "Cannot call instance method '" + nextPodName + "::" + nextClassName + ".x' in static context");
//COMPILER      VerifyErr("static Str f() { return this.x() }",     "Cannot access 'this' in static context");
//COMPILER      VerifyErr("static Str f() { return this.x }",       "Cannot access 'this' in static context");
//COMPILER      VerifyErr("static Int f() { return Int.plus(3); }", "Cannot call instance method 'sys::Int.plus' in static context");
//COMPILER      VerifyErr("static Int f() { return Int.hash() }",   "Cannot call instance method 'sys::Int.hash' in static context");
//COMPILER      VerifyErr("static Int f() { return Int.hash}",      "Cannot call instance method 'sys::Int.hash' in static context");

      members = "Int zero() { return 0; }\n" +
                "Int two() { return 2; }";
      Verify("Bool f() { return zero() > two(); }",   Bool.False);
      Verify("Bool f() { return two()  > zero(); }",  Bool.True);
      Verify("Bool f() { return two()  == zero(); }", Bool.False);
      Verify("Bool f() { return two()  != zero(); }", Bool.True);
      Verify("Bool f() { return two >  zero }",  Bool.True);
      Verify("Bool f() { return two != zero }", Bool.True);
    }

  //////////////////////////////////////////////////////////////////////////
  // Parens
  //////////////////////////////////////////////////////////////////////////

    void VerifyParens()
    {
      members = "Int one() { return 1; }  Int two() { return 2; }";

      // literals wrapped in parens
      Verify("static Bool f() { return (false); }",     MakeBool(false));
      Verify("static Bool f() { return (true); }",      MakeBool(true));
      Verify("static Bool f() { return ((false)); }",   MakeBool(false));
      Verify("static Int f() { return (0); }",          Int.Make(0));
      Verify("static Int f() { return (77); }",         Int.Make(77));
      Verify("static Int f() { return (-5); }",         Int.Make(-5));
      Verify("static Int f() { return ((0)); }",        Int.Make(0));
      Verify("static Str f() { return (\"foo\"); }",    Str.Make("foo"));
      Verify("static Str f() { return (((\"foo\"))); }",Str.Make("foo"));
      Verify("static Bool f(Int a, Int b) { return (a > b); }",     MakeInts(2, 3),  MakeBool(false));
      Verify("static Bool f(Int a, Int b) { return (a) > (b); }",   MakeInts(7, -1), MakeBool(true));
      Verify("static Bool f(Int a, Int b) { return ((a) > (b)); }", MakeInts(7, -1), MakeBool(true));
      Verify("Bool f() { return (one() > two()); }",         MakeBool(false));
      Verify("Bool f() { return (one()) > (two()); }",       MakeBool(false));
      Verify("Bool f() { return ((one()) > (two())); }",     MakeBool(false));

      // TODO
    }

  //////////////////////////////////////////////////////////////////////////
  // Casts
  //////////////////////////////////////////////////////////////////////////

    void VerifyCasts()
    {
      Verify("static Str f(Obj x)   { return (Str)x; }", MakeStrs("hello"), Str.Make("hello"));
      Verify("static Int f(Obj x)   { return (Int)x; }", MakeInts(6), Int.Make(6));
      Verify("static Int f(Obj x)   { return (Int)5; }", MakeInts(6), Int.Make(5));
//GENERICS      Verify("static Str[] f(Obj x) { return (Str[])x; }", new Obj[]{new List(Sys.StrType)}, new List(Sys.StrType));

      /*
      if (jcompiler)
      {
        // various reference based casts
        Verify("Str f(Int x) { return (java.lang.Long.toHexString(x)); }", MakeInts(16), "10");
        Verify("Str f(Int x) { return (Str)java.lang.Long.toHexString(x); }", MakeInts(16), "10");
        Verify("Str f(Int x) { return (fan.sys.Str)java.lang.Long.toHexString(x); }", MakeInts(16), "10");
        Verify("Str f(Int x) { return (Str)(java.lang.Long.toHexString(x)); }", MakeInts(16), "10");
        Verify("Str f(Int x) { return (fan.sys.Str)(java.lang.Long.toHexString(x)); }", MakeInts(16), "10");
        Verify("Str f(Int x) { return ((Str)java.lang.Long.toHexString(x)); }", MakeInts(16), "10");
        Verify("Str f(Int x) { return ((fan.sys.Str)java.lang.Long.toHexString(x)); }", MakeInts(16), "10");
        Verify("Str f(Int x) { return ((Str)(java.lang.Long.toHexString(x))); }", MakeInts(16), "10");
        Verify("Str f(Int x) { return ((fan.sys.Str)(java.lang.Long.toHexString(x))); }", MakeInts(16), "10");

        // test primitive cast matrix

        // int to *
        Verify("Str f(int x) { return java.lang.String.valueOf(x); }",         MakeInts(65), "65");
        Verify("Str f(int x) { return java.lang.String.valueOf((char)x); }",   MakeInts(65), "A");
        Verify("Str f(int x) { return java.lang.Byte.toString((byte)x); }",    MakeInts(1000), String.valueOf((byte)1000));
        Verify("Str f(int x) { return java.lang.Short.toString((short)x); }",  MakeInts(0xFFFFF), String.valueOf((short)0xFFFFF));
        Verify("Str f(int x) { return java.lang.String.valueOf((int)x); }",    MakeInts(-6), "-6");
        Verify("Str f(int x) { return java.lang.String.valueOf((long)x); }",   MakeInts(5), "5");
        Verify("Str f(int x) { return java.lang.String.valueOf((float)x); }",  MakeInts(6), "6.0");
        Verify("Str f(int x) { return java.lang.String.valueOf((double)x); }", MakeInts(7), "7.0");

        // Int to *
        Verify("Str f(Int x) { return java.lang.String.valueOf(x); }",         MakeInts(65), "65");
        Verify("Str f(Int x) { return java.lang.String.valueOf((char)x); }",   MakeInts(65), "A");
        Verify("Str f(Int x) { return java.lang.Byte.toString((byte)x); }",    MakeInts(1000), String.valueOf((byte)1000));
        Verify("Str f(Int x) { return java.lang.Short.toString((short)x); }",  MakeInts(0xFFFFF), String.valueOf((short)0xFFFFF));
        Verify("Str f(Int x) { return java.lang.String.valueOf((int)x); }",    MakeInts(-6), "-6");
        Verify("Str f(Int x) { return java.lang.String.valueOf((long)x); }",   MakeInts(5), "5");
        Verify("Str f(Int x) { return java.lang.String.valueOf((float)x); }",  MakeInts(6), "6.0");
        Verify("Str f(Int x) { return java.lang.String.valueOf((double)x); }", MakeInts(7), "7.0");

        // float to *
        Verify("Str f(float x) { return java.lang.String.valueOf(x); }",         MakeFloats(65), "65.0");
        Verify("Str f(float x) { return java.lang.String.valueOf((char)x); }",   MakeFloats(65), "A");
        Verify("Str f(float x) { return java.lang.Byte.toString((byte)x); }",    MakeFloats(1000), String.valueOf((byte)1000));
        Verify("Str f(float x) { return java.lang.Short.toString((short)x); }",  MakeFloats(0xFFFFF), String.valueOf((short)0xFFFFF));
        Verify("Str f(float x) { return java.lang.String.valueOf((int)x); }",    MakeFloats(-6), "-6");
        Verify("Str f(float x) { return java.lang.String.valueOf((long)x); }",   MakeFloats(5), "5");
        Verify("Str f(float x) { return java.lang.String.valueOf((float)x); }",  MakeFloats(6), "6.0");
        Verify("Str f(float x) { return java.lang.String.valueOf((double)x); }", MakeFloats(7), "7.0");

        // Float to *
        Verify("Str f(Float x) { return java.lang.String.valueOf(x); }",         MakeFloats(65), "65.0");
        Verify("Str f(Float x) { return java.lang.String.valueOf((char)x); }",   MakeFloats(65), "A");
        Verify("Str f(Float x) { return java.lang.Byte.toString((byte)x); }",    MakeFloats(1000), String.valueOf((byte)1000));
        Verify("Str f(Float x) { return java.lang.Short.toString((short)x); }",  MakeFloats(0xFFFFF), String.valueOf((short)0xFFFFF));
        Verify("Str f(Float x) { return java.lang.String.valueOf((int)x); }",    MakeFloats(-6), "-6");
        Verify("Str f(Float x) { return java.lang.String.valueOf((long)x); }",   MakeFloats(5), "5");
        Verify("Str f(Float x) { return java.lang.String.valueOf((float)x); }",  MakeFloats(6), "6.0");
        Verify("Str f(Float x) { return java.lang.String.valueOf((double)x); }", MakeFloats(7), "7.0");

      }
        */
    }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    Bool[] tt = { Bool.True,  Bool.True  };
    Bool[] ft = { Bool.False, Bool.True  };
    Bool[] tf = { Bool.True,  Bool.False };
    Bool[] ff = { Bool.False, Bool.False };

    Bool[] nn = { null,       null };
    Bool[] tn = { Bool.True,  null};
    Bool[] nt = { null,       Bool.True};

    Bool[] ttt = { Bool.True,  Bool.True,  Bool.True  };
    Bool[] ftt = { Bool.False, Bool.True,  Bool.True  };
    Bool[] tft = { Bool.True,  Bool.False, Bool.True  };
    Bool[] ttf = { Bool.True,  Bool.True,  Bool.False };
    Bool[] fft = { Bool.False, Bool.False, Bool.True  };
    Bool[] ftf = { Bool.False, Bool.True,  Bool.False };
    Bool[] tff = { Bool.True,  Bool.False, Bool.False };
    Bool[] fff = { Bool.False, Bool.False, Bool.False };

  }
}