//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   23 Dec 06  Andy Frank  Creation
//

using System.Reflection;
using Fan.Sys;

namespace Fanx.Test
{
  /// <summary>
  /// IsAsTest.
  /// </summary>
  public class IsAsTest : CompileTest
  {

  //////////////////////////////////////////////////////////////////////////
  // Main
  //////////////////////////////////////////////////////////////////////////

    public override void Run()
    {
      VerifyIs();
      VerifyAs();
      // TODO
      //VerifyFits();
      //VerifyList();
    }

  //////////////////////////////////////////////////////////////////////////
  // Is
  //////////////////////////////////////////////////////////////////////////

    void VerifyIs()
    {
      Verify("Bool f() { return true is Bool }",  null, Bool.True);
      //Verify("Bool f() { return 5 is Bool }",  null, Bool.False);
      Verify("Bool f() { return type is Type}",  null, Bool.True);
      //Verify("Bool f() { return type.name is Type}",  null, Bool.False);
      Verify("Bool f() { return type.name is Str}",  null, Bool.True);
      Verify("Bool f() { return type.name.size is Int}",  null, Bool.True);
      Verify("Bool f(Obj o) { return o is Int}",  MakeBools(false), Bool.False);
      Verify("Bool f(Obj o) { return o is Int}",  MakeInts(7), Bool.True);
    }

  //////////////////////////////////////////////////////////////////////////
  // As
  //////////////////////////////////////////////////////////////////////////

    void VerifyAs()
    {
      Verify("Str f(Obj x) { return x as Str }", MakeStrs("foo"), Str.Make("foo"));
      Verify("Str f(Obj x) { return x as Str }", MakeInts(4), null);
      Verify("Str f(Obj x) { return x as Str }", new Obj[] { null }, null);
      Verify("Str f(Obj x) { return x.type.method(\"toStr\").call1(x) as Str }", MakeInts(2), Str.Make("2"));
      Verify("Int f(Obj x) { return x.type.method(\"toStr\").call1(x) as Int }", MakeInts(2), null);
    }

  //////////////////////////////////////////////////////////////////////////
  // Fits
  //////////////////////////////////////////////////////////////////////////

    /*
    void VerifyFits()
    {
      Verify("Bool f() { return true.type.fits(Bool.type) }",  null, Bool.True);
      Verify("Bool f() { return 5.type.fits(Bool.type) }",  null, Bool.False);
      Verify("Bool f() { return 5ms.type.fits(Obj.type) }",  null, Bool.True);
    }
    */

  //////////////////////////////////////////////////////////////////////////
  // List
  //////////////////////////////////////////////////////////////////////////

    /*
    void VerifyList()
    {
      // is Str[,]
      Verify("Bool f() { return Str[,] is Bool  }",   null, Bool.False);
      Verify("Bool f() { return Str[,] is List  }",   null, Bool.True);
      Verify("Bool f() { return Str[,] is Obj[] }",   null, Bool.True);
      Verify("Bool f() { return Str[,] is Str[] }",   null, Bool.True);
      Verify("Bool f() { return Str[,] is Int[] }",   null, Bool.False);
      Verify("Bool f() { return Str[,] is Str[][] }", null, Bool.False);

      // as Str[,]
      Obj x = new List(Sys.StrType);
      Object[] a = { x };
      Verify("Bool    f(Obj x) { return x as Bool  }",   a, null);
      Verify("List    f(Obj x) { return x as List  }",   a, x);
      Verify("Obj[]   f(Obj x) { return x as Obj[] }",   a, x);
      Verify("Str[]   f(Obj x) { return x as Str[] }",   a, x);
      Verify("Int[]   f(Obj x) { return x as Int[] }",   a, null);
      Verify("Str[][] f(Obj x) { return x as Str[][] }", a, null);

      // is [Str[,]]
      Verify("Bool f() { return [Str[,]] is Bool  }",      null, Bool.False);
      Verify("Bool f() { return [Str[,]] is List  }",      null, Bool.True);
      Verify("Bool f() { return [Str[,]] is List[]  }",    null, Bool.True);
      Verify("Bool f() { return [Str[,]] is Str[]  }",     null, Bool.False);
      Verify("Bool f() { return [Str[,]] is Str[][]  }",   null, Bool.True);
      Verify("Bool f() { return [Str[,]] is Obj[][]  }",   null, Bool.True);
      Verify("Bool f() { return [Str[,]] is Int[][]  }",   null, Bool.False);
      Verify("Bool f() { return [Str[,]] is Str[][][] }",  null, Bool.False);

      // as [Str[,]]
      x = new List(Sys.StrType.toListOf());
      a = new Obj[] { x };
      Verify("Bool      f(Obj x) { return x as Bool  }",      a, null);
      Verify("List      f(Obj x) { return x as List  }",      a, x);
      Verify("List[]    f(Obj x) { return x as List[]  }",    a, x);
      Verify("Str[]     f(Obj x) { return x as Str[]  }",     a, null);
      Verify("Str[][]   f(Obj x) { return x as Str[][]  }",   a, x);
      Verify("Obj[][]   f(Obj x) { return x as Obj[][]  }",   a, x);
      Verify("Int[][]   f(Obj x) { return x as Int[][]  }",   a, null);
      Verify("Str[][][] f(Obj x) { return x as Str[][][] }",  a, null);
    }
    */

  }
}