//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   26 Dec 06  Andy Frank  Creation
//

using Fan.Sys;

namespace Fanx.Test
{
  /// <summary>
  /// MethodTest.
  /// </summary>
  public class MethodTest : CompileTest
  {

  //////////////////////////////////////////////////////////////////////////
  // Main
  //////////////////////////////////////////////////////////////////////////

    public override void Run()
    {
      VerifyTyping();
      //VerifyCall();
    }

  //////////////////////////////////////////////////////////////////////////
  // Typing
  //////////////////////////////////////////////////////////////////////////

    public void VerifyTyping()
    {
      VerifySig("|,|", new Type[] { }, Sys.VoidType);

      VerifySig("|->Str|", new Type[] { }, Sys.StrType);

      VerifySig("|->Void|", new Type[] { }, Sys.VoidType);

      VerifySig("|Bool x-> Str|", new Type[] { Sys.BoolType }, Sys.StrType);

      VerifySig("|Str a, Duration b -> Int|",
                new Type[] { Sys.StrType, Sys.DurationType },
                Sys.IntType);

      VerifySig("|Bool a, Int b, Int c, Str d, Type e, Obj f, Str g, Bool h -> Int|",
               new Type[] { Sys.BoolType, Sys.IntType, Sys.IntType, Sys.StrType,
                 Sys.TypeType, Sys.ObjType, Sys.StrType, Sys.BoolType  },
               Sys.IntType);

      VerifySig("|Bool a, Int b, Int c, Str d, Type e, Obj f, Str g, Bool h, Duration i -> Int|",
               new Type[] { Sys.BoolType, Sys.IntType, Sys.IntType, Sys.StrType,
                 Sys.TypeType, Sys.ObjType, Sys.StrType, Sys.BoolType, Sys.DurationType  },
               Sys.IntType);

      //VerifySigErr("|Bool a-> Int,Bool|", "Expected '|', not ','");
    }

    /*
    void VerifySigErr(string sig, string msg)
    {
      VerifyErr("static Obj f(" + sig + " m) { return null }", msg);
    }
    */

    void VerifySig(string sig, Type[] p, Type r)
    {
      // test as method param signature
      Type t = CompileToFanType("class Foo { static Obj f(" + sig + " m) { return null } }");
      Method m = t.Method("f", true);
      MethodParam pars = (MethodParam)m.Params().Get(0);
      VerifySig(pars.ParamType(), p, r);

      // test as type literal
      //t = CompileToFanType("class Foo { static Type f() { return " + sig + ".type } }");
      //m = t.Method("f", true);
      //VerifySig((Type)m.Call0(), p, r);

      // test as closure signature
      System.Type cls;
      if (r.IsVoid())
        cls = CompileToType("class Foo { static Method f() { return " + sig + " { return; } }}");
      else
        cls = CompileToType("class Foo { static Method f() { return " + sig + " { return null; } }}");
      m = (Method)InvokeStatic(cls, "F");
      VerifySig(m.Type(), p, r);
    }

    void VerifySig(Type t, Type[] p, Type r)
    {
      for (int i=0; i<=Method.MaxIndirectParams; i++)
      {
        Method c = t.Method("call"+i, true);

        // Verify return
        Verify(c.ReturnType() == r);

        // Verify p0..pn params
        for (int j=0; j<p.Length && j<i; j++)
        {
          Verify(((MethodParam)c.Params().Get(j)).ParamType() == p[j]);
        }

        // Verify rest left at Obj
        for (int j=p.Length; j<Method.MaxIndirectParams && j<i; j++)
        {
          Verify(((MethodParam)c.Params().Get(j)).ParamType() == Sys.ObjType);
        }
      }
    }

  //////////////////////////////////////////////////////////////////////////
  // Call
  //////////////////////////////////////////////////////////////////////////

    /*
    public void VerifyCall()
      throws Exception
    {
      // static
      Verify("Obj f(|Str s, Bool c -> Type| m) { return m.call([\"sys::Str\", true]) }",
        new Obj[] { Sys.findMethod("sys::Sys.findType", true) }, Sys.StrType);
      Verify("Obj f(|Str s, Bool c -> Type| m) { return m.call([\"sys::Str\", true, 55]) }",
        new Obj[] { Sys.findMethod("sys::Sys.findType", true) }, Sys.StrType);

      // instance
      Verify("Obj f(|Int a, Int b -> Int| m) { return m.call([4, 2])}",
        new Obj[] { Sys.findMethod("sys::Int.plus", true) }, Int.make(6));
      Verify("Obj f(|Int a, Int b -> Int| m) { return m.call([4, 2])}",
        new Obj[] { Sys.findMethod("sys::Int.star", true) }, Int.make(8));
      Verify("Obj f(|Int a, Int b -> Int| m) { return m.call([4, 2, 3])}",
        new Obj[] { Sys.findMethod("sys::Int.star", true) }, Int.make(8));
      Verify("Obj f(|Int a, Int b -> Int| m) { return m.call([-3])}",
        new Obj[] { Sys.findMethod("sys::Int.negate", true) }, Int.make(3));
      Verify("Obj f(|Int a, Int b -> Int| m) { return m.call([4, 2])}",
        new Obj[] { Sys.findMethod("sys::Int.negate", true) }, Int.make(-4));
    }
    */

  }
}