//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   18 Dec 06  Andy Frank  Creation
//

using System.Reflection;
using Fan.Sys;

namespace Fanx.Test
{
  /// <summary>
  /// ParamTeset.
  /// </summary>
  public class ParamTest : CompileTest
  {

  //////////////////////////////////////////////////////////////////////////
  // Main
  //////////////////////////////////////////////////////////////////////////

    public override void Run()
    {
      VerifyStatic();
      VerifyInstance();
      VerifyCtor();
      VerifyStaticCall();
      VerifyInstanceCall();
      VerifyCtorCall();
    }

  //////////////////////////////////////////////////////////////////////////
  // Static
  //////////////////////////////////////////////////////////////////////////

    string _static  =
      "class Foo\n" +
      "{\n" +
      "  static Int f(Int a := 3) { return a }\n" +
      "  static Int g(Int a := 1, Int b := 2) { return a+b }\n" +
      "  static Int h(Int a,Int b:= 2,Int c:=3, Int d:=4, Int e:=5, Int f:=6) { return a+b+c+d+e+f }\n" +
      "}\n";

    public void VerifyStatic()
    {
      System.Type type = CompileToType(_static);
      object o;

      o = type.InvokeMember("F", GetStaticFlags(), null, null, new object[] {});
        Verify(o, Int.Make(3));
      o = type.InvokeMember("F", GetStaticFlags(), null, null, MakeInts(-2));
        Verify(o, Int.Make(-2));

      o = type.InvokeMember("G", GetStaticFlags(), null, null, new object[] {});
        Verify(o, Int.Make(3));
      o = type.InvokeMember("G", GetStaticFlags(), null, null, MakeInts(5));
        Verify(o, Int.Make(7));
      o = type.InvokeMember("G", GetStaticFlags(), null, null, MakeInts(-1, -2));
        Verify(o, Int.Make(-3));

      o = type.InvokeMember("H", GetStaticFlags(), null, null, MakeInts(0));
        Verify(o, Int.Make(20));
      o = type.InvokeMember("H", GetStaticFlags(), null, null, MakeInts(0, 0));
        Verify(o, Int.Make(18));
      o = type.InvokeMember("H", GetStaticFlags(), null, null, MakeInts(0, 0, 0));
        Verify(o, Int.Make(15));
      o = type.InvokeMember("H", GetStaticFlags(), null, null, MakeInts(0, 0, 0, 0));
        Verify(o, Int.Make(11));
      o = type.InvokeMember("H", GetStaticFlags(), null, null, MakeInts(0, 0, 0, 0, 0));
        Verify(o, Int.Make(6));
      o = type.InvokeMember("H", GetStaticFlags(), null, null, MakeInts(0, 0, 0, 0, 0, 0));
        Verify(o, Int.Make(0));
    }

  //////////////////////////////////////////////////////////////////////////
  // Instance
  //////////////////////////////////////////////////////////////////////////

    string instance =
      "class Foo\n" +
      "{\n" +
      "  Int f(Int a := 3) { return a }\n" +
      "  Int g(Int a := 1, Int b := 2) { return a+b }\n" +
      "  Int h(Int a,Int b:= 2,Int c:=3, Int d:=4, Int e:=5, Int f:=6) { return a+b+c+d+e+f }\n" +
      "}\n";

    public void VerifyInstance()
    {
      System.Type type = CompileToType(instance);
      object x = type.InvokeMember("Make", GetStaticFlags(), null, null, new object[0]);
      object o;

      Verify(InvokeInstance(type, x, "F"), Int.Make(3));
      Verify(InvokeInstance(type, x, "F", MakeInts(-2)), Int.Make(-2));

      Verify(InvokeInstance(type, x, "G"), Int.Make(3));
      Verify(InvokeInstance(type, x, "G", MakeInts(5)), Int.Make(7));
      Verify(InvokeInstance(type, x, "G", MakeInts(-1,-2)), Int.Make(-3));

      o = type.InvokeMember("H", GetInstanceFlags(), null, x, MakeInts(0));
        Verify(o, Int.Make(20));
      o = type.InvokeMember("H", GetInstanceFlags(), null, x, MakeInts(0,0));
        Verify(o, Int.Make(18));
      o = type.InvokeMember("H", GetInstanceFlags(), null, x, MakeInts(0,0,0));
        Verify(o, Int.Make(15));
      o = type.InvokeMember("H", GetInstanceFlags(), null, x, MakeInts(0,0,0,0));
        Verify(o, Int.Make(11));
      o = type.InvokeMember("H", GetInstanceFlags(), null, x, MakeInts(0,0,0,0,0));
        Verify(o, Int.Make(6));
      o = type.InvokeMember("H", GetInstanceFlags(), null, x, MakeInts(0,0,0,0,0,0));
        Verify(o, Int.Make(0));
    }

  //////////////////////////////////////////////////////////////////////////
  // Constructor
  //////////////////////////////////////////////////////////////////////////

    string ctor =
      "class Foo\n" +
      "{\n" +
      "  new f(Int a := 3) { x = a }\n" +
      "  new g(Int a := 1, Int b := 2) { x = a+b }\n" +
      "  new h(Int a,Int b:= 2,Int c:=3, Int d:=4, Int e:=5, Int f:=6) { x = a+b+c+d+e+f }\n" +
      "  Int x\n" +
      "}\n";

    public void VerifyCtor()
    {
      System.Type type = CompileToType(ctor);
      object o;

      o = type.InvokeMember("F", GetStaticFlags(), null, null, new object[] {});
        Verify(Get(o, "x"), Int.Make(3));
      o = type.InvokeMember("F", GetStaticFlags(), null, null, MakeInts(-2));
        Verify(Get(o, "x"), Int.Make(-2));

      o = type.InvokeMember("G", GetStaticFlags(), null, null, new object[] {});
        Verify(Get(o, "x"), Int.Make(3));
      o = type.InvokeMember("G", GetStaticFlags(), null, null, MakeInts(5));
        Verify(Get(o, "x"), Int.Make(7));
      o = type.InvokeMember("G", GetStaticFlags(), null, null, MakeInts(-1, -2));
        Verify(Get(o, "x"), Int.Make(-3));

      o = type.InvokeMember("H", GetStaticFlags(), null, null, MakeInts(0));
        Verify(Get(o, "x"), Int.Make(20));
      o = type.InvokeMember("H", GetStaticFlags(), null, null, MakeInts(0,0));
        Verify(Get(o, "x"), Int.Make(18));
      o = type.InvokeMember("H", GetStaticFlags(), null, null, MakeInts(0,0,0));
        Verify(Get(o, "x"), Int.Make(15));
      o = type.InvokeMember("H", GetStaticFlags(), null, null, MakeInts(0,0,0,0));
        Verify(Get(o, "x"), Int.Make(11));
      o = type.InvokeMember("H", GetStaticFlags(), null, null, MakeInts(0,0,0,0,0));
        Verify(Get(o, "x"), Int.Make(6));
      o = type.InvokeMember("H", GetStaticFlags(), null, null, MakeInts(0,0,0,0,0,0));
        Verify(Get(o, "x"), Int.Make(0));
    }

  //////////////////////////////////////////////////////////////////////////
  // Static Call
  //////////////////////////////////////////////////////////////////////////

    string staticCall  =
      "class Foo\n" +
      "{\n" +
      "  static Int f(Int a := 3) { return a }\n" +
      "  static Int c0() { return f() }\n" +
      "  static Int c1(Int a) { return f(a) }\n" +
      "}\n";

    public void VerifyStaticCall()
    {
      System.Type type = CompileToType(staticCall);
      object o;

      o = type.InvokeMember("C0", GetStaticFlags(), null, null, new object[] {});
        Verify(o, Int.Make(3));
      o = type.InvokeMember("C1", GetStaticFlags(), null, null, MakeInts(-2));
        Verify(o, Int.Make(-2));
    }

  //////////////////////////////////////////////////////////////////////////
  // Instance Call
  //////////////////////////////////////////////////////////////////////////

    string instanceCall  =
      "class Foo\n" +
      "{\n" +
      "  Int f(Int a, Int b := 3) { return a+b }\n" +
      "  Int c0(Int a) { return f(a) }\n" +
      "  Int c1(Int a, Int b) { return f(a, b) }\n" +
      "}\n";

    public void VerifyInstanceCall()
    {
      System.Type type = CompileToType(instanceCall);
      object x = type.InvokeMember("Make", GetStaticFlags(), null, null, new object[0]);
      object o;

      o = type.InvokeMember("C0", GetInstanceFlags(), null, x, MakeInts(-1));
        Verify(o, Int.Make(2));
      o = type.InvokeMember("C1", GetInstanceFlags(), null, x, MakeInts(6, 7));
        Verify(o, Int.Make(13));
    }

  //////////////////////////////////////////////////////////////////////////
  // Ctor Call
  //////////////////////////////////////////////////////////////////////////

    string ctorCall  =
      "class Foo\n" +
      "{\n" +
// TODO - Fix for case when make() comes after c0 and c1 in the code
      "  private new make(Int x := 0) { this.x = x }\n" +
      "  static Foo c0() { return make() }\n" +
      "  static Foo c1(Int x) { return make(x) }\n" +
      "  Int x\n" +
      "}\n";

    public void VerifyCtorCall()
    {
      System.Type type = CompileToType(ctorCall);
      object o;

      o = type.InvokeMember("C0", GetStaticFlags(), null, null, new object[] {});
        Verify(Get(o, "x"), Int.Make(0));
      o = type.InvokeMember("C1", GetStaticFlags(), null, null, MakeInts(77));
        Verify(Get(o, "x"), Int.Make(77));
    }

  //////////////////////////////////////////////////////////////////////////
  // TODO
  //////////////////////////////////////////////////////////////////////////

    // TODO - error checking such as: static Void f(Foo a := this)

  }
}
