//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   23 Dec 06  Andy Frank  Creation
//

using Fan.Sys;

namespace Fanx.Test
{
  /// <summary>
  /// ReflectTest.
  /// </summary>
  public class ReflectTest : CompileTest
  {

  //////////////////////////////////////////////////////////////////////////
  // Main
  //////////////////////////////////////////////////////////////////////////

    public override void Run()
    {
      VerifyTypeParser();
      VerifyBasic();
    }

  //////////////////////////////////////////////////////////////////////////
  // TypeParser
  //////////////////////////////////////////////////////////////////////////

    void VerifyTypeParser()
    {
      Verify(Sys.FindType("sys::Bool", true)     == Sys.BoolType);
      Verify(Sys.FindType("sys::Duration", true) == Sys.DurationType);
      //Verify(Sys.FindType("sys::Int[]", true)    == Sys.IntType.ToListOf());
      VerifyTypeParserErr("");
      VerifyTypeParserErr("x");
      VerifyTypeParserErr("xy");
      VerifyTypeParserErr("xyz");
      VerifyTypeParserErr("x:z");
      VerifyTypeParserErr("xz[");
      VerifyTypeParserErr("xz[]");
      VerifyTypeParserErr("[]");
    }

    void VerifyTypeParserErr(string sig)
    {
      try
      {
        Sys.FindType(sig, false);
        System.Console.WriteLine("didn't fail: " + sig);
        Fail();
      }
      catch(Err.Val err)
      {
        //System.out.println(" -- " + err);
        Verify(err.Err().Message().val.StartsWith("Invalid type signature '" + sig + "'"));
      }
    }

  //////////////////////////////////////////////////////////////////////////
  // Basic
  //////////////////////////////////////////////////////////////////////////

    public void VerifyBasic()
    {
      Type t = CompileToFanType("class Foo { Str f() { return type.name } }");
      Obj obj = (Obj)t.Make();
      Str name = (Str)t.Method("f", true).Call1(obj);
      Verify(name.val == "Foo");
    }

  }
}