//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   27 Nov 06  Andy Frank  Creation
//

using Fan.Sys;

namespace Fanx.Test
{
  /// <summary>
  /// TryTest
  /// </summary>
  public class TryTest : CompileTest
  {

  //////////////////////////////////////////////////////////////////////////
  // Main
  //////////////////////////////////////////////////////////////////////////

    public override void Run()
    {
      VerifyFinally();
      VerifyPeerFinally();
      VerifyCatch();
      VerifyCatchAll();
      VerifyCatchAndCatchAll();
      VerifyNested();
      VerifyTryFinally1();
      VerifyTryFinally2();
      VerifyTryFinally3();
      VerifyTryFinally4();
      VerifyTryFinally5();
      VerifyTryFinally6();
      VerifyTryFinally7();
      VerifyTryFinally8();
      VerifyTryCatchFinally1();
      VerifyTryCatchFinally2();
      VerifyTryCatchFinally3();
      VerifyTryCatchFinally4();
      //VerifyTryCatchFinally5();
      //FuckMe();
    }

  //////////////////////////////////////////////////////////////////////////
  // Finally
  //////////////////////////////////////////////////////////////////////////

    private void VerifyFinally()
    {
      string src =
        "class Foo\n" +
        "{\n" +
        "  static Int[] f(Int[] r, Int a)\n" +
        "  {\n" +
        "    r.add(0)\n" +
        "    try\n" +
        "    {\n" +
        "      r.add(a+1)\n" +
        "    }\n" +
        "    finally\n" +
        "    {\n" +
        "      r.add(99)\n" +
        "    }\n" +
        "    return r\n" +
        "  }\n" +
        "}";

      List r = new List(Sys.IntType, new Obj[] { Int.Make(0), Int.Make(3), Int.Make(99) });
      System.Type type = CompileToType(src);
      object o = type.InvokeMember("F", GetStaticFlags(), null, null, new object[] { r, Int.Make(2) });

      List c = (List)o;
      Verify(c.Get(0).Equals(Int.Make(0))  == Bool.True);
      Verify(c.Get(1).Equals(Int.Make(3))  == Bool.True);
      Verify(c.Get(2).Equals(Int.Make(99)) == Bool.True);
    }

  //////////////////////////////////////////////////////////////////////////
  // PeerFinally
  //////////////////////////////////////////////////////////////////////////

    private void VerifyPeerFinally()
    {
      string src =
        "class Foo\n" +
        "{\n" +
        "  static Int[] f(Int[] r, Int a)\n" +
        "  {\n" +
        "    r.add(0)\n" +
        "    try\n" +
        "    {\n" +
        "      r.add(a+1)\n" +
        "    }\n" +
        "    finally\n" +
        "    {\n" +
        "      r.add(99)\n" +
        "    }\n" +
        "    try\n" +
        "    {\n" +
        "      r.add(a+2)\n" +
        "    }\n" +
        "    finally\n" +
        "    {\n" +
        "      r.add(100)\n" +
        "    }\n" +
        "    return r\n" +
        "  }\n" +
        "}";

      List r = new List(Sys.IntType, new Obj[] {});
      System.Type type = CompileToType(src);
      object o = type.InvokeMember("F", GetStaticFlags(), null, null, new object[] { r, Int.Make(2) });

      Verify(r == o);
      Verify(r.Get(0).Equals(Int.Make(0))   == Bool.True);
      Verify(r.Get(1).Equals(Int.Make(3))   == Bool.True);
      Verify(r.Get(2).Equals(Int.Make(99))  == Bool.True);
      Verify(r.Get(3).Equals(Int.Make(4))   == Bool.True);
      Verify(r.Get(4).Equals(Int.Make(100)) == Bool.True);
    }

  //////////////////////////////////////////////////////////////////////////
  // Catch
  //////////////////////////////////////////////////////////////////////////

    void VerifyCatch()
    {
      string src =
        "class Foo\n" +
        "{\n" +
        "  static Int f(Int a)\n" +
        "  {\n" +
        "    try\n" +
        "    {\n" +
        "      a += 1\n" +
        "      throw Err.make()\n" +
        "    }\n" +
        "    catch (Err err)\n" +
        "    {\n" +
        "      a += 2\n" +
        "    }\n" +
        "    return a\n" +
        "  }\n" +
        "}";

      System.Type type = CompileToType(src);
      object o = type.InvokeMember("F", GetStaticFlags(), null, null, new object[] { Int.Make(5) });
      Verify(o.Equals(Int.Make(8)));
    }

  //////////////////////////////////////////////////////////////////////////
  // CatchAll
  //////////////////////////////////////////////////////////////////////////

    void VerifyCatchAll()
    {
      string src =
        "class Foo\n" +
        "{\n" +
        "  static Int f(Int a)\n" +
        "  {\n" +
        "    try\n" +
        "    {\n" +
        "      a += 2\n" +
        "      throw Err.make()\n" +
        "    }\n" +
        "    catch\n" +
        "    {\n" +
        "      a += 1\n" +
        "    }\n" +
        "    return a\n" +
        "  }\n" +
        "}";

      System.Type type = CompileToType(src);
      object o = type.InvokeMember("F", GetStaticFlags(), null, null, new object[] { Int.Make(3) });
      Verify(o.Equals(Int.Make(6)));
    }

  //////////////////////////////////////////////////////////////////////////
  // CatchAndCatchAll
  //////////////////////////////////////////////////////////////////////////

    void VerifyCatchAndCatchAll()
    {
      string src =
        "class Foo\n" +
        "{\n" +
        "  static Int f(Int a, Err err)\n" +
        "  {\n" +
        "    try\n" +
        "    {\n" +
        "      a += 2\n" +
        "      throw err\n" +
        "    }\n" +
        "    catch (IOErr e)\n" +
        "    {\n" +
        "      a += 2\n" +
        "    }\n" +
        "    catch\n" +
        "    {\n" +
        "      a += 1\n" +
        "    }\n" +
        "    return a\n" +
        "  }\n" +
        "}";

      System.Type type = CompileToType(src);
      object o = type.InvokeMember("F", GetStaticFlags(), null, null, new object[] { Int.Make(3), IndexErr.Make() });
      Verify(o.Equals(Int.Make(6)));

      o = type.InvokeMember("F", GetStaticFlags(), null, null, new object[] { Int.Make(3), IOErr.Make() });
      Verify(o.Equals(Int.Make(7)));
    }

  //////////////////////////////////////////////////////////////////////////
  // Nested
  //////////////////////////////////////////////////////////////////////////

    void VerifyNested()
    {
      string src =
        "class Foo\n" +
        "{\n" +
        "  static Int f(Int a)\n" +
        "  {\n" +
        "    try\n" +
        "    {\n" +
        "      a = 1\n" +
        "      try\n" +
        "      {\n" +
        "        a = 2\n" +
        "      }\n" +
        "      catch\n" +
        "      {\n" +
        "        a = 3\n" +
        "      }\n" +
        "      a = 4\n" +
        "    }\n" +
        "    catch\n" +
        "    {\n" +
        "      a = 5\n" +
        "    }\n" +
        "    return a\n" +
        "  }\n" +
        "}";

      System.Type type = CompileToType(src);
      object o = type.InvokeMember("F", GetStaticFlags(), null, null, new object[] { Int.Make(3) });
      Verify(o.Equals(Int.Make(4)));
    }

  //////////////////////////////////////////////////////////////////////////
  // Try/Finally
  //////////////////////////////////////////////////////////////////////////

    void VerifyTryFinally1()
    {
      string src = "class Foo\n"+
        "{\n"+
        "  static Int[] f(Int[] r, Int a)\n"+
        "  {\n"+
        "    r.add(0)\n"+
        "    try\n"+
        "    {\n"+
        "      r.add(a+1)\n"+
        "    }\n"+
        "    finally\n"+
        "    {\n"+
        "      r.add(99)\n"+
        "    }\n"+
        "    return r\n"+
        "  }\n"+
        "}";

      System.Type type = CompileToType(src);
      List r = new List(Sys.IntType, new Obj[0]);
      object o = type.InvokeMember("F", GetStaticFlags(), null, null, new object[] { r, Int.Make(2) });

      List c = (List)o;
      Verify(c.Get(0).Equals(Int.Make(0))  == Bool.True);
      Verify(c.Get(1).Equals(Int.Make(3))  == Bool.True);
      Verify(c.Get(2).Equals(Int.Make(99)) == Bool.True);

      r.Clear();
      try
      {
        o = type.InvokeMember("F", GetStaticFlags(), null, null, new object[] { r, null });
      }
      catch (System.Reflection.TargetInvocationException ex)
      {
        Verify(ex.InnerException is System.NullReferenceException);
        c = (List)o;
        Verify(c.Get(0).Equals(Int.Make(0))  == Bool.True);
        Verify(c.Get(1).Equals(Int.Make(99)) == Bool.True);
      }
    }

    void VerifyTryFinally2()
    {
      string src = "class Foo\n" +
        "{\n" +
        "  static Int[] f(Int[] r, Bool b)\n" +
        "  {\n" +
        "    r.add(0)\n" +
        "    try\n" +
        "    {\n" +
        "      r.add(1)\n" +
        "      if (b) throw ArgErr.make\n" +
        "      r.add(2)\n" +
        "    }\n" +
        "    finally\n" +
        "    {\n" +
        "      r.add(3)\n" +
        "    }\n" +
        "    r.add(4)\n" +
        "    return r\n" +
        "  }\n" +
        "}";

      System.Type type = CompileToType(src);
      List r = new List(Sys.IntType, new Obj[0]);
      object o = type.InvokeMember("F", GetStaticFlags(), null, null, new object[] { r, Bool.False });

      List c = (List)o;
      Verify(c.Get(0).Equals(Int.Make(0)) == Bool.True);
      Verify(c.Get(1).Equals(Int.Make(1)) == Bool.True);
      Verify(c.Get(2).Equals(Int.Make(2)) == Bool.True);
      Verify(c.Get(3).Equals(Int.Make(3)) == Bool.True);
      Verify(c.Get(4).Equals(Int.Make(4)) == Bool.True);

      r.Clear();
      try
      {
        o = type.InvokeMember("F", GetStaticFlags(), null, null, new object[] { r, Bool.True});
      }
      catch (System.Reflection.TargetInvocationException ex)
      {
        Err err = (ex.InnerException as Err.Val).Err();
        Verify(err.GetType() == System.Type.GetType("Fan.Sys.ArgErr"));

        c = (List)o;
        Verify(c.Get(0).Equals(Int.Make(0)) == Bool.True);
        Verify(c.Get(1).Equals(Int.Make(1)) == Bool.True);
        Verify(c.Get(2).Equals(Int.Make(3)) == Bool.True);
      }
    }

    void VerifyTryFinally3()
    {
      string src = "class Foo\n" +
        "{\n" +
        "  static Int[] f(Int[] r)\n" +
        "  {\n" +
        "    r.add(0)\n" +
        "    try\n" +
        "    {\n" +
        "      r.add(1)\n" +
        "      return r\n" +
        "    }\n" +
        "    finally\n" +
        "    {\n" +
        "      r.add(2)\n" +
        "    }\n" +
        "  }\n" +
        "}";

      System.Type type = CompileToType(src);
      List r = new List(Sys.IntType, new Obj[0]);
      object o = type.InvokeMember("F", GetStaticFlags(), null, null, new object[] { r });

      Verify(r == o);

      List c = (List)o;
      Verify(c.Get(0).Equals(Int.Make(0)) == Bool.True);
      Verify(c.Get(1).Equals(Int.Make(1)) == Bool.True);
      Verify(c.Get(2).Equals(Int.Make(2)) == Bool.True);
    }

    void VerifyTryFinally4()
    {
      string src = "class Foo\n" +
        "{\n" +
        "  static Int[] f(Int[] r)\n" +
        "  {\n" +
        "    r.add(0)\n" +
        "    try\n" +
        "    {\n" +
        "      r.add(1)\n" +
        "      try\n" +
        "      {\n" +
        "        return r.add(2)\n" +
        "      }\n" +
        "      finally\n" +
        "      {\n" +
        "        r.add(3)\n" +
        "      }\n" +
        "      return r.add(4)\n" +
        "    }\n" +
        "    finally\n" +
        "    {\n" +
        "      r.add(5)\n" +
        "    }\n" +
        "    return r.add(6)\n" +
        "  }\n" +
        "}";

      System.Type type = CompileToType(src);
      List r = new List(Sys.IntType, new Obj[0]);
      object o = type.InvokeMember("F", GetStaticFlags(), null, null, new object[] { r });

      Verify(r == o);

      List c = (List)o;
      Verify(c.Get(0).Equals(Int.Make(0)) == Bool.True);
      Verify(c.Get(1).Equals(Int.Make(1)) == Bool.True);
      Verify(c.Get(2).Equals(Int.Make(2)) == Bool.True);
      Verify(c.Get(3).Equals(Int.Make(3)) == Bool.True);
      Verify(c.Get(4).Equals(Int.Make(5)) == Bool.True);
    }

    void VerifyTryFinally5()  // same as testTryFinally4 but Void
    {
      string src = "class Foo\n" +
        "{\n" +
        "  static Void f(Int[] r)\n" +
        "  {\n" +
        "    r.add(0)\n" +
        "    try\n" +
        "    {\n" +
        "      r.add(1)\n" +
        "      try\n" +
        "      {\n" +
        "        r.add(2)\n" +
        "        return\n" +
        "      }\n" +
        "      finally\n" +
        "      {\n" +
        "        r.add(3)\n" +
        "      }\n" +
        "      r.add(4)\n" +
        "      return\n" +
        "    }\n" +
        "    finally\n" +
        "    {\n" +
        "      r.add(5)\n" +
        "    }\n" +
        "    return\n" +
        "  }\n" +
        "}";

      System.Type type = CompileToType(src);
      List r = new List(Sys.IntType, new Obj[0]);
      object o = type.InvokeMember("F", GetStaticFlags(), null, null, new object[] { r });

      Verify(o == null);

      Verify(r.Get(0).Equals(Int.Make(0)) == Bool.True);
      Verify(r.Get(1).Equals(Int.Make(1)) == Bool.True);
      Verify(r.Get(2).Equals(Int.Make(2)) == Bool.True);
      Verify(r.Get(3).Equals(Int.Make(3)) == Bool.True);
      Verify(r.Get(4).Equals(Int.Make(5)) == Bool.True);
    }

    void VerifyTryFinally6()
    {
      string src = "class Foo\n" +
        "{\n" +
        "  static Int[] f(Int[] r)\n" +
        "  {\n" +
        "    r.add(0)\n" +
        "    for (i:=1; i<=3; ++i)\n" +
        "    {\n" +
        "      try\n" +
        "      {\n" +
        "        if (i == 3) throw ArgErr.make\n" +
        "        r.add(10+i)\n" +
        "      }\n" +
        "      finally\n" +
        "      {\n" +
        "        r.add(100+i)\n" +
        "      }\n" +
        "    }\n" +
        "    return r.add(1)\n" +
        "  }\n" +
        "}";

      System.Type type = CompileToType(src);
      List r = new List(Sys.IntType, new Obj[0]);
      try
      {
        type.InvokeMember("F", GetStaticFlags(), null, null, new object[] { r });
      }
      catch (System.Reflection.TargetInvocationException ex)
      {
        Err err = (ex.InnerException as Err.Val).Err();
        Verify(err.GetType() == System.Type.GetType("Fan.Sys.ArgErr"));

        Verify(r.Get(0).Equals(Int.Make(0))   == Bool.True);
        Verify(r.Get(1).Equals(Int.Make(11))  == Bool.True);
        Verify(r.Get(2).Equals(Int.Make(101)) == Bool.True);
        Verify(r.Get(3).Equals(Int.Make(12))  == Bool.True);
        Verify(r.Get(4).Equals(Int.Make(102)) == Bool.True);
        Verify(r.Get(5).Equals(Int.Make(103)) == Bool.True);
      }
    }

    void VerifyTryFinally7()
    {
      string src = "class Foo\n" +
        "{\n" +
        "  static Int[] f(Int[] r)\n" +
        "  {\n" +
        "    r.add(0)\n" +
        "    for (i:=1; true; ++i)\n" +
        "    {\n" +
        "      try\n" +
        "      {\n" +
        "        if (i % 2 == 0) continue\n" +
        "        if (i == 5) break\n" +
        "        r.add(i)\n" +
        "      }\n" +
        "      finally\n" +
        "      {\n" +
        "        r.add(100+i)\n" +
        "      }\n" +
        "    }\n" +
        "    return r.add(99)\n" +
        "  }\n" +
        "}";

      System.Type type = CompileToType(src);
      List r = new List(Sys.IntType, new Obj[0]);
      object o = type.InvokeMember("F", GetStaticFlags(), null, null, new object[] { r });

      Verify(r == o);

      Verify(r.Get(0).Equals(Int.Make(0))   == Bool.True);
      Verify(r.Get(1).Equals(Int.Make(1))   == Bool.True);
      Verify(r.Get(2).Equals(Int.Make(101)) == Bool.True);
      Verify(r.Get(3).Equals(Int.Make(102)) == Bool.True);
      Verify(r.Get(4).Equals(Int.Make(3))   == Bool.True);
      Verify(r.Get(5).Equals(Int.Make(103)) == Bool.True);
      Verify(r.Get(6).Equals(Int.Make(104)) == Bool.True);
      Verify(r.Get(7).Equals(Int.Make(105)) == Bool.True);
      Verify(r.Get(8).Equals(Int.Make(99))  == Bool.True);
    }

    void VerifyTryFinally8()
    {
      string src = "class Foo\n" +
        "{\n" +
        "  static Void f(Int[] r)\n" +
        "  {\n" +
        "    try\n" +
        "    {\n" +
        "      try\n" +
        "      {\n" +
        "        r.add(0)\n" +
        "        for (i:=1; true; ++i)\n" +
        "        {\n" +
        "          try\n" +
        "          {\n" +
        "            try\n" +
        "            {\n" +
        "              if (i % 2 == 0) continue\n" +
        "              if (i == 5) break\n" +
        "              r.add(i)\n" +
        "            }\n" +
        "            finally\n" +
        "            {\n" +
        "              r.add(100+i)\n" +
        "            }\n" +
        "          }\n" +
        "          finally\n" +
        "          {\n" +
        "            r.add(99)\n" +
        "          }\n" +
        "        }\n" +
        "        r.add(999)\n" +
        "      }\n" +
        "      finally\n" +
        "      {\n" +
        "        r.add(9999)\n" +
        "      }\n" +
        "    }\n" +
        "    finally\n" +
        "    {\n" +
        "      r.add(99999)\n" +
        "    }\n" +
        "  }\n" +
        "}";

      //r := Int[,]
      //verifySame(t.method("f").call1(r), null)
      //verifyEq(r, [0, 1, 101, 99, 102, 99, 3, 103, 99, 104, 99, 105, 99, 999, 9999, 99999])

      System.Type type = CompileToType(src);
      List r = new List(Sys.IntType, new Obj[0]);
      object o = type.InvokeMember("F", GetStaticFlags(), null, null, new object[] { r });

      Verify(o == null);

      Verify(r.Get(0).Equals(Int.Make(0))      == Bool.True);
      Verify(r.Get(1).Equals(Int.Make(1))      == Bool.True);
      Verify(r.Get(2).Equals(Int.Make(101))    == Bool.True);
      Verify(r.Get(3).Equals(Int.Make(99))     == Bool.True);
      Verify(r.Get(4).Equals(Int.Make(102))    == Bool.True);
      Verify(r.Get(5).Equals(Int.Make(99))     == Bool.True);
      Verify(r.Get(6).Equals(Int.Make(3))      == Bool.True);
      Verify(r.Get(7).Equals(Int.Make(103))    == Bool.True);
      Verify(r.Get(8).Equals(Int.Make(99))     == Bool.True);
      Verify(r.Get(9).Equals(Int.Make(104))    == Bool.True);
      Verify(r.Get(10).Equals(Int.Make(99))    == Bool.True);
      Verify(r.Get(11).Equals(Int.Make(105))   == Bool.True);
      Verify(r.Get(12).Equals(Int.Make(99))    == Bool.True);
      Verify(r.Get(13).Equals(Int.Make(999))   == Bool.True);
      Verify(r.Get(14).Equals(Int.Make(9999))  == Bool.True);
      Verify(r.Get(15).Equals(Int.Make(99999)) == Bool.True);
    }

  //////////////////////////////////////////////////////////////////////////
  // Try/Catch/Finally
  //////////////////////////////////////////////////////////////////////////

    void VerifyTryCatchFinally1()
    {
      string src = "class Foo\n" +
        "{\n" +
        "  static Int f(Int[] r, Bool raise)\n" +
        "  {\n" +
        "    r.add(0)\n" +
        "    try\n" +
        "    {\n" +
        "      r.add(1)\n" +
        "      if (raise) throw ArgErr.make\n" +
        "      r.add(2)\n" +
        "      return 2\n" +
        "    }\n" +
        "    catch\n" +
        "    {\n" +
        "      r.add(3)\n" +
        "      return 3\n" +
        "    }\n" +
        "    finally\n" +
        "    {\n" +
        "      r.add(4)\n" +
        "    }\n" +
        "  }\n" +
        "}";

      System.Type type = CompileToType(src);
      List r = new List(Sys.IntType, new Obj[0]);
      object o = type.InvokeMember("F", GetStaticFlags(), null, null, new object[] { r, Bool.False });

      Verify(o == Int.Make(2));
      Verify(r.Get(0).Equals(Int.Make(0)) == Bool.True);
      Verify(r.Get(1).Equals(Int.Make(1)) == Bool.True);
      Verify(r.Get(2).Equals(Int.Make(2)) == Bool.True);
      Verify(r.Get(3).Equals(Int.Make(4)) == Bool.True);

      r.Clear();
      o = type.InvokeMember("F", GetStaticFlags(), null, null, new object[] { r, Bool.True});
      Verify(o == Int.Make(3));
      Verify(r.Get(0).Equals(Int.Make(0)) == Bool.True);
      Verify(r.Get(1).Equals(Int.Make(1)) == Bool.True);
      Verify(r.Get(2).Equals(Int.Make(3)) == Bool.True);
      Verify(r.Get(3).Equals(Int.Make(4)) == Bool.True);
    }

    void VerifyTryCatchFinally2() // same as testTryCatchFinally2 but Void
    {
      string src = "class Foo\n" +
        "{\n" +
        "  static Void f(Int[] r, Bool raise)\n" +
        "  {\n" +
        "    r.add(0)\n" +
        "    try\n" +
        "    {\n" +
        "      r.add(1)\n" +
        "      if (raise) throw ArgErr.make\n" +
        "      r.add(2)\n" +
        "    }\n" +
        "    catch\n" +
        "    {\n" +
        "      r.add(3)\n" +
        "    }\n" +
        "    finally\n" +
        "    {\n" +
        "      r.add(4)\n" +
        "    }\n" +
        "  }\n" +
        "}";

      System.Type type = CompileToType(src);
      List r = new List(Sys.IntType, new Obj[0]);
      object o = type.InvokeMember("F", GetStaticFlags(), null, null, new object[] { r, Bool.False });

      Verify(o == null);
      Verify(r.Get(0).Equals(Int.Make(0)) == Bool.True);
      Verify(r.Get(1).Equals(Int.Make(1)) == Bool.True);
      Verify(r.Get(2).Equals(Int.Make(2)) == Bool.True);
      Verify(r.Get(3).Equals(Int.Make(4)) == Bool.True);

      r.Clear();
      o = type.InvokeMember("F", GetStaticFlags(), null, null, new object[] { r, Bool.True});
      Verify(o == null);
      Verify(r.Get(0).Equals(Int.Make(0)) == Bool.True);
      Verify(r.Get(1).Equals(Int.Make(1)) == Bool.True);
      Verify(r.Get(2).Equals(Int.Make(3)) == Bool.True);
      Verify(r.Get(3).Equals(Int.Make(4)) == Bool.True);
    }

    void VerifyTryCatchFinally3()
    {
      string src = "class Foo\n" +
        "{\n" +
        "  static Void f(Int[] r, Err err)\n" +
        "  {\n" +
        "    r.add(0)\n" +
        "    try\n" +
        "    {\n" +
        "      r.add(1)\n" +
        "      if (err != null) throw err\n" +
        "      r.add(2)\n" +
        "    }\n" +
        "    catch\n" +
        "    {\n" +
        "      r.add(3)\n" +
        "      throw err\n" +
        "      r.add(4)\n" +
        "    }\n" +
        "    finally\n" +
        "    {\n" +
        "      r.add(99)\n" +
        "    }\n" +
        "  }\n" +
        "}";

      System.Type type = CompileToType(src);
      List r = new List(Sys.IntType, new Obj[0]);
      object o = type.InvokeMember("F", GetStaticFlags(), null, null, new object[] { r, null });

      Verify(o == null);
      Verify(r.Get(0).Equals(Int.Make(0))  == Bool.True);
      Verify(r.Get(1).Equals(Int.Make(1))  == Bool.True);
      Verify(r.Get(2).Equals(Int.Make(2))  == Bool.True);
      Verify(r.Get(3).Equals(Int.Make(99)) == Bool.True);

      r.Clear();
      try
      {
        o = type.InvokeMember("F", GetStaticFlags(), null, null, new object[] { r, IndexErr.Make() });
      }
      catch (System.Reflection.TargetInvocationException ex)
      {
        Err err = (ex.InnerException as Err.Val).Err();
        Verify(err.GetType() == System.Type.GetType("Fan.Sys.IndexErr"));

        Verify(r.Get(0).Equals(Int.Make(0))  == Bool.True);
        Verify(r.Get(1).Equals(Int.Make(1))  == Bool.True);
        Verify(r.Get(2).Equals(Int.Make(3))  == Bool.True);
        Verify(r.Get(3).Equals(Int.Make(99)) == Bool.True);
      }

      r.Clear();
      try
      {
        o = type.InvokeMember("F", GetStaticFlags(), null, null, new object[] { r, IOErr.Make() });
      }
      catch (System.Reflection.TargetInvocationException ex)
      {
        Err err = (ex.InnerException as Err.Val).Err();
        Verify(err.GetType() == System.Type.GetType("Fan.Sys.IOErr"));

        Verify(r.Get(0).Equals(Int.Make(0))  == Bool.True);
        Verify(r.Get(1).Equals(Int.Make(1))  == Bool.True);
        Verify(r.Get(2).Equals(Int.Make(3))  == Bool.True);
        Verify(r.Get(3).Equals(Int.Make(99)) == Bool.True);
      }
    }

    void VerifyTryCatchFinally4()
    {
      string src = "class Foo\n" +
        "{\n" +
        "  static Void f(Int[] r, Err err)\n" +
        "  {\n" +
        "    r.add(0)\n" +
        "    try\n" +
        "    {\n" +
        "      r.add(1)\n" +
        "      if (err != null) throw err\n" +
        "      r.add(2)\n" +
        "    }\n" +
        "    catch (IOErr e)\n" +
        "    {\n" +
        "      r.add(3)\n" +
        "      throw e\n" +
        "      r.add(4)\n" +
        "    }\n" +
        "    catch\n" +
        "    {\n" +
        "      r.add(5)\n" +
        "      throw err\n" +
        "      r.add(6)\n" +
        "    }\n" +
        "    finally\n" +
        "    {\n" +
        "      r.add(99)\n" +
        "    }\n" +
        "  }\n" +
        "}";

      System.Type type = CompileToType(src);
      List r = new List(Sys.IntType, new Obj[0]);
      object o = type.InvokeMember("F", GetStaticFlags(), null, null, new object[] { r, null });

      Verify(o == null);
      Verify(r.Get(0).Equals(Int.Make(0))  == Bool.True);
      Verify(r.Get(1).Equals(Int.Make(1))  == Bool.True);
      Verify(r.Get(2).Equals(Int.Make(2))  == Bool.True);
      Verify(r.Get(3).Equals(Int.Make(99)) == Bool.True);

      r.Clear();
      try
      {
        o = type.InvokeMember("F", GetStaticFlags(), null, null, new object[] { r, IOErr.Make() });
      }
      catch (System.Reflection.TargetInvocationException ex)
      {
        Err err = (ex.InnerException as Err.Val).Err();
        Verify(err.GetType() == System.Type.GetType("Fan.Sys.IOErr"));

        Verify(r.Get(0).Equals(Int.Make(0))  == Bool.True);
        Verify(r.Get(1).Equals(Int.Make(1))  == Bool.True);
        Verify(r.Get(2).Equals(Int.Make(3))  == Bool.True);
        Verify(r.Get(3).Equals(Int.Make(99)) == Bool.True);
      }

      r.Clear();
      try
      {
        o = type.InvokeMember("F", GetStaticFlags(), null, null, new object[] { r, IndexErr.Make() });
      }
      catch (System.Reflection.TargetInvocationException ex)
      {
        Err err = (ex.InnerException as Err.Val).Err();
        Verify(err.GetType() == System.Type.GetType("Fan.Sys.IndexErr"));

        Verify(r.Get(0).Equals(Int.Make(0))  == Bool.True);
        Verify(r.Get(1).Equals(Int.Make(1))  == Bool.True);
        Verify(r.Get(2).Equals(Int.Make(5))  == Bool.True);
        Verify(r.Get(3).Equals(Int.Make(99)) == Bool.True);
      }
    }

    void VerifyTryCatchFinally5() // torture test
    {
      string src = "class Foo\n" +
        "{\n" +
        "  static Void f(Int[] r)\n" +
        "  {\n" +
        "    r.add(0)\n" +
        "    try\n" +
        "    {\n" +
        "      for (i:=0; i<5; ++i)\n" +
        "      {\n" +
        "        r.add(10+i)\n" +
        "        try\n" +
        "        {\n" +
        "          try\n" +
        "          {\n" +
        "            if (i == 2) throw IOErr.make\n" +
        "            r.add(20+i)\n" +
        "          }\n" +
        "          finally\n" +
        "          {\n" +
        "            r.add(30+i)\n" +
        "          }\n" +
        "\n" +
        "          try\n" +
        "          {\n" +
        "          }\n" +
        "          finally\n" +
        "          {\n" +
        "            r.add(300+i)\n" +
        "          }\n" +
        "        }\n" +
        "        catch\n" +
        "        {\n" +
// TODO - THIS IS NOT EMITTED AS VALID IL....
        "          try\n" +
        "          {\n" +
//        "            r.add(900+i)\n" +
//        "            throw IOErr.make\n" +
//        "            r.add(910+i)\n" +
        "          }\n" +
//        "          catch {}\n" +
        "          finally {}\n" +
//        "          catch (IOErr e)\n" +
//        "          {\n" +
//        "            r.add(920+i)\n" +
//        "          }\n" +
//        "          finally\n" +
//        "          {\n" +
//        "            r.add(930+i)\n" +
//        "          }\n" +
// END TODO
        "          break\n" +
        "        }\n" +
        "        r.add(50+i)\n" +
        "      }\n" +
        "    }\n" +
        "    finally\n" +
        "    {\n" +
        "      r.add(99)\n" +
        "    }\n" +
        "    r.add(999)\n" +
        "  }\n" +
        "}";

      //r := Int[,]
      //t.method("f").call1(r)
      //verifyEq(r, [0, 10, 20, 30, 300, 50, 11, 21, 31, 301, 51, 12, 32, 902, 922, 932, 99, 999])

      System.Type type = CompileToType(src);
      List r = new List(Sys.IntType, new Obj[0]);
      object o = type.InvokeMember("F", GetStaticFlags(), null, null, new object[] { r });

      Verify(o == null);
      Verify(r.Get(0).Equals(Int.Make(0))    == Bool.True);
      Verify(r.Get(1).Equals(Int.Make(10))   == Bool.True);
      Verify(r.Get(2).Equals(Int.Make(20))   == Bool.True);
      Verify(r.Get(3).Equals(Int.Make(30))   == Bool.True);
      Verify(r.Get(4).Equals(Int.Make(300))  == Bool.True);
      Verify(r.Get(5).Equals(Int.Make(50))   == Bool.True);
      Verify(r.Get(6).Equals(Int.Make(11))   == Bool.True);
      Verify(r.Get(7).Equals(Int.Make(21))   == Bool.True);
      Verify(r.Get(8).Equals(Int.Make(31))   == Bool.True);
      Verify(r.Get(9).Equals(Int.Make(301))  == Bool.True);
      Verify(r.Get(10).Equals(Int.Make(51))  == Bool.True);
      Verify(r.Get(11).Equals(Int.Make(12))  == Bool.True);
      Verify(r.Get(12).Equals(Int.Make(32))  == Bool.True);
      Verify(r.Get(13).Equals(Int.Make(902)) == Bool.True);
      Verify(r.Get(14).Equals(Int.Make(922)) == Bool.True);
      Verify(r.Get(15).Equals(Int.Make(932)) == Bool.True);
      Verify(r.Get(16).Equals(Int.Make(99))  == Bool.True);
      Verify(r.Get(17).Equals(Int.Make(999)) == Bool.True);
    }


    void FuckMe() // torture test
    {
      string src = "class Foo\n" +
        "{\n" +
        "  static Void f()\n" +
        "  {\n" +
        "    try {}\n" +
        "    catch\n" +
        "    {\n" +
        "      try {}\n" +
        //"      catch {}\n" +
        "      finally {}\n" +
        "    }\n" +
        "  }\n" +
        "}";

      System.Type type = CompileToType(src);
      List r = new List(Sys.IntType, new Obj[0]);
      type.InvokeMember("F", GetStaticFlags(), null, null, new object[0]);
    }

  }
}