//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   27 Sep 06  Andy Frank  Creation
//

using System;
using System.Diagnostics;
using System.IO;
using System.Reflection;
using Fan.Sys;
using Fanx.Util;

namespace Fanx.Test
{
  /// <summary>
  /// UtilTest.
  /// </summary>
  public class UtilTest : Test
  {
    public override void Run()
    {
      VerifyUpper();
      VerifyGetPodName();
      VerifySplitQName();
    }

    public void VerifyUpper()
    {
      Verify(NameUtil.Upper("foo"), "Foo");
      Verify(NameUtil.Upper("Foo"), "Foo");
      Verify(NameUtil.Upper("fooBar"), "FooBar");
      Verify(NameUtil.Upper("FooBar"), "FooBar");
      Verify(NameUtil.Upper("alpha.beta.gamma"), "Alpha.Beta.Gamma");
      Verify(NameUtil.Upper("Alpha.Beta.gamma"), "Alpha.Beta.Gamma");
      Verify(NameUtil.Upper("Alpha.beta.Gamma"), "Alpha.Beta.Gamma");
      Verify(NameUtil.Upper("alpha.Beta.Gamma"), "Alpha.Beta.Gamma");
    }

    public void VerifyGetPodName()
    {
      Verify(NameUtil.GetPodName("Fan.Sys.Foo"),         "sys");
      Verify(NameUtil.GetPodName("Fan.Sys.Foo.Bar"),     "sys");
      Verify(NameUtil.GetPodName("Fan.Sys.Foo.Bar.Car"), "sys");
      Verify(NameUtil.GetPodName("Fan.Sys.Foo/Val"),     "sys");

      VerifyFail(NameUtil.GetPodName("Fan.Sys"),     "sys");
      VerifyFail(NameUtil.GetPodName("Fan.Sys.Foo"), "Sys");
      VerifyFail(NameUtil.GetPodName("Fan.Sys.Foo"), "andy");
    }

    private void VerifyFail(string a, string b)
    {
      if (a == b) Fail();
    }

    public void VerifySplitQName()
    {
      VerifyQName(NameUtil.SplitQName("Fan"), null, "Fan");
      VerifyQName(NameUtil.SplitQName("Fan.Sys"), "Fan", "Sys");
      VerifyQName(NameUtil.SplitQName("Fan.Sys.Bool"), "Fan.Sys", "Bool");
      VerifyQName(NameUtil.SplitQName("Fan.Sys.Bool.Foo"), "Fan.Sys.Bool", "Foo");
      VerifyQName(NameUtil.SplitQName("Fan.Sys.Bool.Foo/Val"), "Fan.Sys.Bool.Foo", "Val");
    }

    private void VerifyQName(string[] a, string b, string c)
    {
      Verify(a.Length == 2);
      Verify(a[0] == b);
      Verify(a[1] == c);
    }
  }
}