//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   4 Dec 05  Brian Frank  Creation
//

**
** Str represents a sequence of Unicode characters.
**
const final class Str
{

//////////////////////////////////////////////////////////////////////////
// Constructor
//////////////////////////////////////////////////////////////////////////

  **
  ** Private constructor.
  **
  private new make()

//////////////////////////////////////////////////////////////////////////
// Obj Overrides
//////////////////////////////////////////////////////////////////////////

  **
  ** Return true if a Str with exact same char sequence.
  **
  override Bool equals(Obj obj)

  **
  ** Compare based on Unicode character values.
  **
  override Int compare(Obj obj)

  **
  ** The hash for a Str is platform dependent.
  **
  override Int hash()

  **
  ** Return this.
  **
  override Str toStr()

//////////////////////////////////////////////////////////////////////////
// Identity
//////////////////////////////////////////////////////////////////////////

  **
  ** Return if size() == 0.
  **
  Bool isEmpty()

  **
  ** Return number of characters in this string.
  **
  Int size()

  **
  ** Internalize this Str such that two strings which are equal
  ** via the == operator will have the same reference such that
  ** === will be true.
  **
  Str intern()

  **
  ** Return if this Str starts with the specified Str.
  **
  Bool startsWith(Str s)

  **
  ** Return if this Str ends with the specified Str.
  **
  Bool endsWith(Str s)

  **
  ** Return the first occurance of the specified substring searching
  ** forward, starting at the specified offset index.  A negative offset
  ** may be used to access from the end of string.  Return null if no
  ** occurences are found.
  **
  ** Examples:
  **   "abcabc".index("b")     -> 1
  **   "abcabc".index("b", 1)  -> 1
  **   "abcabc".index("b", 3)  -> 4
  **   "abcabc".index("b", -3) -> 4
  **   "abcabc".index("x")     -> null
  **
  Int index(Str s, Int offset := 0)

  **
  ** Reverse index - return the first occurance of the specified
  ** substring searching backward, starting at the specified offset
  ** index.  A negative offset may be used to access from the end
  ** of string.  Return null if no occurences are found.
  **
  ** Examples:
  **   "abcabc".indexr("b")     -> 4
  **   "abcabc".indexr("b", -3) -> 1
  **   "abcabc".indexr("b", 0)  -> null
  **
  Int indexr(Str s, Int offset := -1)

  **
  ** Return if this string contains the specified string.
  ** Convenience for index(s) != null
  **
  Bool contains(Str s)

//////////////////////////////////////////////////////////////////////////
// Operators
//////////////////////////////////////////////////////////////////////////

  **
  ** Get the character at the zero based index as a Unicode code point.
  ** Negative indexes may be used to access from the end of the string.
  ** This method is accessed via the [] operator.
  **
  Int get(Int index)

  **
  ** Return a substring based on the specified range.  Negative indexes
  ** may be used to access from the end of the string.  This method
  ** is accessed via the [] operator.  Throw IndexErr if range illegal.
  **
  ** Examples:
  **   "abcd"[0..2]   -> "abc"
  **   "abcd"[3..3]   -> "d"
  **   "abcd"[-2..-1] -> "cd"
  **   "abcd"[0...2]  -> "ab"
  **   "abcd"[1..-2]  -> "bc"
  **   "abcd"[4..-1]  -> ""
  **
  Str slice(Range range)

  **
  ** Concat the value of obj.toStr
  **
  Str plus(Obj obj)

//////////////////////////////////////////////////////////////////////////
// Iterators
//////////////////////////////////////////////////////////////////////////

  **
  ** Call the specified method for every char in the starting
  ** with index 0 and incrementing up to size-1.  This method is
  ** idempotent.
  **
  ** Example:
  **   "abc".each |Int c| { echo(c.toChar) }
  **
  Void each(|Int ch, Int index| c)

  **
  ** Reverse each - call the specified method for every char in
  ** the string starting with index size-1 and decrementing down
  ** to 0.  This method is idempotent.
  **
  ** Example:
  **   "abc".eachr |Int c| { echo(c.toChar) }
  **
  Void eachr(|Int ch, Int index| c)

//////////////////////////////////////////////////////////////////////////
// Utils
//////////////////////////////////////////////////////////////////////////

  **
  ** Get the a Str containing the specified number of spaces.  Also
  ** see Str.justl and Str.justr to justify an existing string.
  **
  ** Examples:
  **   Str.spaces(1) -> " "
  **   Str.spaces(2) -> "  "
  **
  static Str spaces(Int n)

  **
  ** Return this string with all upper case characters replaced to lower
  ** case.  The case conversion is for ASCII only - see Int.lower.
  **
  ** Example:
  **   "Apple".lower -> "apple"
  **
  Str lower()

  **
  ** Return this string with all lower case characters replaced to upper
  ** case.  The case conversion is for ASCII only - see Int.upper.
  **
  ** Example:
  **   "Foo Bar".upper -> "FOO BAR"
  **
  Str upper()

  **
  ** Return this string with the first character converted upper case.
  ** The case conversion is for ASCII only - see Int.upper.
  **
  ** Example:
  **   "foo".capitalize -> "Foo"
  **
  Str capitalize()

  **
  ** Return this string with the first character converted lower case.
  ** The case conversion is for ASCII only - see Int.lower.
  **
  ** Example:
  **   "Foo".decapitalize -> "foo"
  **
  Str decapitalize()

  **
  ** If size is less than width, then add spaces to the right
  ** to create a left justified string.
  **
  ** Examples:
  **   "xyz".justl(2) -> "xyz"
  **   "xyz".justl(4) -> "xyz "
  **
  Str justl(Int width)

  **
  ** If size is less than width, then add spaces to the left
  ** to create a right justified string.
  **
  ** Examples:
  **   "xyz".justr(2) -> "xyz"
  **   "xyz".justr(4) -> " xyz"
  **
  Str justr(Int width)

  **
  ** Reverse the contents of this string.
  **
  ** Example:
  **   "stressed".reverse -> "desserts"
  **
  Str reverse()

  **
  ** Trim whitespace from the beginning and end of the string.  For the purposes
  ** of this method, whitespace is defined as any character equal to or less
  ** than the 0x20 space character (including ' ', '\r', '\n', and '\t').
  **
  ** Examples:
  **    "foo".trim -> "foo"
  **    "  foo".trim -> "foo"
  **    "  foo\n".trim -> "foo"
  **
  Str trim()

  **
  ** Split a string into a list of substrings using the given separators
  ** TODO: just temp hack
  **
  Str[] split(Str separators)

  **
  ** Replace all occurances of from with to.
  ** TODO: just temp hack
  **
  Str replace(Str from, Str to)

//////////////////////////////////////////////////////////////////////////
// Coersions
//////////////////////////////////////////////////////////////////////////

  **
  ** Convenience for Bool.parse(this, checked).
  **
  Bool toBool(Bool checked := true)

  **
  ** Convenience for Int.parse(this, radix, checked).
  **
  Int toInt(Int radix := 10, Bool checked := true)

  **
  ** Convenience for Float.parse(this, checked).
  **
  Float toFloat(Bool checked := true)

  **
  ** Convenience for Uri.parse(this).
  **
  Uri toUri()

}