//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   3 Jul 06  Brian Frank  Creation
//

**
** Time represents an absolute instance in time.  Fan time is normalized
** as nanosecond ticks since 1 Jan 1970  GMT.
**
const final class Time
{

//////////////////////////////////////////////////////////////////////////
// Constructor
//////////////////////////////////////////////////////////////////////////

  **
  ** Return current time.
  **
  static Time now()

  **
  ** Make for nanosecond ticks since 1 Jan 1970 GMT.
  **
  new make(Int ticks)

//////////////////////////////////////////////////////////////////////////
// Identity
//////////////////////////////////////////////////////////////////////////

  **
  ** Two times are equal if have identical nanosecond ticks.
  **
  override Bool equals(Obj that)

  **
  ** Return nanosecond ticks for the hashcode.
  **
  override Int hash()

  **
  ** Compare based on nanosecond ticks.
  **
  override Int compare(Obj obj)

  **
  ** Return string representation.
  **
  override Str toStr()

//////////////////////////////////////////////////////////////////////////
// Access
//////////////////////////////////////////////////////////////////////////

  **
  ** Return number of nanosecond ticks since 1 Jan 1970.
  **
  Int ticks()

  // TODO: just about everything

//////////////////////////////////////////////////////////////////////////
// Operators
//////////////////////////////////////////////////////////////////////////

  **
  ** Return the delta between this and the given time.
  **
  ** Example:
  **   elapsed := Time.now - startTime
  **
  Duration minus(Time time)

  **
  ** Add (or subtract if duration is negative) a duration
  ** to compute a new time.
  **
  ** Example:
  **   tomorrow  := Time.now + 1days
  **   yesterday := Time.now + -1days
  **
  Time plus(Duration duration)

}