//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   27 Mar 06  Brian Frank  Creation
//

**
** CharsetTest
**
class CharsetTest : Test
{

  Void testIdentity()
  {
    // verify name, toStr, equals, hash

    c1 := Charset.make("UTF-8")
    verifyEq(c1.type, Charset.type)
    verifyEq(c1.type.base, Obj.type)
    verifyEq(c1.name,  "UTF-8")
    verifyEq(c1.toStr, "UTF-8")

    c2 := Charset.make("utf-8")
    verifyEq(c2.name,  "UTF-8")
    verifyEq(c2.toStr, "UTF-8")

    verifyEq(c1, c2)
    verifyEq(c1.hash, c2.hash)
    verifyNotEq(c1, "foo")

    // verify invalid name
    verifyErr(ArgErr.type) |,| { Charset.make("bogus") }
    verifyErr(ArgErr.type) |,| { Charset.make("*^%#!%") }
  }

  Void testStandard()
  {
    // verify standard char encodings
    verifyEq(Charset.utf8().name,   "UTF-8")
    verifySame(Charset.make("UTF-8"), Charset.utf8())
    verifySame(Charset.make("utf-8"), Charset.utf8())

    verifyEq(Charset.utf16BE().name, "UTF-16BE")
    verifySame(Charset.make("UTF-16BE"), Charset.utf16BE())
    verifySame(Charset.make("utf-16be"), Charset.utf16BE())

    verifyEq(Charset.utf16LE().name, "UTF-16LE")
    verifySame(Charset.make("UTF-16LE"), Charset.utf16LE())
    verifySame(Charset.make("utf-16le"), Charset.utf16LE())
  }

}