//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   29 May 06  Brian Frank  Creation
//

**
** ErrTest
**
class ErrTest : Test
{

//////////////////////////////////////////////////////////////////////////
// Type
//////////////////////////////////////////////////////////////////////////

  Void testType()
  {
    err := Err.make
    verifySame(err.type, Err.type)
    verifySame(err.type.base, Obj.type)
    verifyEq(err.type.qname, "sys::Err")
    verify(err is Err)
    verify(err is Obj)
  }

//////////////////////////////////////////////////////////////////////////
// Trace
//////////////////////////////////////////////////////////////////////////

  Void testTrace()
  {
    err := Err.make("foo")
    buf := Buf.make
    err.trace(buf)
    verify(buf.size > 0)
    buf.flip
    str := buf.readAllStr
    verify(str.index("testTrace") != null)
  }

//////////////////////////////////////////////////////////////////////////
// Consturctors
//////////////////////////////////////////////////////////////////////////

  Void testCtor()
  {
    cause := Err.make

    err := Err.make
    verifyEq(err.message, null)
    verifyEq(err.cause, null)
    verifyEq(err.toStr, "sys::Err")

    err = Err.make("foo")
    verifyEq(err.message, "foo")
    verifyEq(err.cause, null)
    verifyEq(err.toStr, "sys::Err: foo")

    err = Err.make("foo", cause)
    verifyEq(err.message, "foo")
    verifySame(err.cause, cause)
    verifyEq(err.toStr, "sys::Err: foo")
  }

//////////////////////////////////////////////////////////////////////////
// All Sys Errs
//////////////////////////////////////////////////////////////////////////

  Void testAllSysErrs()
  {
    cause := Err.make
    Err err

    // ArgErr
    err = ArgErr.make("msg", cause)
    verifySame(err.type, ArgErr.type)
    verifySame(err.type.base, Err.type)
    verifyEq(err.type.qname, "sys::ArgErr")
    verify(err is ArgErr)
    verify(err is Err)
    verify(err is Obj)
    verify(err.message === "msg")
    verify(err.cause === cause)
    err = ArgErr.make
    verify(err.message === null)
    verify(err.cause === null)

    // CastErr
    err = CastErr.make("msg", cause)
    verifySame(err.type, CastErr.type)
    verifySame(err.type.base, Err.type)
    verifyEq(err.type.qname, "sys::CastErr")
    verify(err is CastErr)
    verify(err is Err)
    verify(err is Obj)
    verify(err.message === "msg")
    verify(err.cause === cause)
    err = CastErr.make
    verify(err.message === null)
    verify(err.cause === null)

    // IndexErr
    err = IndexErr.make("msg", cause)
    verifySame(err.type, IndexErr.type)
    verifySame(err.type.base, Err.type)
    verifyEq(err.type.qname, "sys::IndexErr")
    verify(err is IndexErr)
    verify(err is Err)
    verify(err is Obj)
    verify(err.message === "msg")
    verify(err.cause === cause)
    err = IndexErr.make
    verify(err.message === null)
    verify(err.cause === null)

    // IOErr
    err = IOErr.make("msg", cause)
    verifySame(err.type, IOErr.type)
    verifySame(err.type.base, Err.type)
    verifyEq(err.type.qname, "sys::IOErr")
    verify(err is IOErr)
    verify(err is Err)
    verify(err is Obj)
    verify(err.message === "msg")
    verify(err.cause === cause)
    err = IOErr.make
    verify(err.message === null)
    verify(err.cause === null)

    // NullErr
    err = NullErr.make("msg", cause)
    verifySame(err.type, NullErr.type)
    verifySame(err.type.base, Err.type)
    verifyEq(err.type.qname, "sys::NullErr")
    verify(err is NullErr)
    verify(err is Err)
    verify(err is Obj)
    verify(err.message === "msg")
    verify(err.cause === cause)
    err = NullErr.make
    verify(err.message === null)
    verify(err.cause === null)

    // ReadonlyErr
    err = ReadonlyErr.make("msg", cause)
    verifySame(err.type, ReadonlyErr.type)
    verifySame(err.type.base, Err.type)
    verifyEq(err.type.qname, "sys::ReadonlyErr")
    verify(err is ReadonlyErr)
    verify(err is Err)
    verify(err is Obj)
    verify(err.message === "msg")
    verify(err.cause === cause)
    err = ReadonlyErr.make
    verify(err.message === null)
    verify(err.cause === null)

    // UnknownPodErr
    err = UnknownPodErr.make("msg", cause)
    verifySame(err.type, UnknownPodErr.type)
    verifySame(err.type.base, Err.type)
    verifyEq(err.type.qname, "sys::UnknownPodErr")
    verify(err is UnknownPodErr)
    verify(err is Err)
    verify(err is Obj)
    verify(err.message === "msg")
    verify(err.cause === cause)
    err = UnknownPodErr.make
    verify(err.message === null)
    verify(err.cause === null)

    // UnknownSlotErr
    err = UnknownSlotErr.make("msg", cause)
    verifySame(err.type, UnknownSlotErr.type)
    verifySame(err.type.base, Err.type)
    verifyEq(err.type.qname, "sys::UnknownSlotErr")
    verify(err is UnknownSlotErr)
    verify(err is Err)
    verify(err is Obj)
    verify(err.message === "msg")
    verify(err.cause === cause)
    err = UnknownSlotErr.make
    verify(err.message === null)
    verify(err.cause === null)

    // UnknownTypeErr
    err = UnknownTypeErr.make("msg", cause)
    verifySame(err.type, UnknownTypeErr.type)
    verifySame(err.type.base, Err.type)
    verifyEq(err.type.qname, "sys::UnknownTypeErr")
    verify(err is UnknownTypeErr)
    verify(err is Err)
    verify(err is Obj)
    verify(err.message === "msg")
    verify(err.cause === cause)
    err = UnknownTypeErr.make
    verify(err.message === null)
    verify(err.cause === null)

    // UnsupportedErr
    err = UnsupportedErr.make("msg", cause)
    verifySame(err.type, UnsupportedErr.type)
    verifySame(err.type.base, Err.type)
    verifyEq(err.type.qname, "sys::UnsupportedErr")
    verify(err is UnsupportedErr)
    verify(err is Err)
    verify(err is Obj)
    verify(err.message === "msg")
    verify(err.cause === cause)
    err = UnsupportedErr.make
    verify(err.message === null)
    verify(err.cause === null)
  }

//////////////////////////////////////////////////////////////////////////
// Subclassing
//////////////////////////////////////////////////////////////////////////

  Void testSubclassing()
  {
    cause := Err.make

    // create TestOneErr with both params
    err := TestOneErr.make("msg", cause)
    verifySame(err.type, TestOneErr.type)
    verifySame(err.type.base, Err.type)
    verifyEq(err.type.qname, "sysTest::TestOneErr")
    verify(err is TestOneErr)
    verify(err is Err)
    verify(err is Obj)
    verify(err.message === "msg")
    verify(err.cause === cause)
    verifyEq(err.r, -3f)

    // verify TestOneErr with 2 default params
    err = TestOneErr.make()
    verifySame(err.type, TestOneErr.type)
    verify(err.message === null)
    verify(err.cause === null)

    // verify TestOneErr with 1 default params
    err = TestOneErr.make("foobar")
    verifySame(err.type, TestOneErr.type)
    verify(err.message === "foobar")
    verify(err.cause === null)

    // verify TestTwoErr which subclasses from TestOneErr
    err2 := TestTwoErr.make()
    verifySame(err2.type, TestTwoErr.type)
    verifySame(err2.type.base, TestOneErr.type)
    verifySame(err2.type.base.base, Err.type)
    verifyEq(err2.type.qname, "sysTest::TestTwoErr")
    verify(err2 is TestTwoErr)
    verify(err2 is TestOneErr)
    verify(err is Err)
    verify(err is Obj)
    verifyEq(err2.r, -3f)
    verifyEq(err2.i, 77)
    verifyEq(err2.s, "hello world")

    // verify TestIOErr which subclasses from IOErr
    errIO := TestIOErr.make()
    verifySame(errIO.type, TestIOErr.type)
    verifySame(errIO.type.base, IOErr.type)
    verifySame(errIO.type.base.base, Err.type)
    verifyEq(errIO.type.qname, "sysTest::TestIOErr")
    verify(errIO is TestIOErr)
    verify(errIO is IOErr)
    verify(errIO is Err)
    verify(errIO is Obj)
    verifyEq(errIO.s, "memorial day")

    // verify throws works correctly
    verifyErr(TestOneErr.type) |,| { throw TestOneErr.make }
    verifyErr(TestTwoErr.type) |,| { throw TestTwoErr.make }
    verifyErr(TestIOErr.type)  |,| { throw TestIOErr.make }
  }

}

//////////////////////////////////////////////////////////////////////////
// Supplemental classes
//////////////////////////////////////////////////////////////////////////

class TestOneErr : Err
{
  new make(Str msg := null, Err cause := null) : super(msg, cause) {}
  Float r := -3f
}

class TestTwoErr : TestOneErr
{
  new make(Str msg := null, Err cause := null) : super(msg, cause) {}

  Int i := 77
  Str s := "hello world"
}

class TestIOErr : IOErr
{
  new make(Str msg := null, Err cause := null) : super(msg, cause) {}

  Str s := "memorial day"
}