//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   15 Mar 06  Andy Frank  Creation
//

**
** WebReq encapsulates a web request.
**
abstract class WebReq
{

//////////////////////////////////////////////////////////////////////////
// Public
//////////////////////////////////////////////////////////////////////////

  **
  ** The WebEnv for this request.
  **
  abstract WebEnv env()

  **
  ** The HTTP request method in uppercase. Example: GET, POST, PUT.
  **
  **
  abstract Str method()

  **
  ** The request URI.
  **
  ** Examples:
  **   http://www.foo.com/a/b/c
  **   http://www.foo.com/a?q=bar
  **
  abstract Uri uri()

  **
  ** The prefixUri is the portion of the request URI controlled
  ** by the web server.  A prefixUri always ends with a trailing
  ** slash.
  **
  **   uri == prefixUri + suffixUri
  **
  ** Examples:
  **         uri: http://www.foo.com/a/b/c
  **   prefixUri: http://www.foo.com/a/
  **   suffixUri: b/c
  **
  **         uri: http://www.foo.com/a/b/c
  **   prefixUri: http://www.foo.com/
  **   suffixUri: a/b/c
  **
  **         uri: http://www.foo.com/a/b/c?q=bar
  **   prefixUri: http://www.foo.com/a/b/
  **   suffixUri: c?q=bar
  **
  abstract Uri prefixUri()

  **
  ** The suffixUri is the portion of the request URI not
  ** controlled by the web server.  The suffixUri will always
  ** starts after the trailing slash of the prefixUri.
  **
  **   uri == prefixUri + suffixUri
  **
  ** Examples:
  **         uri: http://www.foo.com/a/b/c
  **   prefixUri: http://www.foo.com/a/
  **   suffixUri: b/c
  **
  **         uri: http://www.foo.com/a/b/c
  **   prefixUri: http://www.foo.com/
  **   suffixUri: a/b/c
  **
  **         uri: http://www.foo.com/a/b/c?q=bar
  **   prefixUri: http://www.foo.com/a/b/
  **   suffixUri: c?q=bar
  **
  abstract Uri suffixUri()

  **
  ** Map of HTTP request headers.
  **
  abstract Str:Str headers()

  **
  ** The UserAgent for this request.
  **
  abstract UserAgent userAgent()

  **
  ** Stash allows you to stash objects on the WebReq object
  ** in order to pass data b/w Weblets on the same request.
  **
  abstract Str:Obj stash()

  **
  ** The InStream for this request.
  **
  abstract InStream in()

//////////////////////////////////////////////////////////////////////////
// Debug
//////////////////////////////////////////////////////////////////////////

  **
  ** Print debug info for this request to the OutStream.
  **
  // TODO - use sys::Log
  Void dump(OutStream out := Sys.out)
  {
    out.printLine("--- WebReq.dump ---")
    out.printLine("  method: $method")
    out.printLine("  uri: $uri")
    out.printLine("  prefixUri: $prefixUri")
    out.printLine("  suffixUri: $suffixUri")
    out.printLine("  -- Headers --")
    headers.each |Str v, Str k|
    {
      out.printLine("  $k: $v")
    }
  }

}