//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   15 Mar 06  Andy Frank  Creation
//

**
** Httpd is a simple HTTP server.
**
class Httpd : WebEnv
{

//////////////////////////////////////////////////////////////////////////
// Main
//////////////////////////////////////////////////////////////////////////

  **
  ** Start server.
  **
  static Void main() { Httpd.make.start }

//////////////////////////////////////////////////////////////////////////
// WebEnv
//////////////////////////////////////////////////////////////////////////

  **
  ** Return the prefix URI this server is running under.
  **
  override Uri prefixUri() { return `/` }

  **
  ** Return the product name and version.
  **
  override Str product() { return "fan-httpd/1.0" }

  **
  ** Log this message.
  **
  override Void log(Str msg, Err err := null)
  {
    echo("[$timestamp]  $msg")
    if (err != null) echo(err)
  }

//////////////////////////////////////////////////////////////////////////
// Methods
//////////////////////////////////////////////////////////////////////////

  **
  ** Start server.
  **
  native Void start()

  **
  ** Stop server.
  **
  native Void stop()

  **
  ** Get the current timestamp.
  **
  native Str timestamp()

//////////////////////////////////////////////////////////////////////////
// Callbacks
//////////////////////////////////////////////////////////////////////////

  **
  ** Service a request.
  **
  Void service(InStream in, WebOutStream out)
  {
    // create req/res
    req := HttpdWebReq.make(this, in)
    res := HttpdWebRes.make(this, out);

    // parse req
    line := req.in.readLine

    // trace request info
    start := Duration.now
    log(line)

    try
    {
      // request method and uri
      s := line.split(" ")
      req._method = s[0].upper
      req._uri       = Uri.parse(s[1])
      req._prefixUri = prefixUri
      req._suffixUri = req.uri - prefixUri

      // request headers
      line = req.in.readLine
      while (line.size > 0)
      {
        index := line.index(":")
        key := line[0..index-1].trim
        val := line[index+1..-1].trim
        req._headers[key] = val
        if (key == "User-Agent")
          req._userAgent = UserAgent.make(val)
        line = req.in.readLine
      }

      // debug
      //req.dump

      // server specific response headers
      res.headers["Connection"] = "Close"
      res.headers["Server"] = product

      // cache req/res instances
      Thread.locals["web.req"] = req
      Thread.locals["web.res"] = res

      // call weblet
      type   := Type.find("fandoc::FandocWeblet")
      weblet := (Weblet)type.method("make").call1(null)
      weblet.service
    }
    catch (Err err)
    {
      // print stack trace on server
      err.trace

      // and also send stack trace to client
      res.out.writeChars("<h1>$err.toStr</h1><hr/><pre>")
      err.trace(res.out)
      res.out.writeChars("</pre>")
    }

    // trace response info
    end := Duration.now
    sc  := res._sc
    log("-> $sc ${WebRes.statusMsg[sc]} (${(end-start).toMillis}ms)", null)
  }

}