//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   15 Mar 06  Andy Frank  Creation
//

**
** Httpd implementation of WebRes.
**
internal class HttpdWebRes : WebRes
{

//////////////////////////////////////////////////////////////////////////
// Constructor
//////////////////////////////////////////////////////////////////////////

  new make(WebEnv env, WebOutStream out)
  {
    _env = env
    _out = out
  }

//////////////////////////////////////////////////////////////////////////
// WebRes
//////////////////////////////////////////////////////////////////////////

  override WebEnv env() { return _env }

  override Void statusCode(Int sc)
  {
    if (_isCommitted) throw Err.make("Response already committed")
    if (statusMsg[sc] == null) throw Err.make("Unknown status code: $sc")
    _sc = sc
  }

  override Str:Str headers() { return _headers }
  override Bool isCommitted() { return _isCommitted }

  override WebOutStream out()
  {
    if (!_isCommitted)
    {
      _isCommitted = true
      _headers = _headers.ro
      sc := _sc // TEMP - to get around map bug
      _out.w("HTTP/1.1 $_sc ${statusMsg[sc]}\r\n")
      _headers.each |Str v, Str k| { _out.w("$k: $v\r\n") }
      _out.w("\r\n")
    }
    return _out
  }

  override Void redirect(Int statusCode, Uri uri)
  {
    // User WebRes API to check for commit, and then commit
    this.statusCode(statusCode)
    headers["Location"] = uri.toStr
    headers["Content-Length"] = "0"
    out
  }

  override Void sendError(Int statusCode, Str msg := null)
  {
    // User WebRes API to check for commit, and then commit
    this.statusCode(statusCode)
    headers["Content-Type"] = "text/html"

    out.docType
    out.html
    out.head.title("$statusCode ${statusMsg[statusCode]}").headEnd
    out.body
    out.h1(statusMsg[statusCode])
    if (msg != null)
      out.w(msg).nl
    out.hr
    out.w("<i>$env.product</i>").nl
    out.bodyEnd
    out.htmlEnd
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  internal WebEnv _env
  internal WebOutStream _out
  internal Bool _isCommitted := false
  internal Int _sc := 200
  internal Str:Str _headers := Str:Str[:]

}