//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   15 Mar 06  Andy Frank  Creation
//
package fan.web;

import java.lang.Thread;
import java.io.*;
import java.net.*;
import java.util.*;
import fan.sys.*;

/**
 * Httpd native integration.
 */
public class HttpdPeer
{
//////////////////////////////////////////////////////////////////////////
// Peer
//////////////////////////////////////////////////////////////////////////

  public static HttpdPeer make(Httpd d)
  {
    return new HttpdPeer();
  }

//////////////////////////////////////////////////////////////////////////
// Native Methods
//////////////////////////////////////////////////////////////////////////

  public void start(Httpd httpd)
  {
    try
    {
      int port = 80;
      String product = httpd.product().val;

      ServerSocket socket = new ServerSocket(port);
      httpd.log(Str.make(product + " started on port " + port));
      httpd.log(Str.make("Type 'quit' to stop server"));

      new KeyboardListener(httpd).start();

      while (true)
      {
        try
        {
          Socket conn = socket.accept();
          BufferedInputStream in = new BufferedInputStream(conn.getInputStream());
          BufferedOutputStream out = new BufferedOutputStream(conn.getOutputStream());
          service(httpd, in, out);
        }
        catch (Exception e) { e.printStackTrace(); }
      }
    }
    catch (Exception e) { e.printStackTrace(); }
  }

  public void stop(Httpd httpd)
  {
    httpd.log(Str.make("Stopping server..."));
    httpd.log(Str.make("Stopped"));
    System.exit(0);
  }

  public Str timestamp(Httpd httpd)
  {
    Date d  = new Date();
    String s = (d.getYear()+1900) + "-" + pad(d.getMonth()+1) + "-" + pad(d.getDate());
    return Str.make(s + " " + pad(d.getHours()) + ":" + pad(d.getMinutes()) + ":" + pad(d.getSeconds()));
  }

//////////////////////////////////////////////////////////////////////////
// Methods
//////////////////////////////////////////////////////////////////////////

  private String pad(int v)
  {
    return (v < 10) ? "0" + v : "" + v;
  }

  private void service(Httpd httpd, InputStream in, OutputStream out)
    throws Exception
  {
    httpd.service(new SysInStream(in), new WebSysOutStream(out));
    out.flush();
    out.close();
  }

//////////////////////////////////////////////////////////////////////////
// KeyboardListener
//////////////////////////////////////////////////////////////////////////

  class KeyboardListener extends Thread
  {
    public KeyboardListener(Httpd httpd)
    {
      this.httpd = httpd;
    }

    public void run()
    {
      try
      {
        BufferedReader in = new BufferedReader(new InputStreamReader(System.in));
        while (true)
        {
          String s = in.readLine();
          if (s.equals("quit")) httpd.stop();
        }
      }
      catch (Exception e) { e.printStackTrace(); }
    }

    private Httpd httpd;
  }

}